(*
 * Coq code for the paper
 *
 *       A Syntactic Approach to Foundational Proof-Carrying Code
 *                            
 *
 *   by Zhong Shao, Valery Trifonov, Stefan Monnier, Nadeem A. Hamid, 
 *                          and Zhaozhong Ni
 *
 * (for Coq version 7.3)
 *)


(***************************************************************************)
(* Simplified Typed TAL *)
(***************************************************************************)

(***************************************************************************)
(* Utilities *)
Require Export Coq.Lists.List.
Require Export map.


(***************************************************************************)
(* Syntax of TAL *)

Require Export talreg.

Notation label := nat (only parsing).
Notation int := nat (only parsing).
Notation initflag := bool (only parsing).


(* the allocation pointer register is not merged into the normal register
   file - it just contains a label with a special type *)
Definition AP := nat.   
Inductive APTy : Set :=
  | fresh : APTy
  | used : nat -> APTy.


(* lifted types *)
Inductive OmegaL : nat -> Set :=
  | varL : forall i : nat, OmegaL (S i)
  | liftL : forall i : nat, OmegaL i -> OmegaL (S i)
  | inttyL : OmegaL 0
  | codetyL : forall i : nat, OmegaL_Map i -> APTy -> OmegaL i
  | tuptyL : forall i : nat, OmegaL_List i -> list bool -> OmegaL i
  | rectyL : forall i : nat, OmegaL (S i) -> OmegaL i
with OmegaL_Map : nat -> Set :=
  | memptyL : forall i : nat, OmegaL_Map i
  | melemL : forall i : nat, Reg -> OmegaL i -> OmegaL_Map i -> OmegaL_Map i
with OmegaL_List : nat -> Set :=
  | nilL : forall i : nat, OmegaL_List i
  | consL : forall i : nat, OmegaL i -> OmegaL_List i -> OmegaL_List i.

Notation OmegaR := (OmegaL 1) (only parsing).

(* the actual types *)
Inductive Omega : Set :=
  | intty : Omega
  | codety : Map Reg Omega -> APTy -> Omega
  | tupty : list Omega -> list bool -> Omega
  | recty : OmegaL 1 -> Omega.

Load "rftal".

Notation RegFileTy := (Map Reg Omega) (only parsing).

Inductive WordVal : Set :=
  | wl : nat -> WordVal
  | wi : nat -> WordVal
  | wuninit : Omega -> WordVal
  | wfold : WordVal -> Omega -> WordVal.


Inductive Instr : Set :=
  | add : Reg -> Reg -> Reg -> Instr
  | addi : Reg -> Reg -> nat -> Instr
  | alloc : Reg -> list Omega -> Instr
  | bgt : Reg -> Reg -> nat -> Instr
  | bump : nat -> Instr
  | fold : Reg -> Omega -> Reg -> Instr
  | ld : Reg -> Reg -> nat -> Instr
  | mov : Reg -> Reg -> Instr
  | movi : Reg -> nat -> Instr
  | movl : Reg -> nat -> Instr
  | st : Reg -> nat -> Reg -> Instr
  | unfold : Reg -> Reg -> Instr.

Inductive InstrSeq : Set :=
  | iseq : Instr -> InstrSeq -> InstrSeq
  | jd : nat -> InstrSeq
  | jmp : Reg -> InstrSeq.


Inductive HeapVal : Set :=
  | tuple : list WordVal -> HeapVal
  | code : Map Reg Omega -> APTy -> InstrSeq -> HeapVal.


(* RegFile environment ---------------*)
Definition RegFile := Map Reg WordVal.

(* RegFileTy defined with Omega above *)


(* Heap environment ------------------*)
Definition Heap := WFMap nat HeapVal.

Notation hunwrap := (unwrapMap nat HeapVal) (only parsing).
Notation hsize := (msize nat HeapVal) (only parsing).
Notation hindom := (mindom nat HeapVal) (only parsing).
Notation hnindom := (mnotindom nat HeapVal) (only parsing).
Notation hlookup := (mlookup nat HeapVal) (only parsing).
Notation hupdate := (mupdate nat HeapVal) (only parsing).
Notation hextend := (mextend nat HeapVal) (only parsing).
Notation hupdext := (mupdext nat HeapVal) (only parsing).


(* Heap type environment -------------*)
Definition HeapTy := WFMap nat Omega.

Notation htunwrap := (unwrapMap nat Omega) (only parsing).
Notation htsize := (msize nat Omega) (only parsing).
Notation htindom := (mindom nat Omega) (only parsing).
Notation htnindom := (mnotindom nat Omega) (only parsing).
Notation htlookup := (mlookup nat Omega) (only parsing).
Notation htupdate := (mupdate nat Omega) (only parsing).
Notation htextend := (mextend nat Omega) (only parsing).
Notation htupdext := (mupdext nat Omega) (only parsing).


(* Register stuff ------------------- *)
Notation reglookup := (mlookup _ _) (only parsing).
Notation regupdext := (mupdext _ _) (only parsing).


(* TAL Program -----------------------*)
Definition Program := (Heap * (RegFile * (AP * InstrSeq)))%type.


(***************************************************************************)
(* more stuff for the heap *)
Notation HeapMap := (Map nat HeapVal) (only parsing).

Inductive OrdHeap : Map nat HeapVal -> Prop :=
  | ordheap0 : OrdHeap (mempty _ _)
  | ordheap1 :
      forall (l : nat) (hv : HeapVal) (h : Map nat HeapVal),
      msize _ _ h l -> OrdHeap h -> OrdHeap (melem _ _ l hv h).

Inductive PropSubHeap : Map nat HeapVal -> Map nat HeapVal -> Prop :=
  | propsubh0 :
      forall (h : Map nat HeapVal) (l : nat) (hv : HeapVal),
      PropSubHeap h (melem _ _ l hv h)
  | propsubh1 :
      forall (h h' : Map nat HeapVal) (l : nat) (hv : HeapVal),
      PropSubHeap h h' -> PropSubHeap h (melem _ _ l hv h').

Inductive SubHeap : Map nat HeapVal -> Map nat HeapVal -> Prop :=
  | subheap0 : forall h : Map nat HeapVal, SubHeap h h
  | subheap1 :
      forall h h' : Map nat HeapVal, PropSubHeap h h' -> SubHeap h h'.




(***************************************************************************)
(* length and positions of an instruction sequence *)
(* and other utilities *)
Inductive ISubDepth : InstrSeq -> InstrSeq -> nat -> Prop :=
  | isubd0 : forall I : InstrSeq, ISubDepth I I 0
  | isubdS :
      forall (I I' : InstrSeq) (i : Instr) (n : nat),
      ISubDepth I' I n -> ISubDepth I' (iseq i I) (S n).

Fixpoint lenInstrSeq (I : InstrSeq) : nat :=
  match I with
  | iseq _ I' => S (lenInstrSeq I')
  | _ => 1
  end.

Fixpoint getInstrPos (n : nat) (I : InstrSeq) {struct n} : InstrSeq :=
  match n with
  | O => I
  | S n' => match I with
            | iseq _ I' => getInstrPos n' I'
            | _ => I
            end
  end.

Inductive ListNth (A : Set) : list A -> nat -> A -> Prop :=
  | listnth0 : forall (V : list A) (a : A), ListNth A (a :: V) 0 a
  | listnth1 :
      forall (V : list A) (a a' : A) (n : nat),
      ListNth A V n a' -> ListNth A (a :: V) (S n) a'.


Fixpoint makeUninitTup (V : list Omega) : list WordVal :=
  match V with
  | nil => nil (A:=WordVal)
  | t :: V' => wuninit t :: makeUninitTup V'
  end.

Fixpoint makeUninitTupty (V : list Omega) : list bool :=
  match V with
  | nil => nil (A:=bool)
  | t :: V' => false :: makeUninitTupty V'
  end.

Inductive updatetuple :
list WordVal -> nat -> WordVal -> list WordVal -> Prop :=
  | updtup0 :
      forall (V : list WordVal) (w w' : WordVal),
      updatetuple (w :: V) 0 w' (w' :: V)
  | updtup1 :
      forall (V V' : list WordVal) (w w' : WordVal) (i : nat),
      updatetuple V i w' V' -> updatetuple (w :: V) (S i) w' (w :: V').


Inductive updatetupty (* used for static semantics *)
: list bool -> nat -> list bool -> Prop :=
  | updtpt0 :
      forall (V : list bool) (b : bool), updatetupty (b :: V) 0 (true :: V)
  | updtpt1 :
      forall (V V' : list bool) (b : bool) (i : nat),
      updatetupty V i V' -> updatetupty (b :: V) (S i) (b :: V').




(***************************************************************************)
(* Operational Semantics *)
Inductive Eval : Program -> Program -> Prop :=
  | ev_add :
      forall (H : Heap) (R R' : RegFile) (r31 : AP) 
        (I' : InstrSeq) (rd rs rs' : Reg) (rsval rsval' : nat),
      mlookup _ _ R rs (wi rsval) ->
      mlookup _ _ R rs' (wi rsval') ->
      mupdext _ _ R rd (wi (rsval + rsval')) R' ->
      Eval (H, (R, (r31, iseq (add rd rs rs') I'))) (H, (R', (r31, I')))
  | ev_addi :
      forall (H : Heap) (R R' : RegFile) (r31 : AP) 
        (I' : InstrSeq) (rd rs : Reg) (i rsval : nat),
      mlookup _ _ R rs (wi rsval) ->
      mupdext _ _ R rd (wi (rsval + i)) R' ->
      Eval (H, (R, (r31, iseq (addi rd rs i) I'))) (H, (R', (r31, I')))
  | ev_alloc :
      forall (H H' : Heap) (R R' : RegFile) (r31 : AP) 
        (I' : InstrSeq) (h h' : Map nat HeapVal) (wfh : mWF _ _ h)
        (wfh' : mWF _ _ h') (rd : Reg) (V : list Omega),
      H = wfmap _ _ h wfh ->
      H' = wfmap _ _ h' wfh' ->
      mupdext _ _ R rd (wl r31) R' ->
      mextend nat HeapVal h r31 (tuple (makeUninitTup V)) h' ->
      Eval (H, (R, (r31, iseq (alloc rd V) I'))) (H', (R', (r31, I')))
  | ev_bgt0 :
      forall (H : Heap) (R : RegFile) (r31 : AP) (I' : InstrSeq)
        (rs rt : Reg) (l rsval rtval : nat),
      mlookup _ _ R rs (wi rsval) ->
      mlookup _ _ R rt (wi rtval) ->
      rsval <= rtval ->
      Eval (H, (R, (r31, iseq (bgt rs rt l) I'))) (H, (R, (r31, I')))
  | ev_bgt1 :
      forall (H : Heap) (R : RegFile) (r31 : AP) (I' : InstrSeq)
        (G : Map Reg Omega) (T : APTy) (I'' : InstrSeq) 
        (rs rt : Reg) (l rsval rtval : nat),
      mlookup _ _ R rs (wi rsval) ->
      mlookup _ _ R rt (wi rtval) ->
      rsval > rtval ->
      mlookup nat HeapVal (unwrapMap nat HeapVal H) l (code G T I'') ->
      Eval (H, (R, (r31, iseq (bgt rs rt l) I'))) (H, (R, (r31, I'')))
  | ev_bump :
      forall (H : Heap) (R : RegFile) (r31 : AP) (I' : InstrSeq) (i l : nat),
      msize nat HeapVal (unwrapMap nat HeapVal H) l ->
      Eval (H, (R, (r31, iseq (bump i) I'))) (H, (R, (l, I')))
  | ev_fold :
      forall (H : Heap) (R R' : RegFile) (r31 : AP) 
        (I' : InstrSeq) (rd rs : Reg) (t : Omega) (rsval : WordVal),
      mlookup _ _ R rs rsval ->
      mupdext _ _ R rd (wfold rsval t) R' ->
      Eval (H, (R, (r31, iseq (fold rd t rs) I'))) (H, (R', (r31, I')))
  | ev_jd :
      forall (H : Heap) (R : RegFile) (r31 : AP) (l : nat)
        (G : Map Reg Omega) (T : APTy) (I' : InstrSeq),
      mlookup nat HeapVal (unwrapMap nat HeapVal H) l (code G T I') ->
      Eval (H, (R, (r31, jd l))) (H, (R, (r31, I')))
  | ev_jmp :
      forall (H : Heap) (R : RegFile) (r31 : AP) (r : Reg) 
        (l : nat) (G : Map Reg Omega) (T : APTy) (I' : InstrSeq),
      mlookup _ _ R r (wl l) ->
      mlookup nat HeapVal (unwrapMap nat HeapVal H) l (code G T I') ->
      Eval (H, (R, (r31, jmp r))) (H, (R, (r31, I')))
  | ev_ld :
      forall (H : Heap) (R R' : RegFile) (r31 : AP) 
        (I' : InstrSeq) (rd rs : Reg) (i l : nat) (V : list WordVal)
        (v : WordVal),
      mlookup _ _ R rs (wl l) ->
      mlookup nat HeapVal (unwrapMap nat HeapVal H) l (tuple V) ->
      ListNth _ V i v ->
      mupdext _ _ R rd v R' ->
      Eval (H, (R, (r31, iseq (ld rd rs i) I'))) (H, (R', (r31, I')))
  | ev_mov :
      forall (H : Heap) (R R' : RegFile) (r31 : AP) 
        (I' : InstrSeq) (rd rs : Reg) (v : WordVal),
      mlookup _ _ R rs v ->
      mupdext _ _ R rd v R' ->
      Eval (H, (R, (r31, iseq (mov rd rs) I'))) (H, (R', (r31, I')))
  | ev_movi :
      forall (H : Heap) (R R' : RegFile) (r31 : AP) 
        (I' : InstrSeq) (rd : Reg) (i : nat),
      mupdext _ _ R rd (wi i) R' ->
      Eval (H, (R, (r31, iseq (movi rd i) I'))) (H, (R', (r31, I')))
  | ev_movl :
      forall (H : Heap) (R R' : RegFile) (r31 : AP) 
        (I' : InstrSeq) (rd : Reg) (l : nat),
      mupdext _ _ R rd (wl l) R' ->
      Eval (H, (R, (r31, iseq (movl rd l) I'))) (H, (R', (r31, I')))
  | ev_store :
      forall (H H' : Heap) (R : RegFile) (r31 : AP) 
        (I' : InstrSeq) (h h' : Map nat HeapVal) (wfh : mWF _ _ h)
        (wfh' : mWF _ _ h') (rd rs : Reg) (i l : nat) 
        (V V' : list WordVal) (w : WordVal),
      H = wfmap _ _ h wfh ->
      H' = wfmap _ _ h' wfh' ->
      mlookup _ _ R rd (wl l) ->
      mlookup _ _ R rs w ->
      mlookup nat HeapVal h l (tuple V) ->
      updatetuple V i w V' ->
      mupdate nat HeapVal h l (tuple V') h' ->
      Eval (H, (R, (r31, iseq (st rd i rs) I'))) (H', (R, (r31, I')))
  | ev_unfold :
      forall (H : Heap) (R R' : RegFile) (r31 : AP) 
        (I' : InstrSeq) (rd rs : Reg) (v : WordVal) 
        (t : Omega),
      mlookup _ _ R rs (wfold v t) ->
      mupdext _ _ R rd v R' ->
      Eval (H, (R, (r31, iseq (unfold rd rs) I'))) (H, (R', (r31, I'))).




(***************************************************************************)
(* Static Semantics *)

(* register file subtyping *)                                  (*  G <= G' *)
Inductive RegFileSubtype : Map Reg Omega -> Map Reg Omega -> Prop :=
    weaken :
      forall G G' : Map Reg Omega,
      (forall (r : Reg) (t : Omega), mlookup _ _ G' r t -> mlookup _ _ G r t) ->
      RegFileSubtype G G'. 


(* subtyping *)                                              (* |- t <= t' *)
Inductive SubtypeB : bool -> bool -> Prop :=
  | subb_refl : forall i : bool, SubtypeB i i
  | subb_true : SubtypeB true false.


Inductive SubtypeBList : list bool -> list bool -> Prop :=
  | subb_nil : SubtypeBList nil nil
  | subb_cons :
      forall (L1 L2 : list bool) (i1 i2 : bool),
      SubtypeB i1 i2 ->
      SubtypeBList L1 L2 -> SubtypeBList (i1 :: L1) (i2 :: L2).


Inductive Subtype : Omega -> Omega -> Prop :=
  | reflex : forall t : Omega, Subtype t t
  | tuple_sub :
      forall (ol : list Omega) (il1 il2 : list bool),
      SubtypeBList il1 il2 -> Subtype (tupty ol il1) (tupty ol il2).


(* well-formed word values of type *)                  (* HT |- w : t wval *)
Inductive WFWordVal : HeapTy -> WordVal -> Omega -> Prop :=
  | int_wval : forall (HT : HeapTy) (i : nat), WFWordVal HT (wi i) intty
  | label_wval :
      forall (HT : HeapTy) (l : nat) (t t' : Omega),
      mlookup nat Omega (unwrapMap nat Omega HT) l t' ->
      Subtype t' t -> WFWordVal HT (wl l) t
  | fold_word_wval :
      forall (HT : HeapTy) (w : WordVal) (t : OmegaL 1) (t' : Omega),
      RUnlift (RUnfold t) = t' ->
      WFWordVal HT w t' -> WFWordVal HT (wfold w (recty t)) (recty t).


(* well-formed initialized word values of type *)
Inductive WFWordValinit : HeapTy -> WordVal -> Omega -> bool -> Prop :=
  | init :
      forall (HT : HeapTy) (w : WordVal) (t : Omega) (f : bool),
      WFWordVal HT w t -> WFWordValinit HT w t f
  | uninit :
      forall (HT : HeapTy) (t : Omega), WFWordValinit HT (wuninit t) t false.



(* well-formed instruction sequences *)                     (* HT; G |- I *)
Inductive WFInstrSeq : HeapTy -> Map Reg Omega -> APTy -> InstrSeq -> Prop :=
  | s_add :
      forall (HT : HeapTy) (G G' : Map Reg Omega) (T : APTy) 
        (I : InstrSeq) (rd rs rs' : Reg),
      mlookup _ _ G rs intty ->
      mlookup _ _ G rs' intty ->
      mupdext _ _ G rd intty G' ->
      WFInstrSeq HT G' T I -> WFInstrSeq HT G T (iseq (add rd rs rs') I)
  | s_addi :
      forall (HT : HeapTy) (G G' : Map Reg Omega) (T : APTy) 
        (I : InstrSeq) (rd rs : Reg) (i : nat),
      mlookup _ _ G rs intty ->
      mupdext _ _ G rd intty G' ->
      WFInstrSeq HT G' T I -> WFInstrSeq HT G T (iseq (addi rd rs i) I)
  | s_alloc :
      forall (HT : HeapTy) (G G' : Map Reg Omega) (I : InstrSeq) 
        (rd : Reg) (n : nat) (V : list Omega),
      n = length V ->
      mupdext _ _ G rd (tupty V (makeUninitTupty V)) G' ->
      WFInstrSeq HT G' (used n) I ->
      WFInstrSeq HT G fresh (iseq (alloc rd V) I)
  | s_bgt :
      forall (HT : HeapTy) (G G' : Map Reg Omega) (T : APTy) 
        (I : InstrSeq) (rs rt : Reg) (l : nat),
      mlookup _ _ G rs intty ->
      mlookup _ _ G rt intty ->
      mlookup nat Omega (unwrapMap nat Omega HT) l (codety G' T) ->
      RegFileSubtype G G' ->
      WFInstrSeq HT G T I -> WFInstrSeq HT G T (iseq (bgt rs rt l) I)
  | s_bump :
      forall (HT : HeapTy) (G : Map Reg Omega) (I : InstrSeq) (n : nat),
      WFInstrSeq HT G fresh I -> WFInstrSeq HT G (used n) (iseq (bump n) I)
  | s_fold :
      forall (HT : HeapTy) (G G' : Map Reg Omega) (T : APTy) 
        (I : InstrSeq) (rd rs : Reg) (f : OmegaL 1) 
        (rst : Omega),
      mlookup _ _ G rs rst ->
      RUnlift (RUnfold f) = rst ->
      mupdext _ _ G rd (recty f) G' ->
      WFInstrSeq HT G' T I ->
      WFInstrSeq HT G T (iseq (fold rd (recty f) rs) I)
  | s_jd :
      forall (HT : HeapTy) (G G' : Map Reg Omega) (T : APTy) (l : nat),
      mlookup nat Omega (unwrapMap nat Omega HT) l (codety G' T) ->
      RegFileSubtype G G' -> WFInstrSeq HT G T (jd l)
  | s_jmp :
      forall (HT : HeapTy) (G G' : Map Reg Omega) (T : APTy) (r : Reg),
      mlookup _ _ G r (codety G' T) ->
      RegFileSubtype G G' -> WFInstrSeq HT G T (jmp r)
  | s_ld :
      forall (HT : HeapTy) (G G' : Map Reg Omega) (T : APTy) 
        (I : InstrSeq) (rd rs : Reg) (i : nat) (ol : list Omega)
        (il : list bool) (t : Omega),
      mlookup _ _ G rs (tupty ol il) ->
      ListNth _ ol i t ->
      ListNth _ il i true ->
      mupdext _ _ G rd t G' ->
      WFInstrSeq HT G' T I -> WFInstrSeq HT G T (iseq (ld rd rs i) I)
  | s_mov :
      forall (HT : HeapTy) (G G' : Map Reg Omega) (T : APTy) 
        (I : InstrSeq) (rd rs : Reg) (t : Omega),
      mlookup _ _ G rs t ->
      mupdext _ _ G rd t G' ->
      WFInstrSeq HT G' T I -> WFInstrSeq HT G T (iseq (mov rd rs) I)
  | s_movi :
      forall (HT : HeapTy) (G G' : Map Reg Omega) (T : APTy) 
        (I : InstrSeq) (rd : Reg) (i : nat),
      mupdext _ _ G rd intty G' ->
      WFInstrSeq HT G' T I -> WFInstrSeq HT G T (iseq (movi rd i) I)
  | s_movl :
      forall (HT : HeapTy) (G G' : Map Reg Omega) (T : APTy) 
        (I : InstrSeq) (rd : Reg) (l : nat) (t t' : Omega),
      mlookup nat Omega (unwrapMap nat Omega HT) l t ->
      Subtype t t' ->
      mupdext _ _ G rd t' G' ->
      WFInstrSeq HT G' T I -> WFInstrSeq HT G T (iseq (movl rd l) I)
  | s_st :
      forall (HT : HeapTy) (G G' : Map Reg Omega) (T : APTy) 
        (I : InstrSeq) (rd rs : Reg) (i : nat) (V V' : list bool)
        (Ts : list Omega) (t : Omega),
      mlookup _ _ G rd (tupty Ts V) ->
      mlookup _ _ G rs t ->
      ListNth _ Ts i t ->
      updatetupty V i V' ->
      mupdate _ _ G rd (tupty Ts V') G' ->
      WFInstrSeq HT G' T I -> WFInstrSeq HT G T (iseq (st rd i rs) I)
  | s_unfold :
      forall (HT : HeapTy) (G G' : Map Reg Omega) (T : APTy) 
        (I : InstrSeq) (rd rs : Reg) (f : OmegaL 1) 
        (unf rst : Omega),
      mlookup _ _ G rs rst ->
      rst = recty f ->
      RUnlift (RUnfold f) = unf ->
      mupdext _ _ G rd unf G' ->
      WFInstrSeq HT G' T I -> WFInstrSeq HT G T (iseq (unfold rd rs) I).


Inductive WFWordValinitList :
HeapTy -> list WordVal -> list Omega -> list bool -> Prop :=
  | wfwvilist0 : forall HT : HeapTy, WFWordValinitList HT nil nil nil
  | wfwvilist1 :
      forall (HT : HeapTy) (wl : list WordVal) (tl : list Omega)
        (il : list bool) (w : WordVal) (t : Omega) 
        (i : bool),
      WFWordValinit HT w t i ->
      WFWordValinitList HT wl tl il ->
      WFWordValinitList HT (w :: wl) (t :: tl) (i :: il).


(* well-formed heap values of type *)
Inductive WFHeapVal : HeapTy -> HeapVal -> Omega -> Prop :=
  | tuple_wf :
      forall (HT : HeapTy) (wl : list WordVal) (tl : list Omega)
        (il : list bool),
      WFWordValinitList HT wl tl il -> WFHeapVal HT (tuple wl) (tupty tl il)
  | code_wf :
      forall (HT : HeapTy) (G : Map Reg Omega) (I : InstrSeq) (T : APTy),
      WFInstrSeq HT G T I -> WFHeapVal HT (code G T I) (codety G T).


(* well-formed heap type *)
Inductive WFHeap : Heap -> HeapTy -> Prop :=
    heap_wf :
      forall (H : Heap) (HT : HeapTy),
      (exists s : _,
         msize nat HeapVal (unwrapMap nat HeapVal H) s /\
         msize nat Omega (unwrapMap nat Omega HT) s /\
         (forall (n : nat) (h : HeapVal),
          mlookup nat HeapVal (unwrapMap nat HeapVal H) n h -> n < s) /\
         (forall (n : nat) (t : Omega),
          mlookup nat Omega (unwrapMap nat Omega HT) n t -> n < s) /\
         (forall n : nat,
          n < s ->
          exists h : _, mlookup nat HeapVal (unwrapMap nat HeapVal H) n h) /\
         (forall n : nat,
          n < s ->
          exists t : _, mlookup nat Omega (unwrapMap nat Omega HT) n t) /\
         (forall (n : nat) (h : HeapVal) (t : Omega),
          mlookup nat HeapVal (unwrapMap nat HeapVal H) n h ->
          mlookup nat Omega (unwrapMap nat Omega HT) n t -> WFHeapVal HT h t) /\
         OrdHeap (unwrapMap nat HeapVal H)) -> WFHeap H HT.
		   


(* well-formed allocation pointer *)
Inductive WFap : HeapTy -> AP -> APTy -> Prop :=
  | fresh_wf :
      forall (HT : HeapTy) (l : AP),
      msize nat Omega (unwrapMap nat Omega HT) l -> WFap HT l fresh
  | used_wf :
      forall (HT : HeapTy) (l : AP) (n : nat) (tl : list Omega)
        (ol : list bool),
      n = length tl ->
      msize nat Omega (unwrapMap nat Omega HT) (S l) ->
      WFWordVal HT (wl l) (tupty tl ol) -> WFap HT l (used n).


(* strip folds from word value *)
Fixpoint stripWV (w : WordVal) : WordVal :=
  match w with
  | wfold w' t => stripWV w'
  | _ => w
  end.

(* well-formed register file of type *)
Inductive WFRegFile : HeapTy -> RegFile -> Map Reg Omega -> Prop :=
    regfile_wf :
      forall (HT : HeapTy) (R : RegFile) (G : Map Reg Omega),
      (forall (r : Reg) (t : Omega),
       mlookup _ _ G r t ->
       exists w : _, mlookup _ _ R r w /\ WFWordVal HT w t) ->
      (forall (r : Reg) (v : WordVal) (l n : nat),
       mlookup _ _ R r v ->
       stripWV v = wl l ->
       msize nat Omega (unwrapMap nat Omega HT) n -> l < n) ->
      WFRegFile HT R G.

(* well-formed program *)
Inductive WFProgram : Program -> Prop :=
    program_wf :
      forall (H : Heap) (HT : HeapTy) (R : RegFile) 
        (G : Map Reg Omega) (l : AP) (t : APTy) (I : InstrSeq),
      WFHeap H HT ->
      WFRegFile HT R G ->
      WFap HT l t ->
      WFInstrSeq HT G t I ->
      (exists l : _,
         (exists G' : _,
            (exists T' : _,
               (exists I' : _,
                  (exists n : _,
                     mlookup nat HeapVal (unwrapMap nat HeapVal H) l
                       (code G' T' I') /\ ISubDepth I I' n))))) ->
      WFProgram (H, (R, (l, I))).


(***************************************************************************)
(* Proofs *)

Load "lemmas_ftal".


Theorem preserv :
 forall P P' : Program, WFProgram P -> Eval P P' -> WFProgram P'.
Proof.
intros; inversion H; inversion H0.
(* add case: *)
apply (preserv_add H8 R0 I' r31 rd rs rs' (H8, (R', (r31, I'))));
 replace (H8, (R0, (r31, iseq (add rd rs rs') I'))) with P; 
 auto; replace (H8, (R0, (r31, iseq (add rd rs rs') I'))) with P;
 replace (H8, (R', (r31, I'))) with P'; auto.
(* addi case: *)
apply (preserv_addi H8 R0 I' r31 rd rs i (H8, (R', (r31, I'))));
 replace (H8, (R0, (r31, iseq (addi rd rs i) I'))) with P; 
 auto; replace (H8, (R0, (r31, iseq (addi rd rs i) I'))) with P;
 replace (H8, (R', (r31, I'))) with P'; auto.
(* alloc case: *)
apply (preserv_alloc H8 R0 I' r31 rd V (H', (R', (r31, I')))).
replace (H8, (R0, (r31, iseq (alloc rd V) I'))) with P.
auto.
replace (H8, (R0, (r31, iseq (alloc rd V) I'))) with P.
replace (H', (R', (r31, I'))) with P'.
auto.
(* 2 bgt cases: *)
apply (preserv_bgt H8 R0 I' r31 (H8, (R0, (r31, I'))) rs rt l0).
replace (H8, (R0, (r31, iseq (bgt rs rt l0) I'))) with P.
auto.
replace (H8, (R0, (r31, iseq (bgt rs rt l0) I'))) with P.
replace (H8, (R0, (r31, I'))) with P'.
auto.
apply (preserv_bgt H8 R0 I' r31 (H8, (R0, (r31, I''))) rs rt l0).
replace (H8, (R0, (r31, iseq (bgt rs rt l0) I'))) with P.
auto.
replace (H8, (R0, (r31, iseq (bgt rs rt l0) I'))) with P.
replace (H8, (R0, (r31, I''))) with P'.
auto.
(* bump case; *)
apply (preserv_bump H8 R0 I' r31 i (H8, (R0, (l0, I')))).
replace (H8, (R0, (r31, iseq (bump i) I'))) with P.
auto.
replace (H8, (R0, (r31, iseq (bump i) I'))) with P.
replace (H8, (R0, (l0, I'))) with P'.
auto.
(* fold case: *)
apply (preserv_fold H8 R0 I' r31 (H8, (R', (r31, I'))) rd rs t0).
replace (H8, (R0, (r31, iseq (fold rd t0 rs) I'))) with P.
auto.
replace (H8, (R0, (r31, iseq (fold rd t0 rs) I'))) with P.
replace (H8, (R', (r31, I'))) with P'.
auto.
(* jd case: *)
apply (preserv_jd H8 R0 r31 l0 (H8, (R0, (r31, I'))));
 replace (H8, (R0, (r31, jd l0))) with P; auto;
 replace (H8, (R0, (r31, jd l0))) with P;
 replace (H8, (R0, (r31, I'))) with P'; auto.
(* jmp case: *)
apply (preserv_jmp H8 R0 r31 r (H8, (R0, (r31, I'))));
 replace (H8, (R0, (r31, jmp r))) with P; auto;
 replace (H8, (R0, (r31, jmp r))) with P;
 replace (H8, (R0, (r31, I'))) with P'; auto.
(* ld case: *)
apply (preserv_ld H8 R0 I' r31 (H8, (R', (r31, I'))) rd rs i).
rewrite H13.
auto.
rewrite H13.
rewrite H14.
auto.
(* mov, movi cases: *)
apply (preserv_mov H8 R0 r31 I' (H8, (R', (r31, I'))) rd rs).
rewrite H11.
auto.
rewrite H11.
rewrite H12.
auto.
apply (preserv_movi H8 R0 r31 I' (H8, (R', (r31, I'))) rd i).
rewrite H10.
auto.
rewrite H10.
rewrite H11.
auto.
(* movl case: *)
apply (preserv_movl H8 R0 r31 I' (H8, (R', (r31, I'))) rd l0);
 replace (H8, (R0, (r31, iseq (movl rd l0) I'))) with P; 
 auto; replace (H8, (R0, (r31, iseq (movl rd l0) I'))) with P;
 replace (H8, (R', (r31, I'))) with P'; auto.
(* st case: *)
apply (preserv_st H8 R0 I' r31 rd rs i (H', (R0, (r31, I'))));
 replace (H8, (R0, (r31, iseq (st rd i rs) I'))) with P; 
 auto; replace (H8, (R0, (r31, iseq (st rd i rs) I'))) with P;
 replace (H', (R0, (r31, I'))) with P'; auto.
(* unfold case: *)
apply (preserv_unfold H8 R0 I' r31 (H8, (R', (r31, I'))) rd rs).
replace (H8, (R0, (r31, iseq (unfold rd rs) I'))) with P.
auto.
replace (H8, (R0, (r31, iseq (unfold rd rs) I'))) with P.
replace (H8, (R', (r31, I'))) with P'.
auto.
Qed.




Theorem progress :
 forall P : Program, WFProgram P -> exists P' : _, Eval P P'.
Proof.
intros.
inversion H.
clear H5.
inversion H4.
apply progress_add.
rewrite H13.
rewrite H6.
auto.
apply progress_addi.
rewrite H12.
rewrite H6.
auto.
apply progress_alloc.
rewrite H12.
rewrite H6.
auto.
apply progress_bgt.
rewrite H14.
rewrite H6.
auto.
apply progress_bump.
rewrite H10.
rewrite H6.
auto.
apply progress_fold.
rewrite H13.
rewrite H6.
auto.
apply progress_jd.
rewrite H11.
rewrite H6.
auto.
apply progress_jmp.
rewrite H11.
rewrite H6.
auto.
apply progress_ld.
rewrite H14.
rewrite H6.
auto.
apply progress_mov.
rewrite H12. rewrite H6. auto.
apply progress_movi.
rewrite H11. rewrite H6. auto.
apply progress_movl.
rewrite H13.
rewrite H6.
auto.
apply progress_st.
rewrite H15.
rewrite H6.
auto.
apply progress_unfold.
rewrite H14.
rewrite H6.
auto.
Qed.




(***************************************************************************)

