(*
 * Coq code for the paper
 *
 *       A Syntactic Approach to Foundational Proof-Carrying Code
 *                            
 *
 *   by Zhong Shao, Valery Trifonov, Stefan Monnier, Nadeem A. Hamid, 
 *                          and Zhaozhong Ni
 *
 * (for Coq version 7.3)
 *)
(* Thanks to Yves Bertot, Clement Renard at Inria for help with Coq *)

(***************************************************************************)
(* Definition of mapping structure and related utility functions *)
(***************************************************************************)

Require Import natutil.

(***************************************************************************)

(* Map structure ------------------------ *)
Inductive Map (A B : Set) : Set :=
  | mempty : Map A B
  | melem : A -> B -> Map A B -> Map A B.

Definition is_mempty (A B : Set) (m : Map A B) :=
  match m with
  | mempty => True
  | _ => False
  end.


(* size of mapping - probably only makes sense if mapping is well-formed *)
Inductive msize (A B : Set) : Map A B -> nat -> Prop :=
  | msize0 : msize A B (mempty A B) 0
  | msize1 :
      forall (a : A) (b : B) (m : Map A B) (n : nat),
      msize A B m n -> msize A B (melem A B a b m) (S n).

Fixpoint msizef (A B : Set) (m : Map A B) {struct m} : nat :=
  match m with
  | mempty => 0
  | melem _ _ m' => S (msizef _ _ m')
  end.


(* element is in the domain of the mapping *)
Inductive mindom (A B : Set) : Map A B -> A -> Prop :=
  | minfound :
      forall (a : A) (b : B) (m : Map A B), mindom A B (melem A B a b m) a
  | minnext :
      forall (a a' : A) (b : B) (m : Map A B),
      a <> a' -> mindom A B m a' -> mindom A B (melem A B a b m) a'.


(* an element is not in the domain of the map *)
Inductive mnotindom (A B : Set) : Map A B -> A -> Prop :=
  | mnotindom0 : forall a' : A, mnotindom A B (mempty A B) a'
  | mnotindom1 :
      forall (a a' : A) (b : B) (m : Map A B),
      a <> a' -> mnotindom A B m a' -> mnotindom A B (melem A B a b m) a'.


(* well-formedness means all distinct elements unequal *)
Inductive mWF (A B : Set) : Map A B -> Prop :=
  | mwf0 : mWF A B (mempty A B)
  | mwf1 :
      forall (a : A) (b : B) (m : Map A B),
      mnotindom A B m a -> mWF A B m -> mWF A B (melem A B a b m).


(* a well-formed map wrapper *)
Inductive WFMap (A B : Set) : Set :=
    wfmap : forall m : Map A B, mWF A B m -> WFMap A B.

(* utilities for working with well-formed maps *)
Definition unwrapMap (A B : Set) (m : WFMap A B) : 
  Map A B := match m with
             | wfmap m' mwf => m'
             end.
Definition emptyWFMap (A B : Set) : WFMap A B :=
  wfmap _ _ (mempty _ _) (mwf0 _ _).


(* Used to prove that a pair is in this mapping *)
Inductive mlookup (A B : Set) : Map A B -> A -> B -> Prop :=
  | mlookfound :
      forall (a : A) (b : B) (m : Map A B), mlookup A B (melem A B a b m) a b
  | mlooknext :
      forall (a a' : A) (b b' : B) (m : Map A B),
      mlookup A B m a' b' -> a <> a' -> mlookup A B (melem A B a b m) a' b'.


(* updates the mapping of an element that is already in the domain *)
Inductive mupdate (A B : Set) : Map A B -> A -> B -> Map A B -> Prop :=
  | mupdfound :
      forall (a : A) (b b' : B) (m : Map A B),
      mupdate A B (melem A B a b' m) a b (melem A B a b m)
  | mupdnext :
      forall (a a' : A) (b b' : B) (m m' : Map A B),
      mupdate A B m a b m' ->
      a <> a' -> mupdate A B (melem A B a' b' m) a b (melem A B a' b' m').


(* extends the mapping, adding a new element to the domain *)
Inductive mextend (A B : Set) : Map A B -> A -> B -> Map A B -> Prop :=
    mext :
      forall (a : A) (b : B) (m : Map A B),
      mnotindom A B m a -> mextend A B m a b (melem A B a b m).


(* sometimes we don't care exactly what happens to the mapping *)
Inductive mupdext (A B : Set) : Map A B -> A -> B -> Map A B -> Prop :=
  | mupdext0 :
      forall (a : A) (b : B),
      mupdext A B (mempty A B) a b (melem A B a b (mempty A B))
  | mupdext1 :
      forall (a a' : A) (b b' : B) (m m' : Map A B),
      mupdext A B m a b m' ->
      a <> a' -> mupdext A B (melem A B a' b' m) a b (melem A B a' b' m')
  | mupdext2 :
      forall (a : A) (b b' : B) (m : Map A B),
      mupdext A B (melem A B a b' m) a b (melem A B a b m).



(* -------------------------------------------------------------------- *)
(* THEOREMS ABOUT MAPS *)
(*
msize_always
msize_eq
eq_map_wf
mwf_next
mwf_elem_neq
mwf_anyelem_neq
mindom_next
mindom_imp_mlookup
mindom_imp_not_mnotindom
min_or_not_nat_wf
min_or_not
not_min_and_not_wf
mnotindom_imp_not_mindom 
mnotindom_imp_mextend 
mnotindom_mextend_imp_mnotindom
mindom_mnotindom_neq
mlookup_next 
mlookup_imp_mindom 
mlookup_eq
mindom_imp_mupdate
mlookup_imp_mupdate
mupdate_refl
mupdate_imp_mindom_bef
mupdate_imp_mindom_aft
mupdate_imp_mlookup_ex
mupdate_imp_mlookup
mupdate_imp_mupdext
mupdate_preserv
mupdate_preserv_rev
mupdate_preserv_rev_ex
mupdate_size_same
mupdate_mnotindom_preserv
mupdate_always_wf
mupdate_same_eq
mupdate_preserv_mindom_rev
mupdate_preserv_mindom
mextend_imp_mlookup
mextend_wf 
mextend_preserv
mextend_preserv_rev
mextend_size_Sn
mupdext_always
mupdext_preserv
mupdext_preserv_rev
mupdext_preserv_rev_mindom
mupdext_imp_mlookup
mupdext_same_eq

ext_mupdext_two_preserv
ext_map_oneoff_updext_same_oneoff
ext_map_eq_updext_eq
ext_map_lookup_eq



*)




(* a map always has a size *)
Lemma msize_always :
 forall (A B : Set) (m : Map A B), exists l : _, msize A B m l.
Proof.
intros.
induction m as [| a b m Hrecm].
split with 0.
constructor.

inversion_clear Hrecm.
split with (S x).
constructor.
auto.
Qed.


(* a map has a unique size *)
Lemma msize_eq :
 forall (A B : Set) (m : Map A B) (a a' : nat),
 msize A B m a -> msize A B m a' -> a = a'.
Proof.
simple induction m.
intros.
inversion H.
inversion H0.
auto.
intros.
inversion H0.
inversion H1.
cut (n = n0).
intros.
auto.
apply (H n n0 H6 H11).
Qed.


(* well-formedness of equal maps *)
Lemma eq_map_wf :
 forall (A B : Set) (m m' : Map A B) (a : A) (b : B),
 mWF A B m -> m = melem A B a b m' -> mWF A B (melem A B a b m').
Proof.
simple induction m.
intros.
absurd (mempty A B = melem A B a b m').
discriminate; auto.
auto.
intros.
constructor.
inversion H0.
replace m' with m0.
replace a0 with a.
auto.
inversion H1.
auto.
inversion H1.
auto.
inversion H0.
replace m' with m0.
auto.
injection H1; auto.
Qed.



(* subsets of a well-formed map are well-formed *)
Lemma mwf_next :
 forall (A B : Set) (a : A) (b : B) (m : Map A B),
 mWF _ _ (melem _ _ a b m) -> mWF _ _ m.
intros.
inversion H.
auto.
Defined.


(* subsequent elements in a well-formed map are not equal *)
Lemma mwf_elem_neq :
 forall (A B : Set) (a a' : A) (b b' : B) (m : Map A B),
 mWF A B (melem A B a b (melem A B a' b' m)) -> a <> a'.
Proof.
intros.
inversion H.
inversion H2.
auto.
Qed.


(* any two distinct elements in a well-formed map are not equal *)
Lemma mwf_anyelem_neq :
 forall (A B : Set) (a a' : A) (b : B) (m : Map A B),
 mWF A B (melem A B a b m) -> mindom A B m a' -> a <> a'.
Proof.
intros.
induction m as [| a0 b0 m Hrecm].
inversion H0.
inversion H0.
cut (a <> a0).
intros.
rewrite <- H1.
auto.
apply (mwf_elem_neq A B a a0 b b0 m H).
cut (mWF A B (melem A B a b m)).
intros.
apply (Hrecm H7 H6).
clear H1.
clear a1.
clear H3.
clear a'0.
clear H4.
clear m0.
clear H2.
clear b1.
apply mwf1.
inversion H.
inversion H3.
auto.
cut (mWF A B (melem A B a0 b0 m)).
intros.
apply (mwf_next A B a0 b0 m H1).
apply (mwf_next A B a b (melem A B a0 b0 m) H).
Qed.


(* if a' is in the domain of the map and a not-equal a' then a' must be
   in the domain of the remainder *)
Lemma mindom_next :
 forall (A B : Set) (m : Map A B) (a a' : A) (b : B),
 mindom A B (melem A B a b m) a' -> a <> a' -> mindom A B m a'.
Proof.
intros.
inversion H.
tauto.
auto.
Qed.


(* if a is in the domain of the map, it must be mapped to something *)
Lemma mindom_imp_mlookup :
 forall (A B : Set) (m : Map A B) (a : A),
 mindom A B m a -> exists b : _, mlookup A B m a b.
Proof.
intros.
induction m as [| a0 b m Hrecm].
inversion H.
inversion H.
split with b.
constructor.
cut (exists b : B, mlookup A B m a b).
intros.
inversion_clear H6.
split with x.
constructor.
auto.
auto.
apply (Hrecm H5).
Qed.


(* if l is in the domain, it can't not be in the domain *)
Lemma mindom_imp_not_mnotindom :
 forall (A B : Set) (m : Map A B) (l : A),
 (forall a a' : A, a = a' \/ a <> a') ->
 mindom A B m l -> ~ mnotindom A B m l.
Proof.
intros.
induction m as [| a b m Hrecm].
inversion H0.
cut (l = a \/ l <> a); intros; auto.
inversion_clear H1.
inversion_clear H0.
intro.
inversion H0; absurd (l <> l); auto.
rewrite H2 in H1; absurd (a <> a); auto.
intro.
inversion H1.
apply Hrecm; auto.
inversion H0; [ rewrite H9 in H2; absurd (l <> l); auto | auto ].
Qed.


(* an element is either in or not in a (well-formed) map *)
Lemma min_or_not_nat_wf :
 forall (B : Set) (m : Map nat B) (l : nat),
 mWF _ _ m -> mindom nat B m l \/ mnotindom nat B m l.
Proof.
intros.
induction m as [| a b m Hrecm].
right.
constructor.
inversion H.
cut (mindom nat B m l \/ mnotindom nat B m l).
intros.
inversion H5.
left.
apply minnext.
apply (mwf_anyelem_neq nat B a l b m H H6).
auto.
cut (a = l \/ a <> l).
intros.
inversion_clear H7.
left.
rewrite H8.
constructor.
right.
apply mnotindom1.
auto.
auto.
apply eqorneq_nat.
apply (Hrecm H4).
Qed.


(* an element is either in or not in a map if decidable equality *)
Lemma min_or_not :
 forall (A B : Set) (m : Map A B) (l : A)
   (P : forall x x' : A, {x = x'} + {x <> x'}),
 {mindom A B m l} + {mnotindom A B m l}.
simple induction m; intros.
right; constructor; auto.
elim (P l a); intros.
rewrite a0; left; constructor; auto.
elim (H l P); intros.
left; constructor; auto.
right; constructor; auto.
Qed.


(* an element is not both in and not in a (well-formed) map *)
Lemma not_min_and_not_wf :
 forall (B : Set) (m : Map nat B) (l : nat),
 mWF _ _ m -> ~ (mindom nat B m l /\ mnotindom nat B m l).
intros.
induction m as [| a b m Hrecm].
unfold not in |- *.
intros.
inversion H0.
inversion H1.
unfold not in |- *.
intros.
inversion H0.
cut (~ mnotindom nat B (melem nat B a b m) l).
intros.
apply H3.
auto.
apply mindom_imp_not_mnotindom.
apply eqorneq_nat; auto.
auto.
Qed.


(* an element not in the map implies it is not in the map (duh...) *)
Lemma mnotindom_imp_not_mindom :
 forall (B : Set) (m : Map nat B) (l : nat),
 mnotindom nat B m l -> ~ mindom nat B m l.
Proof.
simple induction m; intros; unfold not in |- *; intros.
inversion H0.
compare l a; intros.
rewrite e in H0; inversion H0.
absurd (a <> a); auto.
inversion H0.
generalize (H _ H7); intros.
inversion H1.
rewrite H9 in H6; absurd (l <> l); auto.
apply H8; auto.
Qed.


(* an element not in the map means the map can be extended with it *)
Lemma mnotindom_imp_mextend :
 forall (A B : Set) (m : Map A B) (a : A) (b : B),
 mnotindom _ _ m a -> exists m' : _, mextend _ _ m a b m'.
Proof.
intros.
split with (melem _ _ a b m).
constructor.
auto.
Qed.


(* extending the map with a different element means the element is still
   not in the map *)
Lemma mnotindom_mextend_imp_mnotindom :
 forall (A B : Set) (m m' : Map A B) (a a' : A) (b : B),
 mnotindom A B m a' -> mextend A B m a b m' -> a <> a' -> mnotindom A B m' a'.
Proof.
intros.
inversion H0.
constructor.
auto.
auto.
Qed.



Lemma mindom_mnotindom_neq :
 forall (A B : Set) (m : Map A B) (wfm : mWF _ _ m) (a b : A),
 (forall x x' : A, x = x' \/ x <> x') ->
 mindom _ _ m a -> mnotindom _ _ m b -> a <> b.
Proof.
simple induction m.
intros.
inversion H0.
intros.
cut (a = a0 \/ a <> a0).
intros.
inversion_clear H3.
rewrite H4 in H2.
inversion H2.
auto.
apply H.
inversion wfm.
auto.
auto.
inversion H1.
rewrite H3 in H4.
absurd (a0 <> a0); auto.
auto.
inversion H2.
auto.
auto.
Qed.



(* if the current element is not equal, then must be in the remaining *)
Lemma mlookup_next :
 forall (A B : Set) (m : Map A B) (a a' : A) (b b' : B),
 mlookup A B (melem A B a b m) a' b' -> a <> a' -> mlookup A B m a' b'.
intros.
inversion H.
tauto.
auto.
Qed.


(* if you can look up an element, it must be in the domain *)
Lemma mlookup_imp_mindom :
 forall (A B : Set) (m : Map A B) (a : A) (b : B),
 mlookup A B m a b -> mindom A B m a.
Proof.
intros.
elim H.
intros.
constructor.
intros.
constructor.
auto.
auto.
Defined.


(* lookup is a function (dependent on the charctrstcs of the domain type) *)
Lemma mlookup_eq :
 forall (A B : Set) (a : A) (b b' : B) (m : Map A B),
 (forall x x' : A, x = x' \/ x <> x') ->
 mlookup A B m a b -> mlookup A B m a b' -> b = b'.
Proof.
intros A B a b b' m P H H0.
induction m as [| a0 b0 m Hrecm].
inversion H.
cut (a = a0 \/ a <> a0).
intros.
inversion_clear H1.
rewrite <- H2 in H.
rewrite <- H2 in H0.
cut (b = b0 /\ b' = b0).
intros.
inversion H1.
rewrite H3.
rewrite H4.
auto.
inversion H.
inversion H0.
split.
auto.
rewrite <- H11.
auto.
split.
auto.
absurd (a <> a).
auto.
auto.
absurd (a <> a).
auto.
auto.
cut (mlookup A B m a b').
cut (mlookup A B m a b).
auto.
apply mlookup_next with (a := a0) (b := b0).
auto.
auto.
apply mlookup_next with (a := a0) (b := b0).
auto.
auto.
apply P.
Qed.


(* mlookup is injective *)
Lemma mlookup_neq_dom_neq :
 forall (A B : Set) (m : Map A B) (a a' : A) (b b' : B),
 (forall x x' : A, x = x' \/ x <> x') ->
 mlookup _ _ m a b -> mlookup _ _ m a' b' -> b <> b' -> a <> a'.
Proof.
simple induction m.
intros.
inversion H0.
intros.
cut (a0 = a' \/ a0 <> a').
intros.
inversion_clear H4.
rewrite H5 in H1.
move H2 after H1.
inversion H1.
inversion H2.
rewrite H14 in H9.
rewrite H9 in H3.
absurd (b0 <> b0).
auto.
auto.
rewrite H4 in H16.
absurd (a' <> a').
auto.
auto.
inversion H2.
rewrite H12 in H11.
absurd (a' <> a').
auto.
auto.
rewrite H5.
apply H with b0 b'.
auto.
auto.
auto.
auto.
auto.
auto.
Qed.



(* if an element is previously looked up in the map, it can be updated *)
Lemma mlookup_imp_mupdate :
 forall (A B : Set) (m : Map A B) (a : A) (b b' : B),
 mlookup A B m a b -> exists m' : _, mupdate A B m a b' m'.
Proof.
intros.
induction m as [| a0 b0 m Hrecm].
inversion H.
inversion H.
split with (melem A B a b' m).
apply mupdfound.
cut (mlookup A B m a b).
intro.
cut (exists m' : Map A B, mupdate A B m a b' m').
intros.
elim H8.
intros.
split with (melem A B a0 b0 x).
apply mupdnext.
auto.
auto.
apply (Hrecm H7).
auto.
Qed.


(* if an element is in the map, it can be updated *)
Lemma mindom_imp_mupdate :
 forall (A B : Set) (m : Map A B) (a : A) (b : B),
 mindom A B m a -> exists m' : _, mupdate A B m a b m'.
intros.
cut (exists b' : _, mlookup _ _ m a b');
 [ intros | apply mindom_imp_mlookup; auto ].
inversion_clear H0; apply mlookup_imp_mupdate with x; auto.
Qed.



(* updating with the same thing maintains the same map *)
Lemma mupdate_refl :
 forall (A B : Set) (m : Map A B) (a : A) (b : B),
 (forall x x' : A, x = x' \/ x <> x') ->
 mlookup A B m a b -> mupdate A B m a b m.
Proof.
simple induction m.
intros.
inversion H0.
intros.
cut (a = a0 \/ a <> a0).
intros.
inversion_clear H2.
rewrite H3.
rewrite H3 in H1.
inversion H1.
constructor.
absurd (a0 <> a0).
auto.
auto.
inversion H1.
rewrite H2 in H3.
absurd (a0 <> a0).
auto.
auto.
apply mupdnext.
apply H.
auto.
auto.
auto.
auto.
Qed.



(* if an element has been updated, it must have had a previous value *)
Lemma mupdate_imp_mlookup_ex :
 forall (A B : Set) (m m' : Map A B) (a : A) (b : B)
   (P : forall x x' : A, x = x' \/ x <> x'),
 mupdate A B m a b m' -> exists b' : _, mlookup A B m a b'.
Proof.
simple induction m.
simple induction m'.
intros.
inversion H.
intros.
inversion H0.
intros.
cut (a = a0 \/ a <> a0).
intros.
inversion_clear H1.
split with b.
rewrite H2.
constructor.
cut (exists b' : B, mlookup A B m0 a0 b').
intros.
inversion_clear H1.
split with x.
apply mlooknext.
auto.
auto.
inversion H0.
rewrite H1 in H2.
absurd (a0 <> a0).
auto.
auto.
apply (H m'0 a0 b0).
auto.
auto.
auto.
Qed.


(* if an element is updated, it must be lookup-able exactly the same way
   in the new map *)
Lemma mupdate_imp_mlookup :
 forall (A B : Set) (m m' : Map A B) (a : A) (b : B)
   (P : forall x x' : A, x = x' \/ x <> x'),
 mupdate A B m a b m' -> mlookup A B m' a b.
Proof.
simple induction m.
simple induction m'.
intros.
inversion H.
intros.
inversion H0.
intros.
cut (a = a0 \/ a <> a0).
intros.
inversion_clear H1.
rewrite H2 in H0.
inversion H0.
apply mlookfound.
absurd (a0 <> a0).
auto.
auto.
inversion H0.
rewrite H1 in H2.
absurd (a0 <> a0).
auto.
auto.
apply mlooknext.
apply (H m'0 a0 b0).
auto.
auto.
auto.
auto.
Qed.




Lemma mupdate_imp_mindom_bef :
 forall (A B : Set) (m m' : Map A B) (a : A) (b : B)
   (P : forall x x' : A, x = x' \/ x <> x'),
 mupdate A B m a b m' -> mindom A B m a.
intros.
elim (mupdate_imp_mlookup_ex _ _ m m' a b P H); intros.
apply mlookup_imp_mindom with x; auto.
Qed.


Lemma mupdate_imp_mindom_aft :
 forall (A B : Set) (m m' : Map A B) (a : A) (b : B)
   (P : forall x x' : A, x = x' \/ x <> x'),
 mupdate A B m a b m' -> mindom A B m' a.
intros.
apply mlookup_imp_mindom with b; apply mupdate_imp_mlookup with m; auto.
Qed.






(* mupdate implies mupdext *)
Lemma mupdate_imp_mupdext :
 forall (A B : Set) (m m' : Map A B) (a : A) (b : B)
   (P : forall x x' : A, x = x' \/ x <> x'),
 mupdate A B m a b m' -> mupdext A B m a b m'.
Proof.
simple induction m.
simple induction m'.
intros.
inversion H.
intros.
inversion H0.
intros.
cut (a = a0 \/ a <> a0).
intros.
inversion_clear H1.
rewrite H2 in H0.
inversion H0.
rewrite H2.
constructor.
absurd (a0 <> a0).
auto.
auto.
inversion H0.
rewrite H1 in H2.
absurd (a0 <> a0).
auto.
auto.
apply mupdext1.
auto.
auto.
auto.
Qed.



Lemma mupdate_preserv :
 forall (A B : Set) (m m' : Map A B) (a a' : A) (b b' : B),
 mupdate A B m a' b' m' -> mlookup A B m a b -> a <> a' -> mlookup A B m' a b.
Proof.
intros.
induction H as [a0 b0 b' m| a0 a' b0 b' m m' H HrecH H2].
apply mlooknext.
inversion H0.
rewrite H5 in H1.
absurd (a <> a).
auto.
auto.
auto.
auto.
inversion H0.
constructor.
apply mlooknext.
apply HrecH.
auto.
auto.
auto.
Qed.


Lemma mupdate_preserv_rev :
 forall (A B : Set) (m m' : Map A B) (a a' : A) (b b' : B)
   (P : forall x x' : A, x = x' \/ x <> x'),
 mupdate A B m a' b' m' -> mlookup A B m' a b -> a <> a' -> mlookup A B m a b.
Proof.
intros.
induction H as [a0 b0 b' m| a0 a' b0 b' m m' H HrecH H2].
inversion H0.
rewrite H5 in H1.
absurd (a <> a).
auto.
auto.
apply mlooknext.
auto.
auto.
cut (a' = a \/ a' <> a).
intros.
inversion_clear H3.
rewrite H4 in H0.
inversion H0.
replace a' with a.
constructor.
absurd (a <> a).
auto.
auto.
apply mlooknext.
inversion H0.
rewrite H3 in H4.
absurd (a <> a).
auto.
auto.
apply HrecH.
auto.
auto.
auto.
auto.
Qed.


Lemma mupdate_preserv_rev_ex :
 forall (A B : Set) (m m' : Map A B) (a a' : A) (b b' : B)
   (P : forall x x' : A, x = x' \/ x <> x'),
 mupdate A B m a' b' m' ->
 mlookup A B m' a b -> exists c : _, mlookup A B m a c.
Proof.
simple induction m.
simple induction m'.
intros; inversion H.
intros; inversion H0.
simple induction m'.
intros.
inversion H0.
intros.
cut (a = a1 \/ a <> a1).
intros.
inversion_clear H3.
replace a with a1.
split with b.
constructor.
cut (exists c : _, mlookup _ _ m0 a1 c).
intros.
inversion_clear H3.
split with x.
apply mlooknext.
auto.
auto.
cut (a = a' \/ a <> a').
intros.
inversion_clear H3.
rewrite H5 in H1.
inversion H1.
rewrite <- H11 in H2.
rewrite <- H5 in H2.
move H4 after H2.
inversion H2.
rewrite H10 in H4.
absurd (a1 <> a1).
auto.
auto.
split with b1.
auto.
absurd (a0 <> a0).
auto.
auto.
inversion H1.
rewrite H3 in H5.
absurd (a' <> a').
auto.
auto.
move H after H14.
inversion H2.
rewrite <- H11 in H15.
rewrite H15 in H4.
absurd (a1 <> a1).
auto.
auto.
apply (H m1 a1 a' b1 b').
auto.
auto.
auto.
auto.
apply P.
Qed.


Lemma mupdate_size_same :
 forall (A B : Set) (m m' : Map A B) (a : A) (b : B) 
   (n : nat) (P : forall x x' : A, x = x' \/ x <> x'),
 mupdate A B m a b m' -> msize A B m n -> msize A B m' n.
Proof.
simple induction m.
intros.
inversion H.
intros.
cut (a = a0 \/ a <> a0).
intros.
inversion_clear H2.
rewrite H3 in H0.
inversion H0.
inversion H1.
apply msize1.
auto.
absurd (a0 <> a0).
auto.
auto.
inversion H0.
rewrite H2 in H3.
absurd (a0 <> a0).
auto.
auto.
move H after H10.
inversion H1.
apply msize1.
apply (H m'0 a0 b0 n0).
auto.
auto.
auto.
auto.
Qed.


Lemma mupdate_size_same_rev :
 forall (A B : Set) (m m' : Map A B) (a : A) (b : B) 
   (n : nat) (P : forall x x' : A, x = x' \/ x <> x'),
 mupdate A B m a b m' -> msize A B m' n -> msize A B m n.
Proof.
simple induction m.
intros.
inversion H.
intros.
cut (a = a0 \/ a <> a0).
intros.
inversion_clear H2.
rewrite H3 in H0.
inversion H0.
rewrite <- H8 in H1.
inversion H1.
constructor.
auto.
absurd (a0 <> a0); auto.
inversion H0.
rewrite <- H8 in H1.
inversion H1.
constructor.
auto.
rewrite <- H8 in H1.
inversion H1.
constructor.
apply (H m'0 a0 b0 n0).
auto.
auto.
auto.
auto.
Qed.



Lemma mupdate_mnotindom_preserv :
 forall (A B : Set) (m m' : Map A B) (a a' : A) (b : B),
 (forall x x' : A, x = x' \/ x <> x') ->
 mupdate A B m a b m' -> mnotindom A B m a' -> mnotindom A B m' a'.
Proof.
simple induction m.
intros.
inversion H0.
intros.
cut (a = a0 \/ a <> a0).
intros.
inversion_clear H3.
rewrite H4 in H1.
inversion H1.
rewrite <- H4.
inversion H2.
apply mnotindom1.
auto.
auto.
absurd (a0 <> a0).
auto.
auto.
inversion H1.
rewrite H3 in H4.
absurd (a0 <> a0).
auto.
auto.
inversion H2.
apply mnotindom1.
auto.
apply H with a0 b0.
auto.
auto.
auto.
auto.
Qed.


(* updating an existing mapping can always be done and maintains 
   wellformed-ness *)
Lemma mupdate_always_wf :
 forall (A B : Set) (m : Map A B) (wfm : mWF _ _ m) (a : A) (b b' : B),
 (forall x x' : A, x = x' \/ x <> x') ->
 mlookup A B m a b' -> exists m' : _, mupdate A B m a b m' /\ mWF A B m'.
Proof.
simple induction m.
intros.
inversion H0.
intros.
cut (a = a0 \/ a <> a0).
intros.
inversion_clear H2.
rewrite H3.
split with (melem A B a0 b0 m0).
split.
constructor.
rewrite <- H3.
inversion wfm.
apply mwf1.
auto.
auto.
cut (exists m' : _, mupdate A B m0 a0 b0 m' /\ mWF A B m').
intros.
elim H2.
intro m''.
intros.
inversion_clear H4.
clear H2.
split with (melem A B a b m'').
split.
apply mupdnext.
auto.
auto.
apply mwf1.
inversion wfm.
apply mupdate_mnotindom_preserv with m0 a0 b0.
auto.
auto.
auto.
auto.
apply H with b'.
inversion wfm.
auto.
auto.
inversion H1.
rewrite H2 in H3.
absurd (a0 <> a0).
auto.
auto.
auto.
auto.
Qed.


(* updating the map in the same way produces equivalent map *)
Lemma mupdate_same_eq :
 forall (A B : Set) (m m' m'' : Map A B) (a : A) (b : B),
 (forall x x' : A, x = x' \/ x <> x') ->
 mupdate A B m a b m' -> mupdate A B m a b m'' -> m' = m''.
Proof.
simple induction m.
intros.
inversion H0.
intros.
cut (a = a0 \/ a <> a0).
intros.
inversion_clear H3.
rewrite H4 in H1.
rewrite H4 in H2.
inversion H1.
inversion H2.
auto.
absurd (a0 <> a0).
auto.
auto.
auto.
absurd (a0 <> a0); auto.
inversion H1.
rewrite H3 in H4.
absurd (a0 <> a0); auto.
inversion H2.
rewrite H12 in H4.
absurd (a0 <> a0); auto.
replace m'1 with m'0.
auto.
apply H with a0 b0.
auto.
auto.
auto.
auto.
Qed.



Lemma mupdate_preserv_mindom_rev :
 forall (A B : Set) (m m' : Map A B) (a a' : A) (b : B)
   (P : forall x x' : A, x = x' \/ x <> x'),
 mupdate _ _ m a b m' -> mindom _ _ m' a' -> mindom _ _ m a'.
Proof.
simple induction m; intros.
inversion H.
cut (a = a' \/ a <> a'); intros; auto.
inversion_clear H2.
rewrite <- H3; inversion H0; constructor.
constructor; auto.
inversion H0.
rewrite <- H8 in H1.
inversion H1.
rewrite H9 in H2; rewrite H2 in H3; absurd (a' <> a'); auto.
auto.
apply H with m'0 a0 b0; auto.
rewrite <- H8 in H1; inversion H1.
rewrite H11 in H3; absurd (a' <> a'); auto.
auto.
Qed.




Lemma mupdate_preserv_mindom :
 forall (A B : Set) (m m' : Map A B) (a a' : A) (b : B)
   (P : forall x x' : A, x = x' \/ x <> x'),
 mupdate _ _ m a b m' -> mindom _ _ m a' -> mindom _ _ m' a'.
simple induction m; intros.
inversion H0.
cut (a = a' \/ a <> a'); intros. 
inversion_clear H2.
rewrite H3 in H0.
inversion H0.
constructor.
constructor.
inversion H1.
rewrite H2 in H3; absurd (a' <> a'); auto.
inversion H0.
rewrite <- H9; constructor; auto.
constructor; auto.
apply H with a0 b0; auto.
auto.
Qed.




(* if you extend, then it must be in the map *)
Lemma mextend_imp_mlookup :
 forall (A B : Set) (m m' : Map A B) (a : A) (b : B),
 (forall x x' : A, x = x' \/ x <> x') ->
 mextend A B m a b m' -> mlookup A B m' a b.
Proof.
intros.
induction m as [| a0 b0 m Hrecm].
inversion H0.
constructor.
inversion H0.
constructor.
Qed.


(* if you extend a well-formed map correctly, it is still well-formed *)
Lemma mextend_wf :
 forall (A B : Set) (m m' : Map A B) (a : A) (b : B),
 mWF _ _ m -> mnotindom _ _ m a -> mextend _ _ m a b m' -> mWF _ _ m'.
Proof.
intros.
inversion H1.
constructor.
auto.
auto.
Qed.




Lemma mextend_preserv :
 forall (A B : Set) (m m' : Map A B) (a a' : A) (b b' : B),
 mextend A B m a' b' m' -> mlookup A B m a b -> a <> a' -> mlookup A B m' a b.
Proof.
intros.
induction H as (a0, b0, m, H).
apply mlooknext.
auto.
auto.
Qed.



(* ... in the reverse order *)
Lemma mextend_preserv_rev :
 forall (A B : Set) (m m' : Map A B) (a a' : A) (b b' : B),
 mextend A B m a' b' m' -> mlookup A B m' a b -> a <> a' -> mlookup A B m a b.
Proof.
simple induction m.
intros.
inversion H.
rewrite <- H6 in H0.
inversion H0.
rewrite H7 in H1.
absurd (a <> a).
auto.
auto.
inversion H12.
intros.
inversion H0.
rewrite <- H7 in H1.
move H2 after H1.
inversion H1.
rewrite H8 in H2.
absurd (a0 <> a0).
auto.
auto.
auto.
Qed.

(* extended map is one longer *)
Lemma mextend_size_Sn :
 forall (A B : Set) (m m' : Map A B) (a : A) (b : B) (s : nat),
 mextend _ _ m a b m' -> msize _ _ m s -> msize _ _ m' (S s).
Proof.
simple induction m.
intros.
inversion H.
inversion H0.
constructor.
constructor.
intros.
inversion H0.
constructor.
auto.
Qed.


(* you can always update or extend a map *)
Lemma mupdext_always :
 forall (A B : Set) (m : Map A B) (a : A) (b : B)
   (P : forall x x' : A, x = x' \/ x <> x'),
 exists m' : _, mupdext A B m a b m'.  
Proof.
intros.
induction m as [| a0 b0 m Hrecm].
split with (melem _ _ a b (mempty _ _)).
constructor.
inversion_clear Hrecm.
cut (a = a0 \/ a <> a0).
intros.
induction H0 as [H0| H0].
split with (melem _ _ a0 b m).
replace a0 with a.
constructor.
split with (melem _ _ a0 b0 x).
apply mupdext1.
auto.
auto.
apply P.
Qed.


(* updext-ing a map preserves the old mapping values for non-affect elements *)
Lemma mupdext_preserv :
 forall (A B : Set) (m m' : Map A B) (a a' : A) (b b' : B),
 mupdext A B m a' b' m' -> mlookup A B m a b -> a <> a' -> mlookup A B m' a b.
Proof.
intros.
induction H as [a0 b0| a0 a' b0 b' m m' H HrecH H2| a0 b0 b' m].
inversion H0.
inversion H0.
constructor.
apply mlooknext.
apply HrecH.
auto.
auto.
auto.
inversion H0.
rewrite H5 in H1.
absurd (a <> a).
auto.
auto.
apply mlooknext.
auto.
auto.
Qed.


(* ... in the reverse order *)
Lemma mupdext_preserv_rev :
 forall (A B : Set) (m m' : Map A B) (a a' : A) (b b' : B),
 mupdext A B m a' b' m' -> mlookup A B m' a b -> a <> a' -> mlookup A B m a b.
Proof.
intros.
induction H as [a0 b0| a0 a' b0 b' m m' H HrecH H2| a0 b0 b' m].
inversion H0.
rewrite H5 in H1.
absurd (a <> a).
auto.
auto.
auto.
inversion H0.
constructor.
apply mlooknext.
auto.
auto.
inversion H0.
rewrite H5 in H1.
absurd (a <> a).
auto.
auto.
apply mlooknext.
auto.
auto.
Qed.



Lemma mupdext_preserv_rev_mindom :
 forall (A B : Set) (m m' : Map A B) (a a' : A) (b' : B),
 (forall x x' : A, x = x' \/ x <> x') ->
 mupdext A B m a' b' m' -> mindom A B m' a -> a <> a' -> mindom A B m a.
Proof.
simple induction m.
intros.
inversion H0.
rewrite <- H6 in H1.
inversion H1.
rewrite H3 in H2.
absurd (a <> a); auto.
auto.
intros.
inversion H1.
rewrite <- H9 in H2.
cut (a = a0 \/ a <> a0).
intros.
inversion_clear H12.
rewrite H13.
constructor.
apply minnext.
auto.
apply H with m'0 a' b'.
auto.
auto.
inversion H2.
rewrite H12 in H13.
absurd (a0 <> a0).
auto.
auto.
auto.
auto.
auto.
rewrite <- H9 in H2.
inversion H2.
rewrite H10 in H3.
absurd (a0 <> a0); auto.
apply minnext.
auto.
auto.
Qed.



(* having updated or extended a map, the element is now in it *)
Lemma mupdext_imp_mlookup :
 forall (A B : Set) (m m' : Map A B) (a : A) (b : B),
 (forall x x' : A, x = x' \/ x <> x') ->
 mupdext A B m a b m' -> mlookup A B m' a b.
Proof.
simple induction m.
intros m' a b P H.
inversion H.
constructor.
intros a b m0 H m' a0 b0 P.
intros.
cut (a = a0 \/ a <> a0).
intros.
inversion_clear H1.
rewrite H2 in H0.
inversion H0.
absurd (a0 <> a0).
auto.
auto.
constructor.
inversion H0.
constructor.
apply (H m'0 a0 b0).
auto.
auto.
auto.
constructor.
apply P.
Qed.




Lemma mupdext_same_eq :
 forall (A B : Set) (m m' m'' : Map A B) (a : A) (b : B),
 (forall x x' : A, x = x' \/ x <> x') ->
 mupdext A B m a b m' -> mupdext A B m a b m'' -> m' = m''.
Proof.
simple induction m.
intros.
inversion H0.
inversion H1.
auto.
intros.
cut (a = a0 \/ a <> a0).
intros.
inversion_clear H3.
rewrite <- H4 in H1.
rewrite <- H4 in H2.
inversion H1.
absurd (a <> a).
auto.
auto.
inversion H2.
absurd (a <> a); auto.
auto.
inversion H1.
inversion H2.
replace m'1 with m'0.
auto.
apply H with a0 b0.
auto.
auto.
auto.
rewrite H12 in H11.
absurd (a0 <> a0); auto.
rewrite H3 in H4.
absurd (a0 <> a0); auto.
auto.
Qed.


(*
Lemma mupdate_imp_mindom : (A,B:Set; m,m':(Map A B); a:A; b:B)
      	       	       	   (mupdate A B m a b m') ->
			   (mindom A B m a).
Proof.
Intros.
Induction H.
Constructor.
Constructor.
Auto.
Auto.
Save.

Lemma mindom_empty_false : (A,B:Set; a:A) (mindom A B (mempty A B) a) -> False.
Proof.
Intros.
Inversion H.
Save.

Fixpoint mupdatef [A,B:Set; m:(Map A B); a:A] 
  :  B -> (mindom A B m a) -> (Map A B)
  := [b]
     <[m:(Map A B)](mindom A B m a) -> (Map A B)>
     Cases m of (mempty) => [min:(mindom A B (mempty A B) a)]
                           (False_rec (Map A B) (mindom_empty_false A B a min))
              | (melem a' b' m') 
                  => [min:(mindom A B (melem A B a' b' m') a)] 
      	       	     Cases min of 
		        (minfound a b m') => m'
                      | (minnext a' a b' m' neqa min') => m'
              	     end
     end.
*)



Lemma ext_mupdext_two_preserv :
 forall (A B : Set) (m m1 m2 : Map A B) (a a' : A) (b1 b2 b' : B),
 (forall x x' : A, x = x' \/ x <> x') ->
 mupdext A B m a b1 m1 ->
 mupdext A B m a b2 m2 ->
 a <> a' -> mlookup A B m1 a' b' -> mlookup A B m2 a' b'.
Proof.
intros.
cut (mlookup _ _ m a' b').
intros.
apply mupdext_preserv with m a b2.
auto.
auto.
auto.
apply mupdext_preserv_rev with m1 a b1.
auto.
auto.
auto.
Qed.


Lemma ext_map_oneoff_updext_same_oneoff :
 forall (A B : Set) (m1 m2 m1' m2' : Map A B) (a a0 : A) (b0 b1 b2 : B),
 (forall x x' : A, x = x' \/ x <> x') ->
 (forall (a' : A) (b' : B),
  a <> a' -> mlookup _ _ m1 a' b' -> mlookup _ _ m2 a' b') ->
 mupdext _ _ m1 a0 b0 m1' ->
 mupdext _ _ m2 a0 b0 m2' ->
 forall (a' : A) (b' : B),
 a <> a' -> mlookup _ _ m1' a' b' -> mlookup _ _ m2' a' b'.
Proof.
intros.
cut (a' = a0 \/ a' <> a0).
intros.
inversion_clear H5.
rewrite H6.
rewrite H6 in H4.
replace b' with b0.
apply mupdext_imp_mlookup with m2.
auto.
auto.
apply (mlookup_eq _ _ a0 b0 b' m1').
auto.
apply mupdext_imp_mlookup with m1; auto.
auto.
cut (mlookup _ _ m2 a' b').
intros.
apply mupdext_preserv with m2 a0 b0.
auto.
auto.
auto.
apply H0.
auto.
apply mupdext_preserv_rev with m1' a0 b0.
auto.
auto.
auto.
auto.
Qed.


Lemma ext_map_eq_updext_eq :
 forall (A B : Set) (m1 m2 m1' m2' : Map A B) (a : A) (b : B),
 (forall x x' : A, x = x' \/ x <> x') ->
 (forall (a' : A) (b' : B), mlookup _ _ m1 a' b' -> mlookup _ _ m2 a' b') ->
 mupdext _ _ m1 a b m1' ->
 mupdext _ _ m2 a b m2' ->
 forall (a' : A) (b' : B), mlookup _ _ m1' a' b' -> mlookup _ _ m2' a' b'.
Proof.
simple induction m1.
simple induction m2.
intros.
inversion H1.
inversion H2.
rewrite <- H7 in H3.
inversion H3.
constructor.
inversion H15.
intros.
inversion H2.
rewrite <- H8 in H4.
inversion H4.
inversion H3.
replace a' with a0.
replace b' with b0.
apply mlooknext.
apply mupdext_imp_mlookup with m.
auto.
auto.
auto.
inversion H4.
constructor.
rewrite H5 in H25.
absurd (a' <> a'); auto.
inversion H13.
simple induction m2.
intros.
absurd
 (forall (a' : A) (b' : B),
  mlookup A B (melem A B a b m) a' b' -> mlookup A B (mempty A B) a' b').
unfold not in |- *; intros.
cut (mlookup _ _ (mempty _ _) a b).
intros.
inversion H6.
apply H5.
constructor.
auto.
intros.
cut (a' = a1 \/ a' <> a1).
intros.
inversion_clear H6.
replace a' with a1.
replace b' with b1.
apply mupdext_imp_mlookup with (melem A B a0 b0 m0).
auto.
auto.
cut (mlookup _ _ m1' a' b1).
intros.
apply (mlookup_eq _ _ a' b1 b' m1').
auto.
auto.
auto.
replace a' with a1.
apply mupdext_imp_mlookup with (melem _ _ a b m).
auto.
auto.
cut (mlookup _ _ (melem A B a0 b0 m0) a' b').
intros.
apply mupdext_preserv with (melem _ _ a0 b0 m0) a1 b1.
auto.
auto.
auto.
apply H2.
apply mupdext_preserv_rev with m1' a1 b1.
auto.
auto.
auto.
auto.
Qed.


Lemma ext_map_lookup_eq :
 forall (A B : Set) (m1 m2 : Map A B) (a : A) (b : B),
 (forall x x' : A, x = x' \/ x <> x') ->
 (forall (a' : A) (b' : B),
  a <> a' -> mlookup _ _ m1 a' b' -> mlookup _ _ m2 a' b') ->
 mlookup _ _ m1 a b ->
 mlookup _ _ m2 a b ->
 forall (a' : A) (b' : B), mlookup _ _ m1 a' b' -> mlookup _ _ m2 a' b'.
Proof.
intros.
cut (a = a' \/ a <> a').
intros.
inversion_clear H4.
rewrite <- H5 in H3.
replace a' with a; replace b' with b.
auto.
apply (mlookup_eq _ _ a b b' m1).
auto.
auto.
auto.
apply H0.
auto.
auto.
auto.
Qed.


Lemma msize_msizef : forall A B m s, msize A B m s -> msizef A B m = s.
Proof.
simple induction m; intros.
inversion H; compute in |- *; auto.
inversion H0.
unfold msizef in |- *; fold msizef in |- *.
apply eq_add_S; auto.
Qed.