 (*
 * Coq code for the paper
 *
 *       A Syntactic Approach to Foundational Proof-Carrying Code
 *                            
 *
 *   by Zhong Shao, Valery Trifonov, Stefan Monnier, Nadeem A. Hamid, 
 *                          and Zhaozhong Ni
 *
 * (for Coq version 7.0)
 *)


(***************************************************************************)
(* Utilities for recursive type encoding *)
(***************************************************************************)
Require Import EqNat.

(* substitution *)
Fixpoint subst_aux (i : nat) (t : OmegaL i) {struct t} :
 forall j : nat, i = S j -> OmegaL j -> OmegaL j :=
  match
    t in (OmegaL i) return (forall j : nat, i = S j -> OmegaL j -> OmegaL j)
  with
  | varL n => fun (j : nat) (_ : S n = S j) (e' : OmegaL j) => e'
  | liftL n t0 =>
      fun (j : nat) (p : S n = S j) (_ : OmegaL j) =>
      eq_rec n OmegaL t0 j (eq_add_S n j p)
  | inttyL =>
      fun (j : nat) (_ : 0 = S j) (e' : OmegaL j) => e' (* not reached *)
  | codetyL n G A =>
      fun (j : nat) (p : n = S j) (e' : OmegaL j) =>
      codetyL j (substmap_aux n G j p e') A
  | tuptyL n L bl =>
      fun (j : nat) (p : n = S j) (e' : OmegaL j) =>
      tuptyL j (substlist_aux n L j p e') bl
  | rectyL n t0 =>
      fun (j : nat) (p : n = S j) (e' : OmegaL j) =>
      rectyL j (subst_aux (S n) t0 (S j) (f_equal S p) (liftL j e'))
  end
 
 with substmap_aux (i : nat) (M : OmegaL_Map i) {struct M} :
 forall j : nat, i = S j -> OmegaL j -> OmegaL_Map j :=
  match
    M in (OmegaL_Map i)
    return (forall j : nat, i = S j -> OmegaL j -> OmegaL_Map j)
  with
  | memptyL n => fun (j : nat) (_ : n = S j) (_ : OmegaL j) => memptyL j
  | melemL n r t M0 =>
      fun (j : nat) (p : n = S j) (e' : OmegaL j) =>
      melemL j r (subst_aux n t j p e') (substmap_aux n M0 j p e')
  end
 
 with substlist_aux (i : nat) (L : OmegaL_List i) {struct L} :
 forall j : nat, i = S j -> OmegaL j -> OmegaL_List j :=
  match
    L in (OmegaL_List i)
    return (forall j : nat, i = S j -> OmegaL j -> OmegaL_List j)
  with
  | nilL n => fun (j : nat) (_ : n = S j) (_ : OmegaL j) => nilL j
  | consL n t L0 =>
      fun (j : nat) (p : n = S j) (e' : OmegaL j) =>
      consL j (subst_aux n t j p e') (substlist_aux n L0 j p e')
  end.

Definition SubstL (n : nat) (T : OmegaL (S n)) :=
  subst_aux (S n) T n (refl_equal (S n)).


Definition RUnfold (T : OmegaL 1) : OmegaL 0 := SubstL 0 T (rectyL 0 T).




(* conversion to the real Omega *)
Fixpoint RUnlift (T : OmegaL 0) : Omega :=
  match T in (OmegaL n) return (n = 0 -> Omega) with
  | varL i =>
      fun p : S i = 0 =>
      False_rec Omega
        (eq_ind (S i)
           (fun e : nat => match e with
                           | O => False
                           | _ => True
                           end) I 0 p)
  | liftL i T0 =>
      fun p : S i = 0 =>
      False_rec Omega
        (eq_ind (S i)
           (fun e : nat => match e with
                           | O => False
                           | _ => True
                           end) I 0 p)
  | inttyL => fun _ : 0 = 0 => intty
  | rectyL i T0 =>
      fun p : i = 0 => recty (eq_rec (S i) OmegaL T0 1 (eq_S i 0 p))
  | codetyL i M A =>
      fun p : i = 0 => codety (RUnlift_Map (eq_rec i OmegaL_Map M 0 p)) A
  | tuptyL i L bl =>
      fun p : i = 0 => tupty (RUnlift_List (eq_rec i OmegaL_List L 0 p)) bl
  end (refl_equal 0)
 
 with RUnlift_Map (M : OmegaL_Map 0) : Map Reg Omega :=
  match M in (OmegaL_Map n) return (n = 0 -> Map Reg Omega) with
  | memptyL i => fun _ : i = 0 => mempty Reg Omega
  | melemL i r T M0 =>
      fun p : i = 0 =>
      melem Reg Omega r (RUnlift (eq_rec i OmegaL T 0 p))
        (RUnlift_Map (eq_rec i OmegaL_Map M0 0 p))
  end (refl_equal 0)
 
 with RUnlift_List (L : OmegaL_List 0) : list Omega :=
  match L in (OmegaL_List n) return (n = 0 -> list Omega) with
  | nilL i => fun _ : i = 0 => nil
  | consL i T L0 =>
      fun p : i = 0 =>
      RUnlift (eq_rec i OmegaL T 0 p)
      :: RUnlift_List (eq_rec i OmegaL_List L0 0 p)
  end (refl_equal 0).

(***************************************************************************)
(***************************************************************************)
(***************************************************************************)

(* samples:
Definition T := (rectyL (S O) (tuptyL (S (S O)) 
      	       	       	       	     (consL (S (S O)) 
      	       	       	       	       	    (varL (S O)) 
      	       	       	       	       	    (nilL (S (S O))))
      	       	       	       	     (cons true (nil ?)))).
*)




