 (*
 * Coq code for the paper
 *
 *       A Syntactic Approach to Foundational Proof-Carrying Code
 *                            
 *
 *   by Zhong Shao, Valery Trifonov, Stefan Monnier, Nadeem A. Hamid, 
 *                          and Zhaozhong Ni
 *
 * (for Coq version 7.0)
 *)


(***************************************************************************)
(* Typed Instruction Set *)
(***************************************************************************)

Require Import EqNat.
Require Import natutil.

(***************************************************************************)
(* Syntax *)

Require Import tisreg.

Notation Word := nat (only parsing).
Notation Mem := (nat -> nat) (only parsing).
Notation _RegFile := (_Reg -> nat) (only parsing).
Notation State := ((nat -> nat) * ((_Reg -> nat) * nat))%type (only parsing).

Inductive _Instr : Set :=
  | _add : _Reg -> _Reg -> _Reg -> _Instr
  | _addi : _Reg -> _Reg -> nat -> _Instr
  | _movi : _Reg -> nat -> _Instr
  | _bgt : _Reg -> _Reg -> nat -> _Instr
  | _jd : nat -> _Instr
  | _jmp : _Reg -> _Instr
  | _ld : _Reg -> _Reg -> nat -> _Instr
  | _st : _Reg -> nat -> _Reg -> _Instr
  | _ill : _Instr.


(***************************************************************************)

(* decode utilities *)
Definition icoded (w : nat) : nat := mod_ w (8).
Definition r1argd (w : nat) : nat := mod_ (div8 w) (32).
Definition r2argd (w : nat) : nat := mod_ (div32 (div8 w)) (32).
Definition r3argd (w : nat) : nat := mod_ (div32 (div32 (div8 w))) (32).
Definition w1argd (w : nat) : nat := div8 w.
Definition w2argd (w : nat) : nat := div32 (div8 w).
Definition w3argd (w : nat) : nat := div32 (div32 (div8 w)).


(* encode utilities *)
Definition icodee (n : nat) : nat := n.
Definition r1arge (n : nat) : nat := mult8 n.
Definition r2arge (n : nat) : nat := mult32 (mult8 n).
Definition r3arge (n : nat) : nat := mult32 (mult32 (mult8 n)).
Definition w1arge (n : nat) : nat := mult8 n.
Definition w2arge (n : nat) : nat := mult32 (mult8 n).
Definition w3arge (n : nat) : nat := mult32 (mult32 (mult8 n)).


(* Decode function *)
Definition Dc (w : nat) : _Instr :=
  match icoded w with
  | O =>
      _add (nat2_reg (r1argd w)) (nat2_reg (r2argd w)) (nat2_reg (r3argd w))
  | S O => _addi (nat2_reg (r1argd w)) (nat2_reg (r2argd w)) (w3argd w)
  | S (S O) => _movi (nat2_reg (r1argd w)) (w2argd w)
  | S (S (S O)) =>
      _bgt (nat2_reg (r1argd w)) (nat2_reg (r2argd w)) (w3argd w)
  | S (S (S (S O))) => _jd (w1argd w)
  | S (S (S (S (S O)))) => _jmp (nat2_reg (r1argd w))
  | S (S (S (S (S (S O))))) =>
      _ld (nat2_reg (r1argd w)) (nat2_reg (r2argd w)) (w3argd w)
  | S (S (S (S (S (S (S O)))))) =>
      _st (nat2_reg (r1argd w)) (w3argd w) (nat2_reg (r2argd w))
  | _ => _ill
  end.


(* Encode function *)
Definition Ec (i : _Instr) : nat :=
  match i with
  | _add rd rs rs' =>
      icodee 0 +
      (r1arge (_reg2nat rd) + (r2arge (_reg2nat rs) + r3arge (_reg2nat rs')))
  | _addi rd rs w =>
      icodee (1) + (r1arge (_reg2nat rd) + (r2arge (_reg2nat rs) + w3arge w))
  | _movi rd w => icodee (2) + (r1arge (_reg2nat rd) + w2arge w)
  | _bgt rd rs w =>
      icodee (3) + (r1arge (_reg2nat rd) + (r2arge (_reg2nat rs) + w3arge w))
  | _jd w => icodee (4) + w1arge w
  | _jmp r => icodee (5) + r1arge (_reg2nat r)
  | _ld rd rs w =>
      icodee (6) + (r1arge (_reg2nat rd) + (r2arge (_reg2nat rs) + w3arge w))
  | _st rd w rs =>
      icodee (7) + (r1arge (_reg2nat rd) + (r2arge (_reg2nat rs) + w3arge w))
  | _ill => 0
  end.


(***************************************************************************)
(* Machine semantics *)
Definition updatereg (R : _Reg -> nat) (rd : _Reg) 
  (v : nat) : _Reg -> nat := fun r : _Reg => if beq_reg r rd then v else R r.

Definition updatemem (M : nat -> nat) (a v : nat) : 
  nat -> nat := fun w : nat => if beq_nat w a then v else M w.

Definition Step (St : (nat -> nat) * ((_Reg -> nat) * nat)) :
  (nat -> nat) * ((_Reg -> nat) * nat) :=
  match St with
  | (M, (R, pc)) =>
      match Dc (M pc) with
      | _add rd rs rs' => (M, (updatereg R rd (R rs + R rs'), S pc))
      | _addi rd rs w => (M, (updatereg R rd (R rs + w), S pc))
      | _movi rd w => (M, (updatereg R rd w, S pc))
      | _bgt rs rt w =>
          if blt_nat (R rt) (R rs) then (M, (R, w)) else (M, (R, S pc))
      | _jd w => (M, (R, w))
      | _jmp r => (M, (R, R r))
      | _ld rd rs w => (M, (updatereg R rd (M (R rs + w)), S pc))
      | _st rd w rs => (updatemem M (R rd + w) (R rs), (R, S pc))
      | _ill => St
      end
  end.

Fixpoint MultiStep (n : nat) :
 (nat -> nat) * ((_Reg -> nat) * nat) -> (nat -> nat) * ((_Reg -> nat) * nat) :=
  match n with
  | O => fun S : (nat -> nat) * ((_Reg -> nat) * nat) => Step S
  | S m =>
      fun S : (nat -> nat) * ((_Reg -> nat) * nat) => Step (MultiStep m S)
  end.

(***************************************************************************)


