(*
 * Coq code for the paper
 *
 *       A Syntactic Approach to Foundational Proof-Carrying Code
 *                            
 *
 *   by Zhong Shao, Valery Trifonov, Stefan Monnier, Nadeem A. Hamid, 
 *                          and Zhaozhong Ni
 *
 * (for Coq version 7.3)
 *)


(***************************************************************************)
(* Translation from Mini-FTAL to machine state *)
(***************************************************************************)

Require Import Le.
Require Import Lt.
Require Import Gt.
Require Import Plus.

Require Import ftal.
Load "tis".


(***************************************************************************)
(* the safety policy *)
Definition SP (S : (nat -> nat) * ((_Reg -> nat) * nat)) :=
  let (M, T') := S in let (R, PC) := T' in Dc (M PC) <> _ill.


Definition Safe (S : (nat -> nat) * ((_Reg -> nat) * nat)) :=
  forall n : nat, SP (MultiStep n S).


(***************************************************************************)
(* ---------- the layout function ----------------- *)

Definition HV_size (hv : HeapVal) : nat :=
  match hv with
  | tuple V => length V
  | code G T I' => lenInstrSeq I'
  end.


Fixpoint Layout_aux (H : Map nat HeapVal) : nat -> nat :=
  match H with
  | mempty => fun l' : nat => 0
  | melem l h H' =>
      fun l' : nat =>
      if blt_nat l l' then Layout_aux H' l' + HV_size h else Layout_aux H' l'
  end.


Definition Layout (H : Heap) : nat -> nat :=
  Layout_aux (unwrapMap nat HeapVal H).


Notation LayoutF := (nat -> nat) (only parsing).


(***************************************************************************)
(* translations *)

(* register translations: *)
Require Import translate_ftal_aux.



Inductive TrWordVal : (nat -> nat) -> WordVal -> nat -> Prop :=
  | trwv_label : forall (L : nat -> nat) (l : nat), TrWordVal L (wl l) (L l)
  | trwv_int : forall (L : nat -> nat) (i : nat), TrWordVal L (wi i) i
  | trwv_uninit :
      forall (L : nat -> nat) (t : Omega) (l : nat),
      TrWordVal L (wuninit t) l
  | trwv_fold :
      forall (L : nat -> nat) (v : WordVal) (t : Omega) (w : nat),
      TrWordVal L v w -> TrWordVal L (wfold v t) w.


Inductive TrWordValList :
(nat -> nat) -> list WordVal -> (nat -> nat) -> nat -> Prop :=
  | trwvl_nil : forall (L M : nat -> nat) (l : nat), TrWordValList L nil M l
  | trwvl_cons :
      forall (L : nat -> nat) (v : WordVal) (V : list WordVal)
        (M : nat -> nat) (w l : nat),
      TrWordVal L v w ->
      M l = w -> TrWordValList L V M (S l) -> TrWordValList L (v :: V) M l.


Inductive TrInstr : (nat -> nat) -> Instr -> _Instr -> Prop :=
  | tri_add :
      forall (L : nat -> nat) (rd rs rt : Reg) (_rd _rs _rt : _Reg),
      TrRegF rd = _rd ->
      TrRegF rs = _rs ->
      TrRegF rt = _rt -> TrInstr L (add rd rs rt) (_add _rd _rs _rt)
  | tri_addi :
      forall (L : nat -> nat) (rd rs : Reg) (_rd _rs : _Reg) (i : nat),
      TrRegF rd = _rd ->
      TrRegF rs = _rs -> TrInstr L (addi rd rs i) (_addi _rd _rs i)
  | tri_alloc :
      forall (L : nat -> nat) (rd : Reg) (_rd : _Reg) (Ts : list Omega),
      TrRegF rd = _rd -> TrInstr L (alloc rd Ts) (_addi _rd _r31 0)
  | tri_bgt :
      forall (L : nat -> nat) (rs rt : Reg) (_rs _rt : _Reg) (l : nat),
      TrRegF rs = _rs ->
      TrRegF rt = _rt -> TrInstr L (bgt rs rt l) (_bgt _rs _rt (L l))
  | tri_bump :
      forall (L : nat -> nat) (i : nat),
      TrInstr L (bump i) (_addi _r31 _r31 i)
  | tri_fold :
      forall (L : nat -> nat) (rd rs : Reg) (t : Omega) (_rd _rs : _Reg),
      TrRegF rd = _rd ->
      TrRegF rs = _rs -> TrInstr L (fold rd t rs) (_addi _rd _rs 0)
  | tri_ld :
      forall (L : nat -> nat) (rd rs : Reg) (i : nat) (_rd _rs : _Reg),
      TrRegF rd = _rd ->
      TrRegF rs = _rs -> TrInstr L (ld rd rs i) (_ld _rd _rs i)
  | tri_mov :
      forall (L : nat -> nat) (rd rs : Reg) (_rd _rs : _Reg),
      TrRegF rd = _rd ->
      TrRegF rs = _rs -> TrInstr L (mov rd rs) (_addi _rd _rs 0)
  | tri_movi :
      forall (L : nat -> nat) (r : Reg) (_r : _Reg) (i : nat),
      TrRegF r = _r -> TrInstr L (movi r i) (_movi _r i)
  | tri_movl :
      forall (L : nat -> nat) (r : Reg) (_r : _Reg) (l : nat),
      TrRegF r = _r -> TrInstr L (movl r l) (_movi _r (L l))
  | tri_st :
      forall (L : nat -> nat) (rd rs : Reg) (i : nat) (_rd _rs : _Reg),
      TrRegF rd = _rd ->
      TrRegF rs = _rs -> TrInstr L (st rd i rs) (_st _rd i _rs)
  | tri_unfold :
      forall (L : nat -> nat) (rd rs : Reg) (_rd _rs : _Reg),
      TrRegF rd = _rd ->
      TrRegF rs = _rs -> TrInstr L (unfold rd rs) (_addi _rd _rs 0).


Inductive TrInstrSeq :
(nat -> nat) -> InstrSeq -> (nat -> nat) -> nat -> Prop :=
  | tris_iseq :
      forall (L : nat -> nat) (i : Instr) (_i : _Instr) 
        (I : InstrSeq) (M : nat -> nat) (l : nat),
      TrInstr L i _i ->
      Dc (M l) = _i -> TrInstrSeq L I M (S l) -> TrInstrSeq L (iseq i I) M l
  | tris_jd :
      forall (L : nat -> nat) (l : nat) (M : nat -> nat) (w : nat),
      Dc (M w) = _jd (L l) -> TrInstrSeq L (jd l) M w
  | tris_jmp :
      forall (L : nat -> nat) (r : Reg) (_r : _Reg) 
        (M : nat -> nat) (w : nat),
      TrRegF r = _r -> Dc (M w) = _jmp _r -> TrInstrSeq L (jmp r) M w.


Inductive TrHeapVal :
(nat -> nat) -> HeapVal -> (nat -> nat) -> nat -> Prop :=
  | trhv_tuple :
      forall (L : nat -> nat) (V : list WordVal) (M : nat -> nat) (l : nat),
      TrWordValList L V M l -> TrHeapVal L (tuple V) M l
  | trhv_code :
      forall (L : nat -> nat) (G : Map Reg Omega) (T : APTy) 
        (I : InstrSeq) (M : nat -> nat) (l : nat),
      TrInstrSeq L I M l -> TrHeapVal L (code G T I) M l.


Inductive TrHeap : (nat -> nat) -> Heap -> (nat -> nat) -> Prop :=
    trheap :
      forall (L : nat -> nat) (H : Heap) (s : nat) (M : nat -> nat),
      msize nat HeapVal (unwrapMap nat HeapVal H) s ->
      (forall (n : nat) (h : HeapVal),
       (*		    (lt n s) ->  *)
       mlookup nat HeapVal (unwrapMap nat HeapVal H) n h ->
       TrHeapVal L h M (L n)) -> TrHeap L H M.


Inductive TrRegFile : (nat -> nat) -> RegFile -> (_Reg -> nat) -> Prop :=
    trregfile :
      forall (L : nat -> nat) (R : RegFile) (_R : _Reg -> nat),
      (forall (r : Reg) (_r : _Reg) (v : WordVal),
       mlookup _ _ R r v -> TrRegF r = _r -> TrWordVal L v (_R _r)) ->
      (*		    (_R _r)=w ->
      		    (TrWordVal L v w)) -> *)
      TrRegFile L R _R.


Inductive TrAP : (nat -> nat) -> AP -> (_Reg -> nat) -> Prop :=
    trap :
      forall (L : nat -> nat) (A : AP) (_R : _Reg -> nat),
      _R _r31 = L A -> TrAP L A _R.  


Inductive TrProgram :
Program -> (nat -> nat) * ((_Reg -> nat) * nat) -> Prop :=
    trprogram :
      forall (H : Heap) (R : RegFile) (A : AP) (I : InstrSeq)
        (M : nat -> nat) (_R : _Reg -> nat) (pc : nat) 
        (L : nat -> nat),
      L = Layout H ->
      TrHeap L H M ->
      TrRegFile L R _R ->
      TrAP L A _R ->
      (exists l : _,
         (exists G : _,
            (exists T : _,
               (exists I' : _,
                  (exists n : _,
                     mlookup nat HeapVal (unwrapMap nat HeapVal H) l
                       (code G T I') /\
                     ISubDepth I I' n /\
                     TrInstrSeq L I' M (L l) /\ L l + n = pc))))) -> 
      (* 		   (TrInstrSeq L I M pc) ->  *)
      TrProgram (H, (R, (A, I))) (M, (_R, pc)).


(***************************************************************************)


(***************************************************************************)
(* FPCC proofs *)


Definition Inv (S : (nat -> nat) * ((_Reg -> nat) * nat)) :=
  exists P : Program, WFProgram P /\ TrProgram P S.
(*Definition Inv := [S:State] (EX P:Program |
     	       	       	      (EXT D:(WFProgram P) |
 			         (TrProgram P S))).*)



(***************************************************************************)
(* utilities *)
Lemma tr_sub_instrseq :
 forall (IS IS' : InstrSeq) (n : nat) (L M : nat -> nat) (l : nat),
 ISubDepth IS' IS n -> TrInstrSeq L IS M l -> TrInstrSeq L IS' M (n + l).
Proof.
simple induction IS.
intros.
inversion H0.
auto.
replace (S n0 + l) with (n0 + S l).
apply H.
auto.
inversion H1.
auto.
apply sym_eq.
apply plus_Snm_nSm.
intros.
inversion H.
replace (0 + l) with l.
auto.
auto.
intros.
inversion H.
replace (0 + l) with l.
auto.
auto.
Qed.


Lemma tr_pc_trinstrseq :
 forall (H : Heap) (I : InstrSeq) (L M : nat -> nat) (pc : nat),
 (exists l : _,
    (exists G : _,
       (exists T : _,
          (exists I' : _,
             (exists n : _,
                mlookup nat HeapVal (unwrapMap nat HeapVal H) l (code G T I') /\
                ISubDepth I I' n /\ TrInstrSeq L I' M (L l) /\ L l + n = pc))))) ->
 TrInstrSeq L I M pc.
Proof.
intros.
inversion_clear H0.
inversion_clear H1.
inversion_clear H0.
inversion_clear H1.
inversion_clear H0.
inversion_clear H1.
inversion_clear H2.
inversion_clear H3.
rewrite <- H4.
replace (L x + x3) with (x3 + L x).
apply tr_sub_instrseq with x2.
auto.
auto.
apply plus_comm.
Qed.





Lemma tr_instrseq_subdepth_S :
 forall (IS' IS : InstrSeq) (i : Instr) (n : nat),
 ISubDepth (iseq i IS) IS' n -> ISubDepth IS IS' (S n).
Proof.
simple induction IS'.
intros.
constructor.
inversion H0.
constructor.
apply H with i1.
auto.
intros.
inversion H.
intros.
inversion H.
Qed.


Lemma tr_pc_Spc :
 forall (H : Heap) (i : Instr) (I : InstrSeq) (L M : nat -> nat) (pc : nat),
 (exists l : _,
    (exists G : _,
       (exists T : _,
          (exists I' : _,
             (exists n : _,
                mlookup nat HeapVal (unwrapMap nat HeapVal H) l (code G T I') /\
                ISubDepth (iseq i I) I' n /\
                TrInstrSeq L I' M (L l) /\ L l + n = pc))))) ->
 exists l : _,
   (exists G : _,
      (exists T : _,
         (exists I' : _,
            (exists n : _,
               mlookup nat HeapVal (unwrapMap nat HeapVal H) l (code G T I') /\
               ISubDepth I I' n /\ TrInstrSeq L I' M (L l) /\ L l + n = S pc)))).
Proof.
intros.
inversion_clear H0.
inversion_clear H1.
inversion_clear H0.
inversion_clear H1.
inversion_clear H0.
inversion_clear H1.
inversion_clear H2.
inversion_clear H3.
split with x.
split with x0.
split with x1.
split with x2.
split with (S x3).
split.
auto.
split.
apply tr_instrseq_subdepth_S with i.
auto.
split.
auto.
replace (L x + S x3) with (S x3 + L x).
unfold plus in |- * .
apply eq_S.
fold plus in |- * .
replace (x3 + L x) with (L x + x3).
auto.
apply plus_comm.
apply plus_comm.
Qed.





(***************************************************************************)
(* Progress theorem *)

Theorem Progress :
 forall S : (nat -> nat) * ((_Reg -> nat) * nat), Inv S -> SP S.
Proof.
intros.
unfold Inv in H.
elim H; clear H; intros P H.
elim H; clear H; intros D H.
inversion D.
clear H5.
rewrite <- H6 in H.
inversion H.
cut (TrInstrSeq L I M pc).
intros.
inversion H4.
rewrite <- H25 in H17.
inversion H17.
red in |- * .
rewrite H30.
inversion H28.
discriminate.

rewrite <- H24 in H17.
inversion H17.
red in |- * .
rewrite H29.
inversion H27.
discriminate.

rewrite <- H24 in H17.
inversion H17.
red in |- * .
rewrite H29.
inversion H27.
discriminate.
auto.

rewrite <- H26 in H17.
red in |- * .
inversion H17.
rewrite H31.
inversion H29.
discriminate.

rewrite <- H22 in H17.
red in |- * .
inversion H17.
rewrite H27.
inversion H25.
discriminate.

rewrite <- H25 in H17.
inversion H17.
red in |- * .
rewrite H30.
inversion H28.
discriminate.

rewrite <- H23 in H17.
red in |- * .
inversion H17.
rewrite H26.
discriminate.
auto.

rewrite <- H23 in H17.
inversion H17.
red in |- * .
rewrite H27.
discriminate.

rewrite <- H26 in H17.
inversion H17.
red in |- * .
rewrite H31.
inversion H29.
discriminate.
auto.

rewrite <- H24 in H17.
inversion H17.
red in |- * .
rewrite H29.
inversion H27.
discriminate.
auto.

rewrite <- H23 in H17.
inversion H17.
red in |- * .
rewrite H28.
inversion H26.
discriminate. auto.

rewrite <- H25 in H17.
inversion H17.
red in |- * .
rewrite H30.
inversion H28.
discriminate. auto.

rewrite <- H27 in H17.
inversion H17.
red in |- * .
rewrite H32.
inversion H30.
discriminate.

rewrite <- H26 in H17.
inversion H17.
red in |- * .
rewrite H31.
inversion H29.
discriminate.

apply tr_pc_trinstrseq with H0.
auto.
Qed.





Lemma Preservation_add :
 forall (S : (nat -> nat) * ((_Reg -> nat) * nat)) 
   (H : Heap) (R : RegFile) (a : AP) (rd rs rs' : Reg) 
   (I : InstrSeq) (P P' : Program),
 P = (H, (R, (a, iseq (add rd rs rs') I))) ->
 WFProgram P ->
 TrProgram P S -> Eval P P' -> WFProgram P' -> TrProgram P' (Step S).
Proof.
intros S H R a rd rs rs' I P P' H0 H1 H2 H3 H4.
rewrite H0 in H1.
inversion H1.
clear H14.
rewrite H0 in H2.
inversion H2.
unfold Step in |- * .
cut (TrInstrSeq L (iseq (add rd rs rs') I) M pc).
intros.
inversion H25.
rewrite <- H30 in H28.
inversion H28.

unfold updatereg in |- * .
rewrite H0 in H3.
inversion H3.
apply trprogram with L.
auto.
auto.
constructor.
intros.
cut (_r = _rd \/ _r <> _rd).
intros.
inversion_clear H56.
rewrite H57.
replace (beq_reg _rd _rd) with true.
cut (v = wi (rsval + rsval')).
intros.
rewrite H56.
replace (_R _rs) with rsval.
replace (_R _rt) with rsval'.
constructor.
inversion H22.
cut (TrWordVal L (wi rsval') (_R _rt)).
intros.
inversion H62.
auto.
apply H58 with rs'.
auto.
auto.
cut (TrWordVal L (wi rsval) (_R _rs)).
intros.
inversion H58.
auto.
inversion H22.
apply H58 with rs.
auto.
auto.
apply (mlookup_eq _ _ rd v (wi (rsval + rsval')) R').
apply eqorneq_reg.
replace rd with r.
auto.
apply trregf_eq_eq with _rd.
rewrite <- H57.
auto.
auto.
apply mupdext_imp_mlookup with R.
apply eqorneq_reg.
auto.
apply sym_eq.
apply beq_reg_eq_true.
auto.
replace (beq_reg _r _rd) with false.
inversion H22.
auto.
apply H56 with r.
apply mupdext_preserv_rev with R' rd (wi (rsval + rsval')).
auto.
auto.
apply trregf_neq_neq with _r _rd.
auto.
auto.
auto.
auto.
auto.
(*apply sym_eq.
apply beq_reg_neq_false.
auto.*)
apply eqorneq_tisreg.
constructor.
replace (beq_reg _r31 _rd) with false.
inversion H23.
auto.
cut (_rd <> _r31).
intros.
apply sym_eq.
apply beq_reg_neq_false.
auto.
apply trreg_neq_31 with rd.
auto.
apply tr_pc_Spc with (add rd rs rs').
auto.
apply tr_pc_trinstrseq with H.
auto.
Qed.



Lemma Preservation_addi :
 forall (S : (nat -> nat) * ((_Reg -> nat) * nat)) 
   (H : Heap) (R : RegFile) (a : AP) (rd rs : Reg) 
   (i : nat) (I : InstrSeq) (P P' : Program),
 P = (H, (R, (a, iseq (addi rd rs i) I))) ->
 WFProgram P ->
 TrProgram P S -> Eval P P' -> WFProgram P' -> TrProgram P' (Step S).
Proof.
intros St.
intros.
rewrite H0 in H1.
inversion H1.
clear H14.
rewrite H0 in H2.
inversion H2.
unfold Step in |- * .
cut (TrInstrSeq L (iseq (addi rd rs i) I) M pc).
intros.
inversion H25.
rewrite H30.
inversion H28.
rewrite H0 in H3.
inversion H3.
apply trprogram with L.
auto.
auto.
constructor.
intros.
unfold updatereg in |- * .
cut (_r = _rd \/ _r <> _rd).
intros.
inversion_clear H54.
rewrite H55.
replace (beq_reg _rd _rd) with true.
cut (v = wi (rsval + i)).
intros.
rewrite H54.
replace (_R _rs) with rsval.
constructor.
inversion H22.
cut (TrWordVal L (wi rsval) (_R _rs)).
intros.
inversion H60.
auto.
apply H56 with rs.
auto.
auto.
apply (mlookup_eq _ _ rd v (wi (rsval + i)) R').
apply eqorneq_reg.
replace rd with r.
auto.
apply trregf_eq_eq with _rd.
rewrite <- H55.
auto.
auto.
apply mupdext_imp_mlookup with R.
apply eqorneq_reg.
auto.
apply sym_eq; auto.
replace (beq_reg _r _rd) with false.
inversion H22.
apply H54 with r.
apply mupdext_preserv_rev with R' rd (wi (rsval + i)); auto.
apply trregf_neq_neq with _r _rd; auto.
auto.
auto.
apply eqorneq_tisreg.
constructor.
unfold updatereg in |- * .
replace (beq_reg _r31 _rd) with false.
inversion H23.
auto.
apply sym_eq.
apply beq_reg_neq_false.
cut (_rd <> _r31).
auto.
apply trreg_neq_31 with rd.
auto.
apply tr_pc_Spc with (addi rd rs i).
auto.
apply tr_pc_trinstrseq with H.
auto.
Qed.





Lemma heap_size_eq_label :
 forall (h : Map nat HeapVal) (wfh : mWF _ _ h) (ht : Map nat Omega)
   (wfht : mWF _ _ ht) (l : nat),
 WFHeap (wfmap _ _ h wfh) (wfmap _ _ ht wfht) ->
 msize _ _ h (S l) -> exists hv : _, (exists h' : _, h = melem _ _ l hv h').
Proof.
intros.
inversion H.
inversion_clear H2.
inversion_clear H5.
inversion_clear H6.
inversion_clear H7.
inversion_clear H8.
inversion_clear H9.
inversion_clear H10.
inversion_clear H11.
inversion H0.
split with b.
split with m.
unfold unwrapMap in H2.
rewrite <- H13 in H2.
inversion H2.
replace l with n0.
replace a with n0.
auto.
apply sym_eq.
apply eq_add_S.
rewrite H16.
unfold unwrapMap in H12.
rewrite <- H13 in H12.
inversion H12.
apply (msize_eq _ _ (melem nat HeapVal a b m)).
constructor.
auto.
auto.
apply (msize_eq _ _ m).
auto.
auto.
Qed.


Lemma heap_label_eq_size :
 forall (l : nat) (hv : HeapVal) (h : Map nat HeapVal) 
   (wf : mWF _ _ h) (wf' : mWF _ _ (melem _ _ l hv h)) 
   (G G' : HeapTy),
 WFHeap (wfmap _ _ h wf) G ->
 WFHeap (wfmap _ _ (melem _ _ l hv h) wf') G' -> msize _ _ h l.
Proof.
intros.
inversion_clear H.
inversion_clear H0.
unfold unwrapMap in H2.
unfold unwrapMap in H1.
inversion_clear H2.
inversion_clear H.
inversion_clear H2.
inversion_clear H3.
inversion_clear H4.
inversion_clear H5.
inversion_clear H6.
inversion_clear H1.
inversion_clear H6.
inversion_clear H8.
inversion_clear H9.
inversion_clear H10.
inversion_clear H11.
clear H7.
clear H12.
clear H9.
clear H6.
clear H3.
clear H.
clear H5.
replace l with x; auto.
cut (x0 = S x).
intros.
rewrite H in H1.
rewrite H in H8.
rewrite H in H10.
cut (S x <= l \/ l < S x).
intros.
inversion_clear H3.
cut (l < S x).
intros.
absurd (S x <= l).
apply lt_not_le.
auto.
auto.
apply H8 with hv.
constructor.
red in H5.
cut (l = x \/ S l < S x).
intros.
inversion_clear H3.
auto.
cut (l < x).
intros.
inversion wf'.
absurd (mindom _ _ h l).
apply mnotindom_imp_not_mindom.
auto.
auto.
cut (exists h0 : _, mlookup _ _ h l h0).
intros.
inversion_clear H14.
apply mlookup_imp_mindom with x1.
auto.
apply H4.
auto.
apply lt_S_n.
auto.
cut (S l < S x \/ l = x).
intros.
inversion_clear H3.
right.
auto.
left; auto.
cut (l < x \/ l = x).
intros.
inversion_clear H3.
left.
apply lt_n_S.
auto.
right; auto.
cut (l <= x).
intros.
apply le_lt_or_eq.
auto.
apply le_S_n.
auto.
apply le_or_lt.
inversion H1.
replace n with x.
auto.
apply (msize_eq _ _ h x n).
auto.
auto.
Qed.




Lemma tr_extend_layout_aux_eq_preserv :
 forall (h h' : Map nat HeapVal) (wfh : mWF _ _ h) 
   (wfh' : mWF _ _ h') (l : nat) (hv : HeapVal),
 h' = melem _ _ l hv h -> msize _ _ h l -> Layout_aux h l = Layout_aux h' l.
Proof.
simple induction h.
intros.
inversion H0.
rewrite H.
rewrite <- H2.
unfold Layout_aux in |- * .
unfold blt_nat in |- * .
auto.
intros.
rewrite H0.
unfold Layout_aux in |- * .
replace (blt_nat l l) with false.
auto.
apply sym_eq.
apply blt_eq_false.
Qed.



Lemma tr_extend_layout_aux_preserv :
 forall (h h' : Map nat HeapVal) (wfh : mWF _ _ h) 
   (wfh' : mWF _ _ h') (l n : nat) (hv : HeapVal),
 h' = melem _ _ l hv h ->
 msize _ _ h l -> n < l -> Layout_aux h n = Layout_aux h' n.
Proof.
simple induction h.
intros.
inversion H0.
rewrite <- H3 in H1.
inversion H1.
intros.
clear H.
apply sym_eq.
rewrite H0.
unfold Layout_aux at 1 in |- * .
replace (blt_nat l n) with false.
auto.
apply sym_eq.
apply blt_false.
auto.
Qed.


Lemma tr_extend_layout_preserv :
 forall (h h' : Map nat HeapVal) (wfh : mWF _ _ h) 
   (wfh' : mWF _ _ h') (l n : nat) (hv : HeapVal),
 mextend _ _ h l hv h' ->
 msize _ _ h l ->
 n < l -> Layout (wfmap _ _ h' wfh') n = Layout (wfmap _ _ h wfh) n.
Proof.
intros h.
elim h.
intros.
inversion H0.
rewrite <- H3 in H1. 
inversion H1.
clear h.
intros a b h.
intros.
unfold Layout in |- *; unfold unwrapMap in |- * .
inversion H0.
unfold Layout_aux at 1 in |- * .
replace (blt_nat l n) with false.
unfold Layout_aux in |- * .
auto.
apply sym_eq.
apply blt_false.
auto.
Qed.



Lemma tr_extend_heap_uninitwordlist :
 forall (Ts : list Omega) (H : Heap) (M : nat -> nat) (l : nat),
 TrWordValList (Layout H) (makeUninitTup Ts) M l.
Proof.
simple induction Ts.
intros.
constructor.
intros.
unfold makeUninitTup in |- * .
fold makeUninitTup in |- * .
apply trwvl_cons with (M l0).
constructor.
auto.
apply H.
Qed.



Lemma tr_extend_instr_preserv :
 forall (i : Instr) (I : InstrSeq) (R : Map Reg Omega) 
   (T : APTy) (_i : _Instr) (l : nat) (Ts : list Omega)
   (h h' : Map nat HeapVal) (wf : mWF _ _ h) (wf' : mWF _ _ h')
   (G G' : HeapTy),
 h' = melem _ _ l (tuple (makeUninitTup Ts)) h ->
 WFHeap (wfmap _ _ h wf) G ->
 WFHeap (wfmap _ _ h' wf') G' ->
 WFInstrSeq G R T (iseq i I) ->
 TrInstr (Layout_aux h) i _i -> TrInstr (Layout_aux h') i _i.
Proof.
simple induction i.
intros.
inversion H3.
constructor.
auto.
auto.
auto.
intros.
inversion H3.
constructor.
auto.
auto.
intros.
inversion H3.
constructor.
auto.

intros. inversion H3. (* bgt *)
replace (Layout_aux h n) with (Layout_aux h' n).
constructor.
auto.
auto.
apply sym_eq.
apply tr_extend_layout_aux_preserv with l (tuple (makeUninitTup Ts)).
auto.
auto.
auto.
cut (msize _ _ h' (S l)).
intros.
rewrite H in H11.
inversion H11.
auto.
inversion H1.
unfold unwrapMap in H12.
inversion_clear H12.
inversion_clear H15.
inversion_clear H16.
inversion_clear H17.
inversion_clear H18.
inversion_clear H19.
inversion_clear H20.
inversion_clear H21.
replace (S l) with x.
auto.
apply (msize_eq _ _ h').
auto.
rewrite H.
constructor.
rewrite H in H22.
inversion H22.
auto.
inversion H2.
inversion H0.
inversion_clear H24.
inversion_clear H27.
inversion_clear H28.
inversion_clear H29.
inversion_clear H30.
inversion_clear H31.
inversion_clear H32.
inversion_clear H33.
replace l with x.
apply H29 with (codety G'0 T).
auto.
inversion H1.
inversion_clear H35.
inversion_clear H38.
inversion_clear H39.
inversion_clear H40.
inversion_clear H41.
inversion_clear H42.
inversion_clear H43.
inversion_clear H44.
apply eq_add_S.
replace (S x) with x0.
replace (S l) with x0.
auto.
apply (msize_eq _ _ h').
auto.
unfold unwrapMap in H45.
rewrite H in H45.
rewrite H.
constructor.
inversion H45.
auto.
apply (msize_eq _ _ h').
auto.
rewrite H.
constructor.
auto.


intros.  (* bump *)
inversion H3.
constructor.

intros. inversion H3. (* fold *)
constructor.
auto.
auto.

intros. inversion H3. (* ld *)
constructor.
auto.
auto.
intros. inversion H3. (* mov *)
constructor.
auto.
auto.
intros. inversion H3. (* movi *)
constructor.
auto.
intros. (* movl *)
inversion H3.
replace (Layout_aux h n) with (Layout_aux h' n).
constructor.
auto.
apply sym_eq.
apply tr_extend_layout_aux_preserv with l (tuple (makeUninitTup Ts)).
auto.
auto.
auto.
cut (msize _ _ h' (S l)).
intros.
rewrite H in H9.
inversion H9.
auto.
inversion H1.
inversion_clear H10.
inversion_clear H13.
inversion_clear H14.
inversion_clear H15.
inversion_clear H16.
inversion_clear H17.
inversion_clear H18.
inversion_clear H19.
replace (S l) with x.
auto.
apply sym_eq.
apply (msize_eq _ _ h').
rewrite H.
constructor.
unfold unwrapMap in H20.
rewrite H in H20.
inversion H20.
auto.
auto.

inversion H2.
inversion H0.
inversion_clear H20.
inversion_clear H23.
inversion_clear H24.
inversion_clear H25.
inversion_clear H26.
inversion_clear H27.
inversion_clear H28.
inversion_clear H29.
replace l with x.
apply H25 with t.
auto.
clear H30.
inversion H1.
inversion_clear H30.
inversion_clear H33.
inversion_clear H34.
inversion_clear H35.
inversion_clear H36.
inversion_clear H37.
inversion_clear H38.
inversion_clear H39.
apply eq_add_S.
replace (S x) with x0.
replace (S l) with x0.
auto.
apply sym_eq.
apply (msize_eq _ _ h').
unfold unwrapMap in H40.
rewrite H in H40.
rewrite H.
constructor.
inversion H40.
auto.
auto.
apply (msize_eq _ _ h').
auto.
rewrite H.
constructor.
auto.

intros.
inversion H3.
constructor.
auto.
auto.

intros. inversion H3. (* unfold *)
constructor.
auto.
auto.
Qed.



Lemma tr_extend_instrseq_preserv :
 forall (i : InstrSeq) (M : nat -> nat) (l : nat) (Ts : list Omega)
   (h h' : Map nat HeapVal) (wf : mWF _ _ h) (wf' : mWF _ _ h')
   (m : Map Reg Omega) (a : APTy) (D : nat) (G G' : HeapTy),
 h' = melem _ _ l (tuple (makeUninitTup Ts)) h ->
 WFHeap (wfmap _ _ h wf) G ->
 WFHeap (wfmap _ _ h' wf') G' ->
 WFInstrSeq G m a i ->
 TrInstrSeq (Layout_aux h) i M D -> TrInstrSeq (Layout_aux h') i M D.
Proof.
simple induction i.
intros.
inversion H4.
apply tris_iseq with _i.
apply (tr_extend_instr_preserv i0 i1 m a _i l Ts h h' wf wf' G G').
auto.
auto.
auto.
auto.
auto.
auto.
inversion H3.
apply (H M l Ts h h' wf wf' G'0 a (S D) G G').
auto.
auto.
auto.
auto.
auto.
auto.
apply (H M l Ts h h' wf wf' G'0 a (S D) G G').
auto.
auto.
auto.
auto.
auto.
apply (H M l Ts h h' wf wf' G'0 (used n) (S D) G G').
auto.
auto.
auto.
auto.
auto.
auto.
apply (H M l Ts h h' wf wf' m a (S D) G G').
auto.
auto.
auto.
auto.
auto.
apply (H M l Ts h h' wf wf' m fresh (S D) G G').
auto.
auto.
auto.
auto.
auto.
apply (H M l Ts h h' wf wf' G'0 a (S D) G G').
auto.
auto.
auto.
auto.
auto.
apply (H M l Ts h h' wf wf' G'0 a (S D) G G'). 
auto.
auto.
auto.
auto.
auto.
apply (H M l Ts h h' wf wf' G'0 a (S D) G G'). (* ld *)
auto.
auto.
auto.
auto.
auto.
apply (H M l Ts h h' wf wf' G'0 a (S D) G G').
auto.
auto.
auto.
auto.
auto.
apply (H M l Ts h h' wf wf' G'0 a (S D) G G').
auto.
auto.
auto.
auto.
auto.
apply (H M l Ts h h' wf wf' G'0 a (S D) G G'). 
auto.
auto.
auto.
auto.
auto.
apply (H M l Ts h h' wf wf' G'0 a (S D) G G'). 
auto.
auto.
auto.
auto.
auto.

intros. 
constructor.
inversion H3.
replace (Layout_aux h' n) with (Layout_aux h n).
auto.
apply tr_extend_layout_aux_preserv with l (tuple (makeUninitTup Ts)).
auto.
auto.
auto.
cut (mWF nat HeapVal (melem nat HeapVal l (tuple (makeUninitTup Ts)) h)).
intros.
apply (heap_label_eq_size l (tuple (makeUninitTup Ts)) h wf H9 G G').
auto.
constructor.
inversion H1.
unfold unwrapMap in |- * .
rewrite <- H.
unfold unwrapMap in H11.
auto.
apply eq_map_wf with h'.
auto.
auto.
inversion H2.
inversion H0.
inversion_clear H16.
inversion_clear H19.
inversion_clear H20.
inversion_clear H21.
inversion_clear H22.
replace l with x.
apply H21 with (codety G'0 a).
auto.
unfold unwrapMap in H16.
apply (msize_eq _ _ h x l).
auto.
cut (mWF nat HeapVal (melem nat HeapVal l (tuple (makeUninitTup Ts)) h)).
intros.
apply heap_label_eq_size with (tuple (makeUninitTup Ts)) wf H22 G G'.
auto.
constructor.
inversion H1.
unfold unwrapMap in |- * .
rewrite <- H.
unfold unwrapMap in H25.
auto.
apply eq_map_wf with h'.
auto.
auto.

intros. inversion H3. (* jmp *)
apply tris_jmp with _r.
auto.
auto.
Qed.



Lemma wordval_strip_not_wfold :
 forall v v' : WordVal,
 v' = stripWV v -> ~ (exists w : _, (exists t : _, v' = wfold w t)).
Proof.
simple induction v.
intros.
intro.
inversion_clear H0.
inversion_clear H1.
rewrite H0 in H.
discriminate H.
intros.
intro.
inversion_clear H0.
inversion_clear H1.
rewrite H0 in H.
discriminate H.
intros.
intro.
inversion_clear H0.
inversion_clear H1.
rewrite H0 in H.
discriminate H.
intros.
unfold stripWV in H0.
fold stripWV in H0.
apply H.
auto.
Qed.



Lemma tr_wordval_strip :
 forall (L : nat -> nat) (v : WordVal) (w : nat),
 TrWordVal L v w -> exists v' : _, v' = stripWV v /\ TrWordVal L v' w.
Proof.
simple induction v.
intros.
split with (wl n).
auto.
intros.
split with (wi n).
auto.
intros.
split with (wuninit o).
auto.
intros.
inversion H0.
cut (exists v'' : _, v'' = stripWV w /\ TrWordVal L v'' w0).
intros.
inversion_clear H6.
split with x.
auto.
apply H.
auto.
Qed.


Lemma tr_wordval_strip_alt :
 forall (L : nat -> nat) (v : WordVal) (w : nat),
 TrWordVal L v w -> TrWordVal L (stripWV v) w.
Proof.
intros.
cut (exists v' : _, v' = stripWV v /\ TrWordVal L v' w).
intros.
inversion_clear H0.
inversion_clear H1.
rewrite <- H0.
auto.
apply tr_wordval_strip.
auto.
Qed.


Lemma tr_strip_wordval :
 forall (L : nat -> nat) (v v' : WordVal) (w : nat),
 v' = stripWV v -> TrWordVal L v' w -> TrWordVal L v w.
Proof.
simple induction v.
intros.
rewrite H in H0.
auto.
intros.
rewrite H in H0.
auto.
intros.
rewrite H in H0.
auto.
intros.
constructor.
apply H with v'.
auto.
auto.
Qed.



Lemma tr_extend_wordval_preserv_aux :
 forall (w l : nat) (Ts : list Omega) (h h' : Map nat HeapVal)
   (wf : mWF _ _ h) (wf' : mWF _ _ h') (wv' wv : WordVal) 
   (HT HT' : HeapTy),
 h' = melem _ _ l (tuple (makeUninitTup Ts)) h ->
 WFHeap (wfmap _ _ h wf) HT ->
 WFHeap (wfmap _ _ h' wf') HT' ->
 wv' = stripWV wv ->
 (forall l n : nat,
  stripWV wv = wl l -> msize _ _ (unwrapMap nat Omega HT) n -> l < n) ->
 TrWordVal (Layout_aux h) wv' w -> TrWordVal (Layout_aux h') wv' w.
Proof.
simple induction wv'.
intros.
inversion H4.
replace (Layout_aux h n) with (Layout_aux h' n).
constructor.
apply sym_eq.
apply tr_extend_layout_aux_preserv with l (tuple (makeUninitTup Ts)).
auto.
auto.
auto.
inversion H1.
inversion_clear H9.
inversion_clear H12.
inversion_clear H13.
inversion_clear H14.
inversion_clear H15.
inversion_clear H16.
inversion_clear H17.
inversion_clear H18.
unfold unwrapMap in H19.
rewrite H in H19.
inversion H19.
auto.
apply H3.
auto.
induction HT as (m, m0).
apply wfh_ht_size_eq with h wf m0.
auto.
inversion H1.
inversion_clear H9.
inversion_clear H12.
inversion_clear H13.
inversion_clear H14.
inversion_clear H15.
inversion_clear H16.
inversion_clear H17.
inversion_clear H18.
unfold unwrapMap in H19.
rewrite H in H19.
inversion H19.
auto.
intros.
inversion H4.
constructor.
intros.
constructor.
intros.
absurd (exists w : _, (exists t : _, stripWV wv = wfold w t)).
apply wordval_strip_not_wfold with wv.
auto.
split with w0.
split with o.
auto.
Qed.


(*    (WFRegFile HT R G) ->
    (mlookup ? ? R r wv) -> *)



Lemma tr_extend_wordvali_preserv :
 forall (w l : nat) (Ts : list Omega) (h h' : Map nat HeapVal)
   (wf : mWF _ _ h) (wf' : mWF _ _ h') (wv : WordVal) 
   (t : Omega) (i : bool) (G G' : HeapTy),
 h' = melem _ _ l (tuple (makeUninitTup Ts)) h ->
 WFHeap (wfmap _ _ h wf) G ->
 WFHeap (wfmap _ _ h' wf') G' ->
 WFWordValinit G wv t i ->
 TrWordVal (Layout_aux h) wv w -> TrWordVal (Layout_aux h') wv w.
Proof.
simple induction wv.
intros.
inversion H3.
replace (Layout_aux h n) with (Layout_aux h' n).
constructor.
apply sym_eq.
apply tr_extend_layout_aux_preserv with l (tuple (makeUninitTup Ts)).
auto.
auto.
auto.
cut (mWF nat HeapVal (melem nat HeapVal l (tuple (makeUninitTup Ts)) h)).
intros wf''.
apply heap_label_eq_size with (tuple (makeUninitTup Ts)) wf wf'' G G'.
auto.
constructor.
inversion H1.
inversion_clear H8.
split with x.
unfold unwrapMap in |- * .
rewrite <- H.
unfold unwrapMap in H11.
auto.
apply eq_map_wf with h'.
auto.
auto.
inversion H2.
inversion H7.
inversion H0.
inversion_clear H18.
inversion_clear H21.
inversion_clear H22.
inversion_clear H23.
inversion_clear H24.
replace l with x.
apply H23 with t'.
auto.
unfold unwrapMap in H13.
cut (msize _ _ h l).
intros.
apply (msize_eq _ _ h x l).
auto.
auto.
cut (mWF nat HeapVal (melem nat HeapVal l (tuple (makeUninitTup Ts)) h)).
intros.
apply heap_label_eq_size with (tuple (makeUninitTup Ts)) wf H24 G G'.
auto.
inversion H1.
constructor.
unfold unwrapMap in |- * .
unfold unwrapMap in H27.
rewrite <- H.
auto.
apply eq_map_wf with h'.
auto.
auto.
intros.
inversion H3.
constructor.
intros.
constructor.

intros.
constructor.
inversion H4.
inversion H3.
inversion H10.
apply H with t' i G G'.
auto.
auto.
auto.
constructor.
auto.
auto.
Qed.



Lemma tr_extend_wordvallist_preserv :
 forall (L : list WordVal) (tl : list Omega) (il : list bool)
   (M : nat -> nat) (l : nat) (Ts : list Omega) (h h' : Map nat HeapVal)
   (wf : mWF _ _ h) (wf' : mWF _ _ h') (D : nat) (G G' : HeapTy),
 h' = melem _ _ l (tuple (makeUninitTup Ts)) h ->
 WFHeap (wfmap _ _ h wf) G ->
 WFHeap (wfmap _ _ h' wf') G' ->
 WFWordValinitList G L tl il ->
 TrWordValList (Layout_aux h) L M D -> TrWordValList (Layout_aux h') L M D.
Proof.
simple induction L.
intros.
constructor.
intros.
inversion H4.
apply trwvl_cons with w.
inversion H3.
apply (tr_extend_wordvali_preserv w l0 Ts h h' wf wf' a t i G G').
auto.
auto.
auto.
auto.
auto.
auto.
inversion H3.
apply (H tl0 il0 M l0 Ts h h' wf wf' (S D) G G').
auto.
auto.
auto.
auto.
auto.
Qed.



Lemma tr_extend_heapval_preserv_aux :
 forall (M : nat -> nat) (l : nat) (Ts : list Omega) 
   (h h' : Map nat HeapVal) (wf : mWF _ _ h) (wf' : mWF _ _ h') 
   (n : nat) (hv : HeapVal) (t : Omega) (G G' : HeapTy),
 h' = melem _ _ l (tuple (makeUninitTup Ts)) h ->
 WFHeap (wfmap _ _ h wf) G ->
 WFHeap (wfmap _ _ h' wf') G' ->
 WFHeapVal G hv t ->
 n < l ->
 TrHeapVal (Layout_aux h) hv M (Layout_aux h n) ->
 TrHeapVal (Layout_aux h') hv M (Layout_aux h' n).
Proof.
simple induction hv.
intros.
inversion H4.
constructor.
inversion H2.
replace (Layout_aux h' n) with (Layout_aux h n).
apply
 (tr_extend_wordvallist_preserv l0 tl il M l Ts h h' wf wf' 
    (Layout_aux h n) G G').
auto. auto. auto. auto. auto. auto.
apply tr_extend_layout_aux_preserv with l (tuple (makeUninitTup Ts)).
auto. auto. auto.
cut (mWF nat HeapVal (melem nat HeapVal l (tuple (makeUninitTup Ts)) h)).
intros wf''.
apply heap_label_eq_size with (tuple (makeUninitTup Ts)) wf wf'' G G'.
auto.
inversion H1.
constructor.
inversion_clear H15.
split with x.
unfold unwrapMap in |- * .
unfold unwrapMap in H18.
rewrite <- H. auto.
apply eq_map_wf with h'.
auto. auto. auto.
intros.
inversion H4.
constructor.
apply
 (tr_extend_instrseq_preserv i M l Ts h h' wf wf' m a (Layout_aux h' n) G G').
auto. auto. auto. inversion H2. auto. auto. 
replace (Layout_aux h' n) with (Layout_aux h n).
auto.
apply tr_extend_layout_aux_preserv with l (tuple (makeUninitTup Ts)).
auto. auto. auto.
cut (mWF nat HeapVal (melem nat HeapVal l (tuple (makeUninitTup Ts)) h)).
intros wf''.
apply heap_label_eq_size with (tuple (makeUninitTup Ts)) wf wf'' G G'.
auto.
constructor.
inversion H1.
inversion_clear H13.
split with x.
unfold unwrapMap in |- * .
rewrite <- H.
unfold unwrapMap in H16.
auto.
apply eq_map_wf with h'.
auto.
auto.
auto.
Qed.



Lemma tr_extend_heapval_preserv :
 forall (M : nat -> nat) (l : nat) (Ts : list Omega) 
   (H H' : Heap) (G G' : HeapTy) (n : nat) (hv : HeapVal) 
   (t : Omega),
 WFHeap H G ->
 WFHeap H' G' ->
 mextend nat HeapVal (unwrapMap nat HeapVal H) l (tuple (makeUninitTup Ts))
   (unwrapMap nat HeapVal H') ->
 WFHeapVal G hv t ->
 n < l ->
 TrHeapVal (Layout H) hv M (Layout H n) ->
 TrHeapVal (Layout H') hv M (Layout H' n).
Proof.
intros.
induction H as (m, m0).
induction H' as (m1, m2).
unfold Layout, unwrapMap in |- * .
apply (tr_extend_heapval_preserv_aux M l Ts m m1 m0 m2 n hv t G G'). auto.
inversion H2.
unfold unwrapMap in H.
auto. auto. auto. auto. auto. auto.
Qed.




Lemma tr_heap_extend_emptytup_aux :
 forall (M : nat -> nat) (l : nat) (Ts : list Omega) 
   (h h' : Map nat HeapVal) (wf : mWF _ _ h) (wf' : mWF _ _ h')
   (wf'' : mWF _ _ (melem _ _ l (tuple (makeUninitTup Ts)) h))
   (G G' : HeapTy),
 h' = melem _ _ l (tuple (makeUninitTup Ts)) h ->
 WFHeap (wfmap _ _ h wf) G ->
 WFHeap (wfmap _ _ h' wf') G' ->
 TrHeap (Layout (wfmap _ _ h wf)) (wfmap _ _ h wf) M ->
 TrHeap (Layout (wfmap _ _ (melem _ _ l (tuple (makeUninitTup Ts)) h) wf''))
   (wfmap _ _ (melem _ _ l (tuple (makeUninitTup Ts)) h) wf'') M ->
 TrHeap (Layout (wfmap _ _ h' wf')) (wfmap _ _ h' wf') M.
Proof.
intros.
unfold Layout, unwrapMap in H3.
inversion H3.
apply trheap with s.
unfold unwrapMap in |- * .
unfold unwrapMap in H5.
rewrite H.
auto.
unfold unwrapMap in |- * .
unfold Layout, unwrapMap in |- * .
rewrite H.
auto.
Qed.



Lemma tr_heap_extend_emptytup :
 forall (M : nat -> nat) (l : nat) (Ts : list Omega) 
   (h : Map nat HeapVal) (wf : mWF _ _ h)
   (wf' : mWF _ _ (melem _ _ l (tuple (makeUninitTup Ts)) h)) 
   (G G' : HeapTy),
 WFHeap (wfmap _ _ h wf) G ->
 WFHeap (wfmap _ _ (melem _ _ l (tuple (makeUninitTup Ts)) h) wf') G' ->
 TrHeap (Layout (wfmap _ _ h wf)) (wfmap _ _ h wf) M ->
 TrHeap (Layout (wfmap _ _ (melem _ _ l (tuple (makeUninitTup Ts)) h) wf'))
   (wfmap _ _ (melem _ _ l (tuple (makeUninitTup Ts)) h) wf') M.
Proof.
intros.
inversion H0.
unfold unwrapMap in H3.
inversion_clear H3.
apply trheap with x.
inversion_clear H6.
auto.
unfold unwrapMap in |- * .
intros.
cut (n < x).
intros.
inversion_clear H6.
inversion H8.
clear H12.
clear m.
clear H11.
clear H6.
clear a.
clear b.
move H7 after H13.
rewrite <- H10 in H7.
cut (n = n0 \/ n < n0).
intros.
inversion_clear H6.
cut (n0 = l).
intros.
rewrite H6 in H11.
replace h0 with (tuple (makeUninitTup Ts)).
constructor.
apply tr_extend_heap_uninitwordlist.
rewrite H11 in H3.
inversion H3.
auto.
absurd (l <> l).
auto.
auto.
cut (msize _ _ h l).
intros.
apply (msize_eq _ _ h n0 l).
auto.
auto.
apply heap_label_eq_size with (tuple (makeUninitTup Ts)) wf wf' G G'.
auto.
auto.
cut (exists t : _, WFHeapVal G h0 t).
intros.
elim H6; clear H6; intros t; intros.
apply tr_extend_heapval_preserv with n0 Ts (wfmap nat HeapVal h wf) G G' t.
auto.
auto.
unfold unwrapMap in |- * .
replace n0 with l.
constructor.
inversion wf'.
auto.
cut (msize _ _ h l).
intros.
apply (msize_eq _ _ h l n0).
auto.
auto.
apply (heap_label_eq_size l (tuple (makeUninitTup Ts)) h wf wf' G G').
auto.
auto.
auto.
auto.
inversion H1.
apply H15.
inversion H3.
cut (l = n0).
intros.
rewrite H24 in H19.
absurd (n0 = n).
cut (n <> n0); intros; auto.
apply lt_neq.
auto.
auto.
cut (msize _ _ h l).
intros.
apply (msize_eq _ _ h l n0).
auto.
auto.
apply (heap_label_eq_size l (tuple (makeUninitTup Ts)) h wf wf' G G').
auto.
auto.
auto.
cut (exists t : _, mlookup nat Omega (unwrapMap nat Omega G) n t).
intros.
elim H6; clear H6; intros t H6.
split with t.
inversion H.
inversion_clear H14.
inversion_clear H17.
inversion_clear H18.
inversion_clear H19.
inversion_clear H20.
inversion_clear H21.
inversion_clear H22.
inversion_clear H23.
apply H22 with n.
unfold unwrapMap in |- * .
inversion H3.
absurd (n = l).
apply lt_neq.
replace l with n0.
auto.
cut (msize _ _ h l).
intros.
apply (msize_eq _ _ h n0 l); auto.
apply (heap_label_eq_size l (tuple (makeUninitTup Ts)) h wf wf' G G').
auto.
auto.
auto.
auto.
auto.
inversion H.
inversion_clear H12.
inversion_clear H16.
inversion_clear H17.
inversion_clear H18.
inversion_clear H19.
inversion_clear H20.
inversion_clear H21.
apply H20.
replace x0 with n0.
auto.
unfold unwrapMap in H12.
apply (msize_eq _ _ h n0 x0).
auto.
auto.
cut (n0 = n \/ n < n0).
intros.
inversion_clear H6.
left; auto.
right; auto.
apply lt_S_eq_or_lt.
auto.
inversion_clear H6.
inversion_clear H8.
inversion_clear H9.
apply H8 with h0.
auto.
Qed.



Lemma heaplookup_wfheapval :
 forall (H : Heap) (HT : HeapTy) (l : nat) (h : HeapVal),
 WFHeap H HT ->
 mlookup nat HeapVal (unwrapMap nat HeapVal H) l h ->
 exists t : _, WFHeapVal HT h t.
Proof.
intro.
elim H; clear H; intros h wfh.
intro.
elim HT; clear HT; intros ht wfht.
intros.
inversion H.
unfold unwrapMap in H2.
cut (exists t : _, mlookup _ _ ht l t).
intros.
inversion_clear H5.
split with x.
inversion_clear H2.
inversion_clear H5.
inversion_clear H7.
inversion_clear H8.
inversion_clear H9.
inversion_clear H10.
inversion_clear H11.
inversion_clear H12.
apply H11 with l.
auto.
auto.
inversion_clear H2.
inversion_clear H5.
inversion_clear H6.
inversion_clear H7.
inversion_clear H8.
inversion_clear H9.
inversion_clear H10.
apply H9.
apply H6 with h0.
auto.
Qed.





Lemma Preservation_alloc :
 forall (S : (nat -> nat) * ((_Reg -> nat) * nat)) 
   (H : Heap) (R : RegFile) (a : AP) (rd : Reg) (Ts : list Omega)
   (I : InstrSeq) (P P' : Program),
 P = (H, (R, (a, iseq (alloc rd Ts) I))) ->
 WFProgram P ->
 TrProgram P S -> Eval P P' -> WFProgram P' -> TrProgram P' (Step S).
Proof.
intros.
rewrite H0 in H1.
inversion H1.
clear H14.
rewrite H0 in H2.
inversion H2.
unfold Step in |- * .
cut (TrInstrSeq L (iseq (alloc rd Ts) I) M pc).
intros.
inversion H25.
rewrite <- H30 in H28.
inversion H28.
rewrite H0 in H3.
inversion H3.
apply trprogram with (Layout H').
auto.
rewrite <- H44 in H4.
inversion H4.
clear H60.
cut (mWF nat HeapVal (melem nat HeapVal a (tuple (makeUninitTup Ts)) h)).
intros.
rewrite H48.
apply (tr_heap_extend_emptytup_aux M a Ts h h' wfh wfh' H60 HT HT0).
inversion H50.
auto.
rewrite <- H47.
auto.
rewrite <- H48.
auto.
rewrite <- H47.
rewrite <- H19.
auto.
apply (tr_heap_extend_emptytup M a Ts h wfh H60 HT HT0).
rewrite <- H47.
auto.
inversion H50.
inversion H56.
constructor.
inversion_clear H67.
split with x.
unfold unwrapMap in |- * .
rewrite H48 in H70.
unfold unwrapMap in H70.
rewrite <- H65 in H70.
auto.
rewrite <- H47.
rewrite <- H19.
auto.
apply eq_map_wf with h'.
auto.
inversion H50.
auto.
unfold updatereg in |- * .
constructor.
intros.
cut (_r = _rd \/ _r <> _rd).
intros.
inversion_clear H53.
rewrite H54.
replace (beq_reg _rd _rd) with true.
cut (v = wl a).
intros.
replace (_R _r31) with (Layout H' a).
replace (Layout H' a + 0) with (Layout H' a).
rewrite H53.
constructor.
apply sym_eq.
apply plus_n_O_n.
inversion H23.
rewrite H19 in H55.
replace (Layout H' a) with (Layout H a).
auto.
unfold Layout in |- * .
unfold unwrapMap in |- * .
rewrite H47.
rewrite H48.
apply
 (tr_extend_layout_aux_eq_preserv h h' wfh wfh' a (tuple (makeUninitTup Ts))).
inversion H50.
auto.
rewrite <- H44 in H4.
inversion H4.
clear H68.
cut (mWF nat HeapVal (melem nat HeapVal a (tuple (makeUninitTup Ts)) h)).
intros.
apply heap_label_eq_size with (tuple (makeUninitTup Ts)) wfh H68 HT HT0.
rewrite <- H47.
auto.
inversion H64.
rewrite H48 in H70.
unfold unwrapMap in H70.
inversion H50.
rewrite <- H77 in H70.
constructor.
auto.
apply eq_map_wf with h'.
auto.
inversion H50.
auto.
cut (r = rd).
intros.
cut (mlookup _ _ R' rd (wl a)).
intros.
apply (mlookup_eq _ _ rd v (wl a) R').
apply eqorneq_reg.
rewrite <- H53.
auto.
auto.
apply mupdext_imp_mlookup with R.
apply eqorneq_reg.
auto.
rewrite H54 in H52.
apply trregf_eq_eq with _rd.
auto.
auto.
apply sym_eq; auto.
replace (beq_reg _r _rd) with false; auto.
inversion H22.
rewrite H19 in H53.
cut (TrWordVal (Layout_aux h) v (_R _r)).
intros.
rewrite H48.
unfold Layout, unwrapMap in |- * .
apply tr_strip_wordval with (stripWV v).
auto.
rewrite <- H44 in H4.
inversion H4.
apply
 (tr_extend_wordval_preserv_aux (_R _r) a Ts h h' wfh wfh' 
    (stripWV v) v HT HT0).
inversion H50.
auto.
rewrite <- H47.
auto.
rewrite <- H48.
auto.
auto.
inversion H11.
intros.
apply H70 with r v.
apply mupdext_preserv_rev with R' rd (wl a).
auto.
auto.
apply trregf_neq_neq with _r _rd; auto.
auto.
auto.
auto.
auto.
auto.
apply tr_wordval_strip_alt.
auto.
unfold Layout, unwrapMap in H53.
rewrite H47 in H53.
apply H53 with r.
apply mupdext_preserv_rev with R' rd (wl a).
auto.
auto.
apply trregf_neq_neq with _r _rd.
auto.
auto.
auto.
auto.
apply eqorneq_tisreg.
constructor.
unfold updatereg in |- * .
replace (beq_reg _r31 _rd) with false.
inversion H23.
rewrite H19 in H51.
rewrite H47 in H51.
rewrite H48.
unfold Layout, unwrapMap in |- * .
unfold Layout, unwrapMap in H51.
replace (Layout_aux h' a) with (Layout_aux h a).
auto.
apply tr_extend_layout_aux_eq_preserv with (tuple (makeUninitTup Ts)).
auto.
auto.
inversion H50.
auto.
rewrite <- H44 in H4.
inversion H4.
clear H64.
cut (mWF nat HeapVal (melem nat HeapVal a (tuple (makeUninitTup Ts)) h)).
intros.
apply heap_label_eq_size with (tuple (makeUninitTup Ts)) wfh H64 HT HT0.
rewrite <- H47.
auto.
inversion H60.
rewrite H48 in H66.
unfold unwrapMap in H66.
inversion H50.
rewrite <- H73 in H66.
constructor.
auto.
apply eq_map_wf with h'.
auto.
inversion H50.
auto.
apply sym_eq.
apply beq_reg_neq_false.
cut (_rd <> _r31).
auto.
apply trreg_neq_31 with rd.
auto.
elim H24; clear H24; intros l1 H24.
elim H24; clear H24; intros GO H24.
elim H24; clear H24; intros T H24.
elim H24; clear H24; intros I'0 H24.
elim H24; clear H24; intros n' H24.
inversion_clear H24.
inversion_clear H52.
inversion_clear H53.
split with l1.
split with GO.
split with T.
split with I'0.
rewrite <- H21 in H2.
clear H21.
clear S.
split with (S n').
split.
rewrite H48.
unfold unwrapMap in |- * .
apply mextend_preserv with h a (tuple (makeUninitTup Ts)).
auto.
rewrite H47 in H51.
auto.
cut (a <> l1).
auto.
apply (mwf_anyelem_neq _ _ a l1 (tuple (makeUninitTup Ts)) h).
inversion H50.
apply eq_map_wf with h'.
auto.
auto.
apply mlookup_imp_mindom with (code GO T I'0).
rewrite H47 in H51.
auto.
split.
apply tr_instrseq_subdepth_S with (alloc rd Ts).
auto.
cut (TrInstrSeq (Layout H') I'0 M (Layout H' l1)).
intros.
split.
auto.
replace (Layout H' l1 + S n') with (S n' + Layout H' l1).
unfold plus in |- * .
fold plus in |- * .
apply eq_S.
replace (n' + Layout H' l1) with (Layout H' l1 + n').
replace (Layout H' l1) with (Layout H l1).
rewrite <- H19.
auto.
unfold Layout, unwrapMap in |- * .
rewrite H47.
rewrite H48.
apply tr_extend_layout_aux_preserv with a (tuple (makeUninitTup Ts)).
auto.
auto.
inversion H50.
auto.
rewrite <- H44 in H4.
inversion H4.
cut (mWF nat HeapVal (melem nat HeapVal a (tuple (makeUninitTup Ts)) h)).
intros.
apply (heap_label_eq_size a (tuple (makeUninitTup Ts)) h wfh H64 HT HT0).
rewrite <- H47.
auto.
rewrite H48 in H59.
inversion H59.
inversion H50.
unfold unwrapMap in H66.
rewrite <- H73 in H66.
constructor.
auto.
apply eq_map_wf with h'.
auto.
inversion H50.
auto.
inversion H10.
inversion_clear H55.
inversion_clear H58.
inversion_clear H59.
inversion_clear H60.
inversion_clear H61.
replace a with x.
apply H59 with (code GO T I'0).
auto.
inversion H13.
rewrite <- H67 in H12.
inversion H12.
apply (msize_eq _ _ (unwrapMap nat Omega HT)).
auto.
auto.
apply plus_comm.
apply plus_comm.
rewrite H19 in H52.
unfold Layout, unwrapMap in |- * .
unfold Layout, unwrapMap in H52.
rewrite H48.
rewrite H47 in H52.
cut (exists t : _, WFHeapVal HT (code GO T I'0) t).
intros.
inversion_clear H21.
inversion_clear H53.
rewrite <- H44 in H4.
inversion H4.
apply (tr_extend_instrseq_preserv I'0 M a Ts h h' wfh wfh' GO T) with HT HT0.
inversion H50.
auto.
rewrite <- H47.
auto.
rewrite <- H48.
auto.
auto.
replace (Layout_aux h' l1) with (Layout_aux h l1).
auto.
apply tr_extend_layout_aux_preserv with a (tuple (makeUninitTup Ts)).
auto.
auto.
inversion H50.
auto.
cut (mWF nat HeapVal (melem nat HeapVal a (tuple (makeUninitTup Ts)) h)).
intros.
apply (heap_label_eq_size a (tuple (makeUninitTup Ts)) h wfh H64 HT HT0).
rewrite <- H47.
auto.
rewrite H48 in H59.
inversion H59.
unfold unwrapMap in H66.
inversion H50.
rewrite <- H73 in H66.
constructor.
auto.
apply eq_map_wf with h'.
auto.
inversion H50.
auto.
inversion H10.
inversion_clear H65.
inversion_clear H68.
inversion_clear H69.
inversion_clear H70.
inversion_clear H71.
replace a with x0.
apply H69 with (code GO T I'0).
auto.
apply (msize_eq _ _ (unwrapMap nat Omega HT)).
auto.
inversion H13.
rewrite <- H77 in H12.
inversion H12.
auto.
apply heaplookup_wfheapval with (wfmap _ _ h wfh) l1.
rewrite <- H47.
auto.
rewrite <- H47.
auto.
apply tr_pc_trinstrseq with H.
auto.
Qed.




Lemma Preservation_bgt :
 forall (S : (nat -> nat) * ((_Reg -> nat) * nat)) 
   (H : Heap) (R : RegFile) (a : AP) (rs rt : Reg) 
   (l : nat) (I : InstrSeq) (P P' : Program),
 P = (H, (R, (a, iseq (bgt rs rt l) I))) ->
 WFProgram P ->
 TrProgram P S -> Eval P P' -> WFProgram P' -> TrProgram P' (Step S).
Proof.
intros St.
intros.
rewrite H0 in H1.
inversion H1.
clear H14.
rewrite H0 in H2.
inversion H2.
unfold Step in |- * .
cut (TrInstrSeq L (iseq (bgt rs rt l) I) M pc).
intros.
inversion H25.
rewrite H30.
inversion H28.
rewrite H0 in H3.
inversion H3.
replace (blt_nat (_R _rt) (_R _rs)) with false.
apply trprogram with L.
auto.
auto.
auto.
auto.
apply tr_pc_Spc with (bgt rs rt l).
auto.
apply sym_eq.
apply le_blt_false.
replace (_R _rs) with rsval.
replace (_R _rt) with rtval.
auto.
inversion H22.
cut (TrWordVal L (wi rtval) (_R _rt)).
intros.
inversion H57.
auto.
apply H53 with rt.
auto.
auto.
cut (TrWordVal L (wi rsval) (_R _rs)).
intros.
inversion H53.
auto.
inversion H22.
apply H53 with rs.
auto.
auto.
replace (blt_nat (_R _rt) (_R _rs)) with true.
apply trprogram with L.
auto.
auto.
auto.
auto.
split with l.
split with G0.
split with T.
split with I''.
split with 0.
split.
auto.
split.
constructor.
split.
inversion H20.
cut (TrHeapVal L (code G0 T I'') M (L l)).
intros.
inversion H60.
auto.
apply H56.
auto.
auto.
apply sym_eq.
apply blt_lt_true.
replace (_R _rt) with rtval.
replace (_R _rs) with rsval.
unfold lt in |- * .
apply gt_le_S.
auto.
cut (TrWordVal L (wi rsval) (_R _rs)).
intros.
inversion H54.
auto.
inversion H22.
apply H54 with rs.
auto.
auto.
cut (TrWordVal L (wi rtval) (_R _rt)).
intros.
inversion H54.
auto.
inversion H22.
apply H54 with rt.
auto.
auto.
apply tr_pc_trinstrseq with H.
auto.
Qed.





Lemma wf_wordlist_len_eq :
 forall (HT : HeapTy) (V : list WordVal) (tl : list Omega) (ol : list bool),
 WFWordValinitList HT V tl ol -> length V = length tl.
Proof.
simple induction V.
intros.
inversion H.
auto.
intros.
inversion H0.
unfold length in |- * .
apply eq_S.
apply H with il.
auto.
Qed.


Lemma wf_tuple_tupty_len_eq :
 forall (HT : HeapTy) (V : list WordVal) (tl : list Omega) (ol : list bool),
 WFHeapVal HT (tuple V) (tupty tl ol) -> length V = length tl.
Proof.
intros.
inversion H.
apply wf_wordlist_len_eq with HT ol.
auto.
Qed.


Lemma layout_gt_eq :
 forall (h : Map nat HeapVal) (wfh : mWF _ _ h) (a : nat),
 (forall b : nat, mindom _ _ h b -> b < a) ->
 Layout_aux h a = Layout_aux h (S a).
Proof.
simple induction h.
intros.
auto.
intros.
unfold Layout_aux in |- * .
replace (blt_nat a a0) with true.
replace (blt_nat a (S a0)) with true.
fold Layout_aux in |- * .
cut (Layout_aux m a0 = Layout_aux m (S a0)).
intros.
rewrite H1.
auto.
apply H.
inversion wfh.
auto.
intros.
apply H0.
apply minnext.
inversion wfh.
cut (b0 <> a).
intros.
auto.
apply mindom_mnotindom_neq with HeapVal m.
auto.
apply eqorneq_nat.
auto.
auto.
auto.
apply sym_eq.
apply blt_lt_true.
cut (a < a0).
intros.
apply lt_S.
auto.
apply H0.
constructor.
apply sym_eq.
apply blt_lt_true.
apply H0.
constructor.
Qed.


Lemma wfheap_next_mindom_imp_lt :
 forall (l a b : nat) (hv : HeapVal) (h H : Map nat HeapVal)
   (wfh : mWF _ _ H) (HT : HeapTy),
 WFHeap (wfmap _ _ H wfh) HT ->
 H = melem _ _ l hv h -> msize _ _ h a -> mindom _ _ h b -> b < a.
Proof.
intros.
inversion H0.
inversion_clear H5.
inversion_clear H8.
inversion_clear H9.
inversion_clear H10.
inversion_clear H11.
inversion_clear H12.
inversion_clear H13.
inversion_clear H14.
replace a with l.
cut (l = b \/ b < l).
intros.
inversion_clear H14.
absurd (l = b).
inversion wfh.
rewrite H1 in H14.
absurd (mempty nat HeapVal = melem nat HeapVal l hv h).
discriminate.
auto.
replace l with a0.
cut (b <> a0).
intros.
auto.
apply mindom_mnotindom_neq with HeapVal m.
auto.
apply eqorneq_nat.
replace m with h.
auto.
rewrite H1 in H18.
injection H18.
auto.
auto.
rewrite H1 in H18.
injection H18.
auto.
auto.
auto.
apply lt_S_eq_or_lt.
replace (S l) with x.
cut
 (exists hv : _,
    mlookup nat HeapVal (unwrapMap nat HeapVal (wfmap nat HeapVal H wfh)) b
      hv).
intros.
inversion_clear H14.
apply H9 with x0.
auto.
unfold unwrapMap in |- * .
apply mindom_imp_mlookup.
rewrite H1.
apply minnext.
cut (b <> l).
auto.
apply mindom_mnotindom_neq with HeapVal h.
inversion wfh.
rewrite H1 in H14.
absurd (mempty nat HeapVal = melem nat HeapVal l hv h).
discriminate.
auto.
replace h with m.
auto.
rewrite H1 in H17.
injection H17.
auto.
apply eqorneq_nat.
auto.
inversion wfh.
rewrite H1 in H14.
absurd (mempty nat HeapVal = melem nat HeapVal l hv h).
discriminate.
auto.
replace h with m.
replace l with a0.
auto.
rewrite H1 in H17.
injection H17.
auto.
rewrite H1 in H17.
injection H17.
auto.
auto.
apply (msize_eq _ _ H).
auto.
unfold unwrapMap in H15.
rewrite H1 in H15.
inversion H15.
rewrite H1.
constructor.
auto.
cut (S a = x).
intros.
cut (S l = x).
intros.
rewrite <- H14 in H16.
apply eq_add_S.
auto.
unfold unwrapMap in |- * .
auto.
apply (msize_eq _ _ H).
rewrite H1.
unfold unwrapMap in H15.
rewrite H1 in H15.
inversion H15.
constructor.
auto.
auto.
apply (msize_eq _ _ H).
rewrite H1.
constructor.
auto.
auto.
Qed.


Lemma wfh_h_size_eq :
 forall (h : Map nat HeapVal) (wfh : mWF _ _ h) (ht : Map nat Omega)
   (wfht : mWF _ _ ht) (l : nat),
 WFHeap (wfmap _ _ h wfh) (wfmap _ _ ht wfht) ->
 msize nat Omega (unwrapMap nat Omega (wfmap _ _ ht wfht)) l ->
 msize nat HeapVal (unwrapMap nat HeapVal (wfmap _ _ h wfh)) l.
Proof.
intros.
inversion H.
inversion_clear H2.
inversion_clear H5.
inversion_clear H6.
replace l with x.
unfold unwrapMap in H2.
auto.
unfold unwrapMap in H5.
apply (msize_eq nat Omega ht).
auto.
auto.
Qed.



Lemma bump_aux :
 forall (H : Heap) (HT : HeapTy) (a n w : nat) (tl : list Omega)
   (ol : list bool),
 WFHeap H HT ->
 msize nat Omega (unwrapMap nat Omega HT) (S a) ->
 Layout H a = w ->
 WFWordVal HT (wl a) (tupty tl ol) -> n = length tl -> Layout H (S a) = w + n.
Proof.
intros H.
elim H; clear H; intros H wfH; intros.
cut (msize _ _ H (S a)).
intros.
inversion H5.
unfold Layout, unwrapMap in |- * .
rewrite <- H7.
cut (a0 = a).
intros.
rewrite H9.
unfold Layout_aux in |- * .
replace (blt_nat a (S a)) with true.
fold Layout_aux in |- * .
replace (Layout_aux m (S a)) with (Layout_aux m a).
unfold Layout, unwrapMap in H2.
rewrite <- H7 in H2.
rewrite <- H2.
rewrite H9.
unfold Layout_aux in |- * .
replace (blt_nat a a) with false.
fold Layout_aux in |- * .
replace (HV_size b) with n.
auto.
inversion H3.
cut (exists ol' : _, t' = tupty tl ol').
intros.
elim H15; clear H15; intros ol'; intros.
rewrite H15 in H11.
cut (exists V : _, mlookup _ _ H a (tuple V)).
intros.
elim H16; clear H16; intros V H16.
replace b with (tuple V).
unfold HV_size in |- * .
cut (WFHeapVal HT (tuple V) (tupty tl ol')).
intros.
replace (length V) with (length tl).
auto.
apply sym_eq.
apply wf_tuple_tupty_len_eq with HT ol'.
auto.
inversion H0.
inversion_clear H18.
inversion_clear H21.
inversion_clear H22.
inversion_clear H23.
inversion_clear H24.
inversion_clear H25.
inversion_clear H26.
inversion_clear H27.
apply H26 with a.
auto.
auto.
rewrite <- H7 in H16.
rewrite H9 in H16.
inversion H16.
auto.
absurd (a <> a); auto.
inversion H0.
inversion_clear H17.
inversion_clear H20.
inversion_clear H21.
inversion_clear H22.
inversion_clear H23.
inversion_clear H24.
inversion_clear H25.
inversion_clear H26.
cut
 (exists h : _,
    mlookup nat HeapVal (unwrapMap nat HeapVal (wfmap _ _ H wfH)) a h).
intros.
inversion H26.
cut (WFHeapVal HT x0 (tupty tl ol')).
intros.
inversion H29.
split with wl.
rewrite H32.
auto.
apply H25 with a.
auto.
auto.
apply H23.
apply H22 with (tupty tl ol').
auto.
apply subtype_tupty with ol.
auto.
apply sym_eq.
apply blt_eq_false.
apply layout_gt_eq.
inversion wfH.
rewrite <- H10 in H7.
absurd (melem nat HeapVal a0 b m = mempty nat HeapVal).
discriminate.
auto.
replace m with m0.
auto.
rewrite <- H12 in H7.
injection H7.
auto.
intros.
apply (wfheap_next_mindom_imp_lt a0 a b0 b m H wfH HT).
auto.
auto.
auto.
auto.
apply sym_eq.
apply blt_lt_true.
apply lt_n_Sn.
apply eq_add_S.
inversion H0.
inversion_clear H10.
inversion_clear H13.
inversion_clear H14.
inversion_clear H15.
inversion_clear H16.
inversion_clear H17.
inversion_clear H18.
inversion_clear H19.
cut (S a0 = x).
intros.
rewrite H19.
cut (S a = x).
intros.
rewrite H21.
auto.
apply (msize_eq _ _ H).
rewrite <- H7.
constructor.
auto.
auto.
apply (msize_eq _ _ H).
unfold unwrapMap in H20.
rewrite <- H7 in H20.
rewrite <- H7.
constructor.
inversion H20.
auto.
auto.
change
  (msize nat HeapVal (unwrapMap nat HeapVal (wfmap nat HeapVal H wfH)) (S a))
 in |- * .
induction HT as (m, m0).
apply wfh_h_size_eq with m m0.
auto.
auto.
Qed.




Lemma Preservation_bump :
 forall (S : (nat -> nat) * ((_Reg -> nat) * nat)) 
   (H : Heap) (R : RegFile) (a : AP) (i : nat) (I : InstrSeq)
   (P P' : Program),
 P = (H, (R, (a, iseq (bump i) I))) ->
 WFProgram P ->
 TrProgram P S -> Eval P P' -> WFProgram P' -> TrProgram P' (Step S).
Proof.
intros St.
intros.
rewrite H0 in H1.
inversion H1.
clear H14.
rewrite H0 in H2.
inversion H2.
unfold Step in |- * .

cut (TrInstrSeq L (iseq (bump i) I) M pc). intros.
inversion H25.
rewrite <- H30 in H28.
inversion H28.
clear H24.

unfold updatereg in |- * .
rewrite H0 in H3.
inversion H3.
apply trprogram with L.
auto.
auto.
constructor.
intros.
cut (_r = _r31 \/ _r <> _r31).
intros.
inversion_clear H46.
rewrite H47 in H45.
absurd (_r31 <> _r31).
auto.
apply trreg_neq_31 with r.
auto.
replace (beq_reg _r _r31) with false.
inversion H22.
apply H46 with r.
auto.
auto.
apply sym_eq.
apply beq_reg_neq_false.
auto.
apply eqorneq_tisreg.
constructor.
replace (beq_reg _r31 _r31) with true.
rewrite H19.
inversion H23.
inversion H13.
rewrite <- H51 in H12.
inversion H12.
rewrite <- H48.
replace l1 with (S a).
apply sym_eq.
apply bump_aux with HT tl ol; auto.
rewrite <- H19; auto.
inversion H10.
inversion_clear H61.
inversion_clear H64.
inversion_clear H65.
clear H66.
induction H as (m, m0).
induction HT as (m1, m2).
replace (S a) with x.
replace l1 with x; auto.
apply (msize_eq _ _ m x l1); auto.
apply (msize_eq _ _ m1 x (S a)); auto.
apply sym_eq; auto.
rewrite <- H21 in H2.
inversion H2.
rewrite H19.
rewrite <- H51.
apply tr_pc_Spc with (bump i); auto.
apply tr_pc_trinstrseq with H; auto.
Qed.




Lemma Preservation_fold :
 forall (S : (nat -> nat) * ((_Reg -> nat) * nat)) 
   (H : Heap) (R : RegFile) (a : AP) (rd rs : Reg) 
   (t : Omega) (I : InstrSeq) (P P' : Program),
 P = (H, (R, (a, iseq (fold rd t rs) I))) ->
 WFProgram P ->
 TrProgram P S -> Eval P P' -> WFProgram P' -> TrProgram P' (Step S).
Proof.
intros St.
intros.
rewrite H0 in H1.
inversion H1.
clear H14.
rewrite H0 in H2.
inversion H2.
unfold Step in |- * .
cut (TrInstrSeq L (iseq (fold rd t rs) I) M pc).
intros.
inversion H25.
rewrite H30.
inversion H28.
rewrite H0 in H3.
inversion H3.
apply trprogram with L.
auto.
auto.
unfold updatereg in |- * .
constructor.
intros.
compare _r _rd.
intros.
replace (beq_reg _r _rd) with true.
replace (_R _rs + 0) with (0 + _R _rs).
unfold plus in |- * .
replace v with (wfold rsval t).
constructor.
inversion H22.
apply H54 with rs.
auto.
auto.
apply (mlookup_eq _ _ r (wfold rsval t) v R').
apply eqorneq_reg.
apply mupdext_imp_mlookup with R.
apply eqorneq_reg.
replace r with rd.
auto.
apply trregf_eq_eq with _r.
rewrite e.
auto.
auto.
auto.
apply plus_comm.
rewrite e.
apply sym_eq; auto.
intros.
replace (beq_reg _r _rd) with false; auto.
inversion H22.
apply H54 with r; auto.
apply mupdext_preserv_rev with R' rd (wfold rsval t); auto.
apply trregf_neq_neq with _r _rd; auto.
constructor.
unfold updatereg in |- * .
replace (beq_reg _r31 _rd) with false.
inversion H23.
auto.
apply sym_eq.
apply beq_reg_neq_false; auto.
cut (_rd <> _r31); auto.
apply trreg_neq_31 with rd; auto.
apply tr_pc_Spc with (fold rd t rs); 
auto.
apply tr_pc_trinstrseq with H;
auto.
Qed.





Lemma Preservation_jmp :
 forall (St : (nat -> nat) * ((_Reg -> nat) * nat)) 
   (H : Heap) (R : RegFile) (a : AP) (r : Reg) (P P' : Program),
 P = (H, (R, (a, jmp r))) ->
 WFProgram P ->
 TrProgram P St -> Eval P P' -> WFProgram P' -> TrProgram P' (Step St).
Proof.
intros.
rewrite H0 in H1.
inversion H1.
clear H14.
rewrite H0 in H2.
inversion H2.
unfold Step in |- * .
cut (TrInstrSeq L (jmp r) M pc).
intros.
inversion H25.
rewrite H29.
rewrite H0 in H3.
inversion H3.
apply trprogram with L.
auto.
auto.
auto.
auto.
split with l0.
split with G0.
split with T.
split with I'.
split with 0.
split.
auto.
split.
constructor.
split.
inversion H20.
cut (TrHeapVal L (code G0 T I') M (L l0)).
intros.
inversion H46.
auto.
apply H42.
auto.
replace (L l0 + 0) with (0 + L l0).
unfold plus in |- * .
inversion H22.
cut (TrWordVal L (wl l0) (_R _r)).
intros.
inversion H44.
auto.
apply H40 with r.
auto.
auto.
apply plus_comm.
apply tr_pc_trinstrseq with H.
auto.
Qed.



Lemma Preservation_jd :
 forall (St : (nat -> nat) * ((_Reg -> nat) * nat)) 
   (H : Heap) (R : RegFile) (a : AP) (l : nat) (P P' : Program),
 P = (H, (R, (a, jd l))) ->
 WFProgram P ->
 TrProgram P St -> Eval P P' -> WFProgram P' -> TrProgram P' (Step St).
Proof.
intros.
rewrite H0 in H1.
inversion H1.
clear H14.
rewrite H0 in H2.
inversion H2.
unfold Step in |- * .
cut (TrInstrSeq L (jd l) M pc).
intros.
inversion H25.
rewrite H28.
rewrite H0 in H3.
inversion H3.
apply trprogram with L.
auto.
auto.
auto.
auto.
split with l.
split with G0.
split with T.
split with I'.
split with 0.
split.
auto.
split.
constructor.
split.
inversion H20.
cut (TrHeapVal L (code G0 T I') M (L l)).
intros.
inversion H44.
auto.
apply H40.
auto.
replace (L l + 0) with (0 + L l).
unfold plus in |- * .
auto.
apply plus_comm.

inversion_clear H24.
inversion_clear H25.
inversion_clear H24.
inversion_clear H25.
inversion_clear H24.
inversion_clear H25.
inversion_clear H26.
inversion_clear H27.
rewrite <- H28.
replace (L x + x3) with (x3 + L x).
apply tr_sub_instrseq with x2.
auto.
auto.
apply plus_comm.
Qed.




Lemma trwordvallist_i_wordval :
 forall (L M : nat -> nat) (V : list WordVal) (i : nat) 
   (v : WordVal) (w : nat),
 TrWordValList L V M w -> ListNth _ V i v -> TrWordVal L v (M (w + i)).
Proof.
simple induction V.
intros.
inversion H0.
clear V.
intros v V.
simple induction i.
intros.
inversion H1.
inversion H0.
replace (M (w + 0)) with w0.
rewrite <- H2.
auto.
replace (w + 0) with (0 + w).
auto.
apply plus_comm.
intros.
replace (w + S n) with (S w + n).
apply H.
inversion H1.
auto.
inversion H2.
auto.
replace (w + S n) with (S n + w).
unfold plus in |- * .
fold plus in |- * .
replace (n + w) with (w + n).
auto.
apply plus_comm.
apply plus_comm.
Qed.



Lemma Preservation_ld :
 forall (S : (nat -> nat) * ((_Reg -> nat) * nat)) 
   (H : Heap) (R : RegFile) (a : AP) (rd rs : Reg) 
   (i : nat) (I : InstrSeq) (P P' : Program),
 P = (H, (R, (a, iseq (ld rd rs i) I))) ->
 WFProgram P ->
 TrProgram P S -> Eval P P' -> WFProgram P' -> TrProgram P' (Step S).
Proof.
intros St.
intros.
rewrite H0 in H1.
inversion H1.
clear H14.
rewrite H0 in H2.
inversion H2.
unfold Step in |- * .
cut (TrInstrSeq L (iseq (ld rd rs i) I) M pc). intros.
inversion H25.
rewrite H30.
inversion H28.
rewrite H0 in H3.
inversion H3.
apply trprogram with L;
auto.
unfold updatereg in |- * .
constructor.
intros.
cut (_r = _rd \/ _r <> _rd); 
intros.
inversion_clear H56.
replace (beq_reg _r _rd) with true; auto.
replace v0 with v; auto.
cut (TrHeapVal L (tuple V) M (L l1));
intros.
inversion H56.
replace (_R _rs) with (L l1).
apply trwordvallist_i_wordval with V;
auto.
inversion H22.
cut (TrWordVal L (wl l1) (_R _rs));
intros.
inversion H67; 
auto.
apply H63 with rs;
auto.
inversion H20.
apply H59;
auto.
apply (mlookup_eq _ _ r v v0 R'); auto.
apply eqorneq_reg.
apply mupdext_imp_mlookup with R.
apply eqorneq_reg.
replace r with rd;
auto.
apply trregf_eq_eq with _r; auto.
rewrite H57;
auto.
replace (beq_reg _r _rd) with false; auto.
inversion H22.
apply H56 with r.
apply mupdext_preserv_rev with R' rd v; auto.
apply trregf_neq_neq with _r _rd;
auto.
auto.
auto.
unfold updatereg in |- * .
constructor.
replace (beq_reg _r31 _rd) with false.
inversion H23;
auto.
apply sym_eq.
apply beq_reg_neq_false.
cut (_rd <> _r31);
auto.
apply trreg_neq_31 with rd;
auto.
apply tr_pc_Spc with (ld rd rs i); auto.
apply tr_pc_trinstrseq with H; auto.
Qed.



Lemma Preservation_mov :
 forall (S : (nat -> nat) * ((_Reg -> nat) * nat)) 
   (H : Heap) (R : RegFile) (a : AP) (rd rs : Reg) 
   (I : InstrSeq) (P P' : Program),
 P = (H, (R, (a, iseq (mov rd rs) I))) ->
 WFProgram P ->
 TrProgram P S -> Eval P P' -> WFProgram P' -> TrProgram P' (Step S).
Proof.
intros St.
intros.
rewrite H0 in H1.
inversion H1.
clear H14.
rewrite H0 in H2.
inversion H2.
unfold Step in |- * .
cut (TrInstrSeq L (iseq (mov rd rs) I) M pc). intros.
inversion H25.
rewrite H30.
inversion H28.
rewrite H0 in H3.
inversion H3.
apply trprogram with L; auto.
unfold updatereg in |- * .
constructor.
intros.
compare _r _rd.
intros.
replace (beq_reg _r _rd) with true.
replace (_R _rs + 0) with (0 + _R _rs).
unfold plus in |- * .
replace v0 with v.
inversion H22.
apply H52 with rs; auto.
apply (mlookup_eq _ _ r v v0 R').
apply eqorneq_reg.
apply mupdext_imp_mlookup with R.
apply eqorneq_reg.
replace r with rd; auto.
apply trregf_eq_eq with _r.
rewrite e; auto.
auto.
auto.
apply plus_comm.
apply sym_eq.
rewrite e.
apply beq_reg_eq_true; auto.
intros.
replace (beq_reg _r _rd) with false.
inversion H22.
apply H52 with r.
apply mupdext_preserv_rev with R' rd v; auto.
apply trregf_neq_neq with _r _rd; auto.
auto.
auto.
constructor.
unfold updatereg in |- * .
replace (beq_reg _r31 _rd) with false.
inversion H23.
auto.
apply sym_eq.
apply beq_reg_neq_false.
cut (_rd <> _r31).
auto.
apply trreg_neq_31 with rd.
auto.
apply tr_pc_Spc with (mov rd rs).
auto.
apply tr_pc_trinstrseq with H.
auto.
Qed.



Lemma Preservation_movi :
 forall (S : (nat -> nat) * ((_Reg -> nat) * nat)) 
   (H : Heap) (R : RegFile) (a : AP) (rd : Reg) (i : nat) 
   (I : InstrSeq) (P P' : Program),
 P = (H, (R, (a, iseq (movi rd i) I))) ->
 WFProgram P ->
 TrProgram P S -> Eval P P' -> WFProgram P' -> TrProgram P' (Step S).
Proof.
intros St.
intros.
rewrite H0 in H1.
inversion H1.
clear H14.
rewrite H0 in H2.
inversion H2.
unfold Step in |- * .
cut (TrInstrSeq L (iseq (movi rd i) I) M pc). intros.
inversion H25.
rewrite H30.
inversion H28.
rewrite H0 in H3.
inversion H3.
apply trprogram with L; auto.
unfold updatereg in |- * .
constructor.
intros.
compare _r0 _r.
intros.
replace (beq_reg _r0 _r) with true.
replace v with (wi i).
constructor.
apply (mlookup_eq _ _ rd (wi i) v R').
apply eqorneq_reg.
apply mupdext_imp_mlookup with R.
apply eqorneq_reg.
auto.
replace rd with r0.
auto.
apply trregf_eq_eq with _r0.
auto.
rewrite e.
auto.
apply sym_eq.
rewrite e.
apply beq_reg_eq_true; auto.
intros.
replace (beq_reg _r0 _r) with false.
inversion H22.
apply H50 with r0.
apply mupdext_preserv_rev with R' rd (wi i); auto.
apply trregf_neq_neq with _r0 _r; auto.
auto.
auto.
constructor.
unfold updatereg in |- * .
replace (beq_reg _r31 _r) with false.
inversion H23.
auto.
apply sym_eq.
apply beq_reg_neq_false.
cut (_r <> _r31).
auto.
apply trreg_neq_31 with rd.
auto.
apply tr_pc_Spc with (movi rd i).
auto.
apply tr_pc_trinstrseq with H.
auto.
Qed.



Lemma Preservation_movl :
 forall (S : (nat -> nat) * ((_Reg -> nat) * nat)) 
   (H : Heap) (R : RegFile) (a : AP) (r : Reg) (l : nat) 
   (I : InstrSeq) (P P' : Program),
 P = (H, (R, (a, iseq (movl r l) I))) ->
 WFProgram P ->
 TrProgram P S -> Eval P P' -> WFProgram P' -> TrProgram P' (Step S).
Proof.
intros St.
intros.
rewrite H0 in H1.
inversion H1.
clear H14.
rewrite H0 in H2.
inversion H2.
unfold Step in |- * .

cut (TrInstrSeq L (iseq (movl r l) I) M pc). intros.
inversion H25.
rewrite <- H30 in H28.
inversion H28.
clear H24.

rewrite H0 in H3.
inversion H3.
apply trprogram with L; auto.
unfold updatereg in |- * .
constructor.
intros.
cut (_r0 = _r \/ _r0 <> _r).
intros.
inversion_clear H49.
rewrite H50.
replace (beq_reg _r _r) with true.
replace v with (wl l).
constructor.
inversion H13.
apply mlookup_eq with Reg r R'.
apply eqorneq_reg.
apply mupdext_imp_mlookup with R.
apply eqorneq_reg.
auto.
replace r with r1; auto.
apply trregf_eq_eq with _r0; auto.
rewrite H50; auto.
apply sym_eq; apply beq_reg_eq_true; auto.
replace (beq_reg _r0 _r) with false; auto.
inversion H22.
apply H49 with r1.
apply mupdext_preserv_rev with R' r (wl l).
auto.
auto.
apply trregf_neq_neq with _r0 _r; auto.
auto.
auto.
constructor.
unfold updatereg in |- * .
replace (beq_reg _r31 _r) with false.
inversion H23.
auto.
apply sym_eq.
apply beq_reg_neq_false.
cut (_r <> _r31).
auto.
apply trreg_neq_31 with r.
auto.
rewrite <- H21 in H2.
inversion H2.
rewrite H19.
rewrite <- H54.
apply tr_pc_Spc with (movl r l).
auto.
apply tr_pc_trinstrseq with H.
auto.
Qed.



Lemma tr_update_layout_preserv :
 forall (h h' : Map nat HeapVal) (l : nat) (hv hv' : HeapVal),
 mlookup _ _ h l hv ->
 mupdate _ _ h l hv' h' ->
 HV_size hv = HV_size hv' -> Layout_aux h = Layout_aux h'.
Proof.
simple induction h.
intros.
inversion H.
intros.
inversion H1.
rewrite H3 in H0.
inversion H0.
unfold Layout_aux in |- * .
rewrite H2.
auto.
absurd (l <> l); auto.
unfold Layout_aux in |- * .
fold Layout_aux in |- * .
replace (Layout_aux m') with (Layout_aux m).
auto.
apply H with l hv hv'.
inversion H0.
rewrite H11 in H10.
absurd (l <> l); auto.
auto.
auto.
auto.
Qed.


Lemma tr_update_tuple_size_eq :
 forall (V V' : list WordVal) (i : nat) (w : WordVal),
 updatetuple V i w V' -> HV_size (tuple V) = HV_size (tuple V').
Proof.
intros.
induction H as [V w w'| V V' w w' i H HrecH].
unfold HV_size in |- * .
auto.
unfold HV_size in |- * .
unfold length in |- * .
unfold HV_size in HrecH.
auto.
Qed.


Lemma tr_update_wordvallist_aux :
 forall (L M : nat -> nat) (V : list WordVal) (n a w : nat),
 a < S n ->
 TrWordValList L V M (S n) -> TrWordValList L V (updatemem M a w) (S n).
Proof.
simple induction V.
intros.
constructor.
intros.
inversion H1.
apply trwvl_cons with w0.
auto.
unfold updatemem in |- * .
replace (beq_nat (S n) a0) with false.
auto.
apply beq_neq_false.
cut (a0 <> S n).
auto.
apply lt_neq.
auto.
apply H.
auto.
auto.
Qed.


Lemma tr_update_wordvallist_aux2 :
 forall (L M : nat -> nat) (i a n : nat) (V V' : list WordVal) 
   (wv : WordVal) (w : nat),
 TrWordVal L wv w ->
 a = n ->
 updatetuple V i wv V' ->
 TrWordValList L V M n -> TrWordValList L V' (updatemem M (a + i) w) n.
Proof.
simple induction i.
intros.
inversion H1.
apply trwvl_cons with w.
auto.
unfold updatemem in |- * .
replace (beq_nat n (a + 0)) with true.
auto.
replace (a + 0) with n.
apply beq_nat_refl.
rewrite H0.
auto.
rewrite <- H3 in H2.
inversion H2.
apply tr_update_wordvallist_aux.
replace (a + 0) with n.
apply lt_n_Sn.
rewrite H0.
auto.
auto.
intros.
inversion H2.
rewrite <- H5 in H3.
inversion H3.
apply trwvl_cons with w1.
auto.
unfold updatemem in |- * .
replace (beq_nat n0 (a + S n)) with false.
auto.
apply beq_neq_false.
rewrite H1.
apply plus_S_neq.
replace (a + S n) with (S a + n).
apply H with V0 wv.
auto.
auto.
auto.
auto.
apply plus_Snm_nSm.
Qed.




Lemma lt_Sm_n : forall m n : nat, S m < n -> m < n.
Proof.
simple induction m.
intros.
inversion H.
auto.
apply lt_O_Sn.
intros.
inversion H0.
auto.
apply lt_n_S.
apply H.
unfold lt in |- * .
apply le_Sn_le.
auto.
Qed.


Lemma label_lt_leftinc_neq :
 forall l l' m : nat, l + m < l' -> forall k : nat, k <= m -> l + k < l'.
Proof.
simple induction m.
intros.
inversion H0.
auto.
intros.
inversion H1.
auto.
apply H.
apply lt_Sm_n.
replace (S (l + n)) with (l + S n).
auto.
auto.
auto.
Qed.


Lemma label_lt_rightinc_neq :
 forall l l' n : nat, l' + n < l -> forall k : nat, l' + n < l + k.
Proof.
simple induction k.
replace (l + 0) with l.
auto.
auto.
intros.
replace (l + S n0) with (S (l + n0)).
apply lt_S.
auto.
auto.
Qed.


Lemma lt_plus_S : forall n k : nat, n < n + S k.
Proof.
simple induction n.
intros.
unfold plus in |- * .
apply lt_O_Sn.
intros.
replace (S n0 + S k) with (S (n0 + S k)).
apply lt_n_S.
apply H.
auto.
Qed.


Lemma lt_prop_aux :
 forall l l' i m n : nat,
 i <= n -> l + m < l' \/ l' + n < l -> l + m < l' + i \/ l' + i < l.
Proof.
intros.
inversion_clear H0.
left.
induction i as [| i Hreci].
replace (l' + 0) with l'.
auto.
auto.
replace (l' + S i) with (S (l' + i)).
apply lt_S.
apply Hreci.
apply le_Sn_le.
auto.
auto.
right.
inversion H.
auto.
cut (i < n).
intros.
apply lt_trans with (l' + n).
apply plus_le_lt_compat.
constructor.
auto.
auto.
unfold lt in |- * .
rewrite <- H2.
apply le_n_S.
auto.
Qed.



Lemma wf_propsubheap_wf :
 forall h h' : Map nat HeapVal, mWF _ _ h -> PropSubHeap h' h -> mWF _ _ h'.
Proof.
simple induction h.
intros.
inversion H0.
intros.
inversion H1.
inversion H0.
auto.
apply H.
inversion H0.
auto.
auto.
Qed.


Lemma wf_subheap_wf :
 forall h h' : Map nat HeapVal, mWF _ _ h -> SubHeap h' h -> mWF _ _ h'.
Proof.
intros.
inversion H0.
auto.
apply wf_propsubheap_wf with h.
auto.
auto.
Qed.

Lemma ord_propsubheap_ord :
 forall h h' : Map nat HeapVal, OrdHeap h -> PropSubHeap h' h -> OrdHeap h'.
Proof.
simple induction h.
intros.
inversion H0.
intros.
inversion H1.
inversion H0.
auto.
apply H.
inversion H0.
auto.
auto.
Qed.


Lemma ord_subheap_ord :
 forall h h' : Map nat HeapVal, OrdHeap h -> SubHeap h' h -> OrdHeap h'.
Proof.
intros.
inversion H0.
auto.
apply ord_propsubheap_ord with h.
auto.
auto.
Qed.


Lemma size_propsubheap_lt :
 forall h h' : Map nat HeapVal,
 PropSubHeap h' h ->
 forall s s' : nat, OrdHeap h -> msize _ _ h s -> msize _ _ h' s' -> s' < s.
Proof.
intros h h' H.
induction H as [h l hv| h h' l hv H HrecH].
intros.
replace s with (S s').
auto.
apply (msize_eq _ _ (melem nat HeapVal l hv h)).
constructor.
auto.
auto.
intros.
inversion H1.
apply lt_S.
apply HrecH.
inversion H0.
auto.
auto.
auto.
Qed.


Lemma ord_propsubheap_mlookup :
 forall (h h' : Map nat HeapVal) (s l : nat) (hv : HeapVal),
 OrdHeap h ->
 PropSubHeap h' h ->
 msize _ _ h' s -> l < s -> mlookup _ _ h l hv -> mlookup _ _ h' l hv.
Proof.
simple induction h.
intros.
inversion H0.
intros.
inversion H1.
inversion H4.
absurd (l = a).
apply lt_neq.
replace a with s.
auto.
inversion H0.
rewrite <- H9 in H16.
apply (msize_eq _ _ h').
auto.
auto.
auto.
auto.
apply H with s.
inversion H0.
auto.
auto.
auto.
auto.
inversion H4.
absurd (l = a).
apply lt_neq.
apply lt_trans with s.
auto.
inversion H0.
apply size_propsubheap_lt with m h'.
auto.
auto.
auto.
auto.
auto.
auto.
Qed.


Lemma ord_propsubheap_mindom :
 forall (h h' : Map nat HeapVal) (s l : nat),
 OrdHeap h ->
 PropSubHeap h' h ->
 msize _ _ h' s -> l < s -> mindom _ _ h l -> mindom _ _ h' l.
Proof.
simple induction h.
intros.
inversion H0.
intros.
inversion H1.
inversion H4.
absurd (l = a).
apply lt_neq.
replace a with s.
auto.
inversion H0.
rewrite <- H9 in H15.
apply (msize_eq _ _ h').
auto.
auto.
auto.
auto.
apply H with s.
inversion H0.
auto.
auto.
auto.
auto.
inversion H4.
absurd (l = a).
apply lt_neq.
apply lt_trans with s.
auto.
inversion H0.
apply size_propsubheap_lt with m h'.
auto.
auto.
auto.
auto.
auto.
auto.
Qed.



Lemma ord_heap_label_notin :
 forall (h : Map nat HeapVal) (s t : nat),
 OrdHeap h -> msize _ _ h s -> s <= t -> mnotindom _ _ h t.
Proof.
simple induction h.
intros.
constructor.
intros.
constructor.
inversion H1.
cut (n < t).
intros.
replace a with n.
apply lt_neq.
auto.
apply (msize_eq _ _ m).
auto.
inversion H0.
auto.
rewrite <- H4 in H2.
auto.
apply H with a.
inversion H0.
auto.
inversion H0.
auto.
apply le_trans with s.
inversion H1.
replace a with n.
auto.
apply (msize_eq _ _ m).
auto.
inversion H0.
auto.
auto.
Qed.



Lemma layout_aux_eq_top_label :
 forall (h : Map nat HeapVal) (wfh : mWF _ _ h) (ordh : OrdHeap h)
   (l s : nat),
 msize _ _ h s ->
 l < s ->
 exists m : _,
   msize _ _ m l /\ SubHeap m h /\ Layout_aux h l = Layout_aux m l.
Proof.
simple induction h.
intros.
inversion H.
rewrite <- H2 in H0.
inversion H0.
intros.
inversion H1.
cut (l = a).
intros.
rewrite H3.
split with m.
split.
inversion ordh.
auto.
split.
constructor.
constructor.
unfold Layout_aux at 1 in |- * .
replace (blt_nat a a) with false.
auto.
apply sym_eq.
apply blt_eq_false.
cut (S a = s).
intros.
rewrite <- H2 in H3.
apply eq_add_S.
auto.
apply (msize_eq _ _ (melem _ _ a b m)).
constructor.
inversion ordh.
auto.
auto.
cut
 (exists m0 : _,
    msize _ _ m0 l /\ SubHeap m0 m /\ Layout_aux m l = Layout_aux m0 l).
intros.
inversion_clear H4.
split with x.
split.
inversion_clear H5.
auto.
split.
inversion_clear H5.
inversion_clear H6.
inversion H5.
constructor.
constructor.
constructor.
constructor.
auto.
inversion_clear H5.
inversion_clear H6.
unfold Layout_aux at 1 in |- * .
replace (blt_nat a l) with false.
auto.
apply sym_eq.
apply blt_false.
rewrite <- H3 in H1.
replace a with m0.
auto.
cut (S a = s).
intros.
apply eq_add_S.
rewrite H3.
rewrite H6.
auto.
apply (msize_eq _ _ (melem _ _ a b m)).
constructor.
inversion ordh.
auto.
auto.
apply H with m0.
inversion wfh.
auto.
inversion ordh.
auto.
rewrite <- H3 in H0.
inversion H0.
auto.
auto.
Qed.



Lemma layout_aux_lt_eq :
 forall (h : Map nat HeapVal) (a a' s : nat),
 OrdHeap h ->
 msize _ _ h s -> s <= a -> s <= a' -> Layout_aux h a = Layout_aux h a'.
Proof.
simple induction h.
intros.
auto.
intros.
unfold Layout_aux in |- * .
replace (blt_nat a a0) with true.
replace (blt_nat a a') with true.
fold Layout_aux in |- * .
replace (Layout_aux m a0) with (Layout_aux m a').
auto.
induction m as [| a1 b0 m Hrecm].
auto.
apply H with a.
inversion H0.
auto.
inversion H0.
auto.
apply le_Sn_le.
replace (S a) with s.
auto.
apply (msize_eq _ _ (melem nat HeapVal a b (melem nat HeapVal a1 b0 m))).
auto.
constructor.
inversion H0.
auto.
apply le_Sn_le.
replace (S a) with s.
auto.
apply (msize_eq _ _ (melem nat HeapVal a b (melem nat HeapVal a1 b0 m))).
auto.
constructor.
inversion H0.
auto.
apply sym_eq.
apply blt_lt_true.
unfold lt in |- * .
replace (S a) with s.
auto.
apply (msize_eq _ _ (melem nat HeapVal a b m)).
auto.
constructor.
inversion H0.
auto.
apply sym_eq.
apply blt_lt_true.
unfold lt in |- * .
replace (S a) with s.
auto.
apply (msize_eq _ _ (melem nat HeapVal a b m)).
auto.
constructor.
inversion H0.
auto.
Qed.


Lemma layout_aux_le_sub :
 forall (h : Map nat HeapVal) (l : nat),
 OrdHeap h -> msize _ _ h (S l) -> Layout_aux h l <= Layout_aux h (S l).
Proof.
simple induction h.
intros.
inversion H0.
intros.
replace a with l.
unfold Layout_aux at 1 in |- * .
replace (blt_nat l l) with false.
fold Layout_aux in |- * .
unfold Layout_aux at 2 in |- * .
replace (blt_nat l (S l)) with true.
fold Layout_aux in |- * .
replace (Layout_aux m (S l)) with (Layout_aux m l).
apply le_plus_l.
apply layout_aux_lt_eq with l.
inversion H0.
auto.
inversion H1.
auto.
auto.
auto.
apply sym_eq.
apply blt_lt_true.
auto.
apply sym_eq.
apply blt_eq_false.
inversion H0.
inversion H1.
apply (msize_eq _ _ m).
auto.
auto.
Qed.



Lemma layout_aux_plus_le_sub :
 forall (h : Map nat HeapVal) (l : nat) (hv : HeapVal),
 OrdHeap h ->
 msize _ _ h (S l) ->
 mlookup _ _ h l hv -> Layout_aux h l + HV_size hv <= Layout_aux h (S l).
Proof.
simple induction h.
intros.
inversion H0.
intros.
replace a with l.
unfold Layout_aux at 1 in |- * .
replace (blt_nat l l) with false.
fold Layout_aux in |- * .
unfold Layout_aux at 2 in |- * .
replace (blt_nat l (S l)) with true.
fold Layout_aux in |- * .
replace (Layout_aux m (S l)) with (Layout_aux m l).
replace b with hv.
auto.
cut (a = l).
intros.
rewrite H3 in H2.
inversion H2.
auto.
absurd (l <> l); auto.
apply eq_add_S.
apply (msize_eq _ _ (melem _ _ a b m)).
constructor.
inversion H0.
auto.
auto.
apply layout_aux_lt_eq with l.
inversion H0.
auto.
inversion H1.
auto.
auto.
auto.
apply sym_eq.
apply blt_lt_true.
auto.
apply sym_eq.
apply blt_eq_false.
apply eq_add_S.
apply (msize_eq _ _ (melem _ _ a b m)).
auto.
constructor.
inversion H0.
auto.
Qed.


Lemma ordheap_mindom_lt_size :
 forall (h : Map nat HeapVal) (l s : nat),
 OrdHeap h -> msize _ _ h s -> mindom _ _ h l -> l < s.
Proof.
simple induction h.
intros.
inversion H1.
intros.
cut (a = l \/ a <> l).
intros.
inversion_clear H3.
rewrite <- H4.
inversion H1.
replace n with a.
auto.
apply (msize_eq _ _ m).
inversion H0.
auto.
auto.
inversion H1.
apply lt_S.
apply H.
inversion H0.
auto.
auto.
inversion H2.
rewrite H9 in H4.
absurd (l <> l); auto.
auto.
apply eqorneq_nat.
Qed.


Lemma layout_aux_propsubheap_eq :
 forall (h h' : Map nat HeapVal) (l s : nat),
 OrdHeap h ->
 PropSubHeap h' h ->
 msize _ _ h' s -> l < s -> Layout_aux h' l = Layout_aux h l.
Proof.
simple induction h.
intros.
inversion H0.
intros.
inversion H1.
unfold Layout_aux at 2 in |- * .
replace (blt_nat a l) with false.
auto.
apply sym_eq.
apply blt_false.
replace a with s.
auto.
apply (msize_eq _ _ m).
rewrite <- H8.
auto.
inversion H0.
auto.
unfold Layout_aux at 2 in |- * .
replace (blt_nat a l) with false.
fold Layout_aux in |- * .
apply H with s.
inversion H0.
auto.
auto.
auto.
auto.
apply sym_eq.
apply blt_false.
apply lt_trans with s.
auto.
apply size_propsubheap_lt with m h'.
auto.
inversion H0.
auto.
inversion H0.
auto.
auto.
Qed.


Lemma layout_aux_top_label_plus_nonoverlap :
 forall (s l : nat) (h : Map nat HeapVal) (hv : HeapVal),
 mWF _ _ h ->
 OrdHeap h ->
 msize _ _ h (S s) ->
 mlookup _ _ h l hv -> l < s -> Layout_aux h l + HV_size hv <= Layout_aux h s.
Proof.
simple induction s.
intros.
inversion H3.
intros.
inversion H2.
cut (S n = a).
intros.
unfold Layout_aux at 2 in |- * .
replace (blt_nat a (S n)) with false.
fold Layout_aux in |- * .
cut (l = n \/ l <> n).
intros.
inversion_clear H9.
rewrite H10.
rewrite <- H8.
inversion H7.
replace a0 with n.
replace hv with b0.
unfold Layout_aux at 1 in |- * .
replace (blt_nat (S n) n) with false.
replace (blt_nat n n) with false.
fold Layout_aux in |- * .
unfold Layout_aux at 2 in |- * .
replace (blt_nat n (S n)) with true.
fold Layout_aux in |- * .
replace (Layout_aux m0 n) with (Layout_aux m0 (S n)).
auto.
apply layout_aux_lt_eq with n.
rewrite <- H11 in H6.
rewrite <- H6 in H1.
inversion H1.
inversion H17.
auto.
auto.
auto.
auto.
apply sym_eq.
apply blt_lt_true.
auto.
apply sym_eq.
apply blt_eq_false.
apply sym_eq.
apply blt_false.
auto.
apply (mlookup_eq _ _ l b0 hv h).
apply eqorneq_nat.
rewrite <- H6.
rewrite <- H11.
replace a0 with l.
apply mlooknext.
apply mlookfound.
rewrite <- H8.
rewrite H10.
auto.
rewrite H10.
replace a0 with n.
auto.
apply (msize_eq _ _ m0).
auto.
rewrite <- H11 in H6.
rewrite <- H6 in H1.
inversion H1.
inversion H17.
auto.
auto.
apply (msize_eq _ _ m0).
auto.
rewrite <- H11 in H6.
rewrite <- H6 in H1.
inversion H1.
inversion H17.
auto.
apply le_trans with (Layout_aux m n).
replace (Layout_aux (melem nat HeapVal a b m) l) with (Layout_aux m l).
apply H.
rewrite <- H6 in H0.
inversion H0.
auto.
rewrite <- H6 in H1.
inversion H1.
auto.
auto.
rewrite <- H6 in H3.
inversion H3.
absurd (l = a).
rewrite <- H8.
apply lt_neq.
auto.
auto.
auto.
apply lt_S_neq.
auto.
auto.
rewrite <- H8.
unfold Layout_aux at 2 in |- * .
replace (blt_nat (S n) l) with false.
auto.
apply sym_eq.
apply blt_false.
auto.
apply layout_aux_le_sub.
rewrite <- H6 in H1.
inversion H1.
auto.
auto.
apply eqorneq_nat.
apply sym_eq.
rewrite H8.
apply blt_eq_false.
rewrite <- H6 in H1.
inversion H1.
apply (msize_eq _ _ m).
auto.
auto.
Qed.


Lemma layout_aux_label_plus_nonoverlap :
 forall (h : Map nat HeapVal) (l l' : nat) (hv : HeapVal),
 mWF _ _ h ->
 OrdHeap h ->
 mlookup _ _ h l hv ->
 mindom _ _ h l' -> l < l' -> Layout_aux h l + HV_size hv <= Layout_aux h l'.
Proof.
intros.
cut
 (exists h' : Map nat HeapVal,
    msize nat HeapVal h' l' /\
    SubHeap h' h /\ Layout_aux h l' = Layout_aux h' l').
intros.
elim H4; clear H4; intros h' H4.
inversion_clear H4.
inversion_clear H6.
inversion H4.
absurd (mnotindom _ _ h l').
apply mindom_imp_not_mnotindom.
apply eqorneq_nat.
auto.
apply ord_heap_label_notin with l'.
auto.
rewrite <- H8.
auto.
auto.
clear H8 H9 h0 h'0.
cut (exists s : _, msize _ _ h s).
intros.
elim H8; clear H8; intros s H8.
replace (Layout_aux h l) with (Layout_aux h' l).
rewrite H7.
inversion H5.
absurd (mnotindom _ _ h' l).
apply mindom_imp_not_mnotindom.
apply eqorneq_nat.
(* Apply wf_subheap_wf with h. 
Auto.
Auto. *)
apply ord_propsubheap_mindom with h l'.
auto.
auto.
auto.
auto.
apply mlookup_imp_mindom with hv.
auto.
rewrite <- H9.
constructor.
cut (n = a).
intros.
cut (l = a \/ l <> a).
intros.
inversion_clear H13.
rewrite H14.
rewrite H12.
apply layout_aux_plus_le_sub.
rewrite H10.
apply ord_subheap_ord with h.
auto.
auto.
auto.
rewrite H10.
rewrite <- H12.
rewrite H11.
auto.
rewrite H10.
rewrite <- H14.
apply ord_propsubheap_mlookup with h l'.
auto.
auto.
auto.
auto.
auto.
cut (l < n).
intros.
rewrite H12.
apply le_trans with (Layout_aux (melem _ _ a b m) a).
apply layout_aux_top_label_plus_nonoverlap.
apply wf_subheap_wf with h.
auto.
rewrite H10.
auto.
apply ord_subheap_ord with h.
auto.
rewrite H10.
auto.
constructor.
rewrite <- H12.
auto.
rewrite H10.
apply ord_propsubheap_mlookup with h l'.
auto.
auto.
auto.
auto.
auto.
rewrite <- H12.
auto.
apply layout_aux_le_sub.
apply ord_subheap_ord with h.
auto.
rewrite H10.
auto.
constructor.
rewrite <- H12.
auto.
apply lt_S_neq.
rewrite H11.
auto.
rewrite H12.
auto.
apply eqorneq_nat.
apply (msize_eq _ _ m).
auto.
cut (OrdHeap (melem _ _ a b m)).
intros.
inversion H12.
auto.
apply ord_subheap_ord with h.
auto.
rewrite H10.
auto.
apply layout_aux_propsubheap_eq with l'.
auto.
auto.
auto.
auto.
apply msize_always.
cut (exists s : nat, msize nat HeapVal h s).
intros.
inversion_clear H4.
apply layout_aux_eq_top_label with x.
auto.
auto.
auto.
apply ordheap_mindom_lt_size with h.
auto.
auto.
auto.
apply msize_always.
Qed.


Lemma updatetuple_list_len_eq :
 forall (V V' : list WordVal) (i : nat) (wv : WordVal),
 updatetuple V i wv V' -> length V = length V'.
Proof.
simple induction V.
intros.
inversion H.
intros.
inversion H0.
unfold length in |- * .
apply eq_S.
auto.
unfold length in |- * .
apply eq_S.
apply H with i0 wv.
auto.
Qed.


Lemma tuple_length_le_aux :
 forall (i : nat) (V V' : list WordVal) (wv : WordVal),
 updatetuple V i wv V' -> i < length V'.
Proof.
simple induction i.
intros.
inversion H.
unfold length in |- * .
apply lt_O_Sn.
intros.
inversion H0.
unfold length in |- * .
fold length in |- * .
apply lt_n_S.
apply H with V0 wv.
auto.
Qed.


Lemma tuple_length_le : forall i j k : nat, i < j -> S (k + i) <= k + j.
Proof.
simple induction i.
intros.
replace (k + 0) with k.
inversion H.
replace (k + 1) with (S k).
auto.
replace (k + 1) with (1 + k).
auto.
apply plus_comm.
replace (k + S m) with (S (k + m)).
apply le_n_S.
apply le_plus_l.
replace (k + S m) with (S m + k).
unfold plus at 2 in |- * .
replace (k + m) with (m + k).
auto.
apply plus_comm.
apply plus_comm.
auto.
intros.
inversion H0.
replace (k + S (S n)) with (S (S (k + n))).
replace (S (k + S n)) with (S (S (k + n))).
auto.
replace (k + S n) with (S n + k).
unfold plus at 2 in |- * .
fold plus in |- * .
replace (n + k) with (k + n).
auto.
apply plus_comm.
apply plus_comm.
replace (k + S (S n)) with (S (S n) + k).
unfold plus at 2 in |- * .
fold plus in |- * .
replace (n + k) with (k + n).
auto.
apply plus_comm.
apply plus_comm.
inversion H1.
replace (k + S n) with (S n + k).
replace (k + S (S (S n))) with (S (S (S n)) + k).
unfold plus in |- * .
auto.
apply plus_comm.
apply plus_comm.
replace (S (k + S n)) with (S (S (k + n))).
replace (k + S (S m0)) with (S (k + S m0)).
apply le_n_S.
apply H.
unfold lt in |- * .
apply le_S.
apply le_trans with (S (S n)).
auto.
auto.
replace (k + S m0) with (S m0 + k).
replace (k + S (S m0)) with (S (S m0) + k).
unfold plus in |- * .
auto.
apply plus_comm.
apply plus_comm.
replace (k + S n) with (S n + k).
replace (k + n) with (n + k).
unfold plus in |- * .
auto.
apply plus_comm.
apply plus_comm.
Qed.


Lemma layout_non_overlap :
 forall (L : nat -> nat) (H : Heap) (HT : HeapTy) (l l' : nat)
   (hv hv' : HeapVal),
 WFHeap H HT ->
 L = Layout H ->
 mlookup nat HeapVal (unwrapMap nat HeapVal H) l hv ->
 mlookup nat HeapVal (unwrapMap nat HeapVal H) l' hv' ->
 l <> l' -> L l + HV_size hv <= L l' \/ L l' + HV_size hv' <= L l.
Proof.
intros.
cut (l < l' \/ l' < l).
intros.
inversion_clear H5.
left.
rewrite H1.
unfold Layout in |- * .
apply layout_aux_label_plus_nonoverlap.
induction H as (m, m0).
unfold unwrapMap in |- * .
auto.
inversion H0.
inversion_clear H7.
inversion_clear H10.
inversion_clear H11.
inversion_clear H12.
inversion_clear H13.
inversion_clear H14.
inversion_clear H15.
inversion_clear H16.
auto.
auto.
apply mlookup_imp_mindom with hv'.
auto.
auto.
right.
rewrite H1.
unfold Layout in |- * .
apply layout_aux_label_plus_nonoverlap.
induction H as (m, m0).
unfold unwrapMap in |- * .
auto.
inversion H0.
inversion_clear H7.
inversion_clear H10.
inversion_clear H11.
inversion_clear H12.
inversion_clear H13.
inversion_clear H14.
inversion_clear H15.
inversion_clear H16.
auto.
auto.
apply mlookup_imp_mindom with hv.
auto.
auto.
apply nat_total_order.
auto.
Qed.



Lemma tr_update_wordvallist_preserv :
 forall (L M : nat -> nat) (l' w : nat) (V : list WordVal) (l : nat),
 l + length V <= l' \/ l' < l ->
 TrWordValList L V M l -> TrWordValList L V (updatemem M l' w) l.
Proof.
simple induction V.
intros.
constructor.
intros.
inversion H1.
apply trwvl_cons with w0.
auto.
unfold updatemem in |- * .
replace (beq_nat l0 l') with false.
auto.
apply beq_neq_false.
inversion H0.
apply lt_neq.
cut (l0 + length (a :: l) = S (l0 + length l)).
intros.
rewrite H11 in H10.
unfold lt in |- * .
apply le_trans with (S (l0 + length l)).
apply le_n_S.
apply le_plus_l.
auto.
unfold length at 1 in |- * .
auto.
cut (l' <> l0).
auto.
apply lt_neq.
auto.
apply H.
inversion H0.
left.
replace (S l0 + length l) with (l0 + length (a :: l)).
auto.
unfold length in |- * .
unfold plus in |- * .
auto.
right.
apply lt_S.
auto.
auto.
Qed.





Lemma tr_update_instrseq_preserv :
 forall (L M : nat -> nat) (l' w : nat) (IS : InstrSeq) (l : nat),
 l + lenInstrSeq IS <= l' \/ l' < l ->
 TrInstrSeq L IS M l -> TrInstrSeq L IS (updatemem M l' w) l.
Proof.
simple induction IS.
intros.
inversion H1.
apply tris_iseq with _i.
auto.
unfold updatemem in |- * .
replace (beq_nat l l') with false.
auto.
apply beq_neq_false.
inversion H0.
cut (l + lenInstrSeq (iseq i i0) = S (l + lenInstrSeq i0)).
intros.
rewrite H11 in H10.
apply lt_neq.
unfold lt in |- * .
apply le_trans with (S (l + lenInstrSeq i0)).
apply le_n_S.
apply le_plus_l.
auto.
unfold lenInstrSeq at 1 in |- * .
auto.
cut (l' <> l).
auto.
apply lt_neq.
auto.
apply H.
inversion H0.
left.
replace (S l + lenInstrSeq i0) with (l + lenInstrSeq (iseq i i0)).
auto.
replace (S l + lenInstrSeq i0) with (l + lenInstrSeq (iseq i i0)).
auto.
unfold lenInstrSeq in |- * .
unfold plus in |- * .
auto.
right.
apply lt_S.
auto.
auto.

intros.
constructor.
inversion H0.
unfold updatemem in |- * .
replace (beq_nat l l') with false.
auto.
apply beq_neq_false.
inversion H.
apply lt_neq.
unfold lenInstrSeq in H6.
cut (l + 1 = S l).
intros.
rewrite H7 in H6.
auto.
replace (l + 1) with (1 + l).
auto.
apply plus_comm.
cut (l' <> l).
auto.
apply lt_neq.
auto.

intros.
inversion H0.
apply tris_jmp with _r.
auto.
unfold updatemem in |- * .
replace (beq_nat l l') with false.
auto.
apply beq_neq_false.
inversion H.
apply lt_neq.
unfold lenInstrSeq in H7.
cut (l + 1 = S l).
intros.
rewrite H8 in H7.
auto.
replace (l + 1) with (1 + l).
auto.
apply plus_comm.
cut (l' <> l).
auto.
apply lt_neq.
auto.
Qed.




Lemma Preservation_st :
 forall (S : (nat -> nat) * ((_Reg -> nat) * nat)) 
   (H : Heap) (R : RegFile) (a : AP) (rd rs : Reg) 
   (i : nat) (I : InstrSeq) (P P' : Program),
 P = (H, (R, (a, iseq (st rd i rs) I))) ->
 WFProgram P ->
 TrProgram P S -> Eval P P' -> WFProgram P' -> TrProgram P' (Step S).
Proof.
intros St.
intros.
rewrite H0 in H1.
inversion H1.
rewrite H0 in H2.
inversion H2.
unfold Step in |- * .
cut (TrInstrSeq L (iseq (st rd i rs) I) M pc).
intros.
inversion H26.
rewrite <- H31 in H29.
inversion H29.
clear H25.
rewrite H0 in H3.
inversion H3.
apply trprogram with L.
rewrite H20.
unfold Layout, unwrapMap in |- * .
rewrite H49.
rewrite H51.
apply tr_update_layout_preserv with l1 (tuple V) (tuple V').
auto.
auto.
apply tr_update_tuple_size_eq with i w.
auto.
inversion H21.
apply trheap with s.
unfold unwrapMap in |- * .
rewrite H51.
apply mupdate_size_same with h l1 (tuple V').
apply eqorneq_nat.
auto.
rewrite H49 in H58.
auto.
intros.
cut (n = l1 \/ n <> l1).
intros.
rewrite H51 in H63.
unfold unwrapMap in H63.
inversion_clear H64.
replace h0 with (tuple V').
cut (TrHeapVal L (tuple V) M (L n)).
intros.
constructor.
inversion H64.
apply tr_update_wordvallist_aux2 with V w.
inversion H23.
apply H71 with rs.
auto.
auto.
cut (TrWordVal L (wl n) (_R _rd)).
intros.
inversion H71.
auto.
inversion H23.
apply H71 with rd.
rewrite H65.
auto.
auto.
auto.
auto.
apply H59.
rewrite H49.
unfold unwrapMap in |- * .
rewrite H65.
auto.

apply mlookup_eq with nat n h'.
apply eqorneq_nat.
apply mupdate_imp_mlookup with h.
apply eqorneq_nat.
rewrite H65.
auto.
auto.
cut (TrHeapVal L h0 M (L n)).
intros.

induction h0 as [l2| m a0 i3].
constructor.
apply tr_update_wordvallist_preserv.
cut (L n + HV_size (tuple l2) <= L l1 \/ L l1 + HV_size (tuple V') <= L n).
intros.
inversion_clear H66.
left.
unfold HV_size in H67.
replace (_R _rd) with (L l1).
apply le_trans with (L l1).
auto.
apply le_plus_l.
inversion H23.
cut (TrWordVal L (wl l1) (_R _rd)).
intros.
inversion H71.
auto.
apply H66 with rd.
auto.
auto.
right.
unfold lt in |- * .
replace (_R _rd) with (L l1).
apply le_trans with (L l1 + HV_size (tuple V')).
unfold HV_size in |- * .
apply tuple_length_le.
apply tuple_length_le_aux with V w.
auto.
auto.
inversion H23.
cut (TrWordVal L (wl l1) (_R _rd)).
intros.
inversion H71.
auto.
apply H66 with rd.
auto.
auto.
rewrite <- H50 in H4.
inversion H4.
apply layout_non_overlap with H' HT0.
auto.
rewrite H20.
unfold Layout in |- * .
apply tr_update_layout_preserv with l1 (tuple V) (tuple V').
rewrite H49.
auto.
rewrite H49.
rewrite H51.
auto.
unfold HV_size in |- * .
apply updatetuple_list_len_eq with i w.
auto.
rewrite H51.
auto.
rewrite H51.
unfold unwrapMap in |- * .
apply mupdate_imp_mlookup with h.
apply eqorneq_nat.
auto.
auto.

inversion H64.
auto.
inversion H64.
constructor.
apply tr_update_instrseq_preserv.
cut
 (L n + HV_size (code m a0 i3) <= L l1 \/ L l1 + HV_size (tuple V') <= L n).
intros.
inversion_clear H73.
left.
unfold HV_size in H74.
replace (_R _rd) with (L l1).
apply le_trans with (L l1).
auto.
apply le_plus_l.
inversion H23.
cut (TrWordVal L (wl l1) (_R _rd)).
intros.
inversion H78.
auto.
apply H73 with rd.
auto.
auto.
right.
unfold lt in |- * .
replace (_R _rd) with (L l1).
apply le_trans with (L l1 + HV_size (tuple V')).
unfold HV_size in |- * .
apply tuple_length_le.
apply tuple_length_le_aux with V w.
auto.
auto.
inversion H23.
cut (TrWordVal L (wl l1) (_R _rd)).
intros.
inversion H78.
auto.
apply H73 with rd.
auto.
auto.
rewrite <- H50 in H4.
inversion H4.
apply layout_non_overlap with H' HT0.
auto.
rewrite H20.
unfold Layout in |- * .
apply tr_update_layout_preserv with l1 (tuple V) (tuple V').
rewrite H49.
auto.
rewrite H49.
rewrite H51.
auto.
unfold HV_size in |- * .
apply updatetuple_list_len_eq with i w.
auto.
rewrite H51.
auto.
rewrite H51.
unfold unwrapMap in |- * .
apply mupdate_imp_mlookup with h.
apply eqorneq_nat.
auto.
auto.
auto.
apply H59.
rewrite H49.
unfold unwrapMap in |- * .
apply mupdate_preserv_rev with h' l1 (tuple V').
apply eqorneq_nat.
auto.
auto.
auto.
apply eqorneq_nat.
auto.

auto.
rewrite <- H22 in H2.
inversion H2.
inversion_clear H69.
inversion_clear H70.
inversion_clear H69.
inversion_clear H70.
inversion_clear H69.
inversion_clear H70.
inversion_clear H71.
inversion_clear H72.
split with x.
split with x0.
split with x1.
split with x2.
split with (S x3).
cut (x <> l1).
intros.
split.
apply mupdate_preserv with h l1 (tuple V').
rewrite H51.
auto.
rewrite H49 in H69.
auto.
auto.
split.
apply tr_instrseq_subdepth_S with (st rd i rs).
auto.
split.
apply tr_update_instrseq_preserv.
cut (_R _rd = L l1).
intros.
cut
 (L x + HV_size (code x0 x1 x2) <= L l1 \/ L l1 + HV_size (tuple V') <= L x).
intros.
inversion_clear H75.
left.
unfold HV_size in H76.
rewrite H74.
apply le_trans with (L l1).
auto.
apply le_plus_l.
right.
unfold lt in |- *; rewrite H74.
apply le_trans with (L l1 + HV_size (tuple V')).
unfold HV_size in |- * .
apply tuple_length_le.
apply tuple_length_le_aux with V w.
auto.
auto.
rewrite <- H50 in H4.
inversion H4.
apply layout_non_overlap with H' HT0.
auto.
rewrite H20.
unfold Layout in |- * .
apply tr_update_layout_preserv with l1 (tuple V) (tuple V').
rewrite H49.
auto.
rewrite H49.
rewrite H51.
auto.
unfold HV_size in |- * .
apply updatetuple_list_len_eq with i w.
auto.
rewrite H51.
apply mupdate_preserv with h l1 (tuple V').
auto.
rewrite H49 in H69.
auto.
auto.
rewrite H51.
apply mupdate_imp_mlookup with h.
apply eqorneq_nat.
auto.
auto.
inversion H23.
cut (TrWordVal L (wl l1) (_R _rd)).
intros.
inversion H78.
auto.
apply H74 with rd.
auto.
auto.
rewrite H20.
rewrite <- H64.
auto.
replace (L x + S x3) with (S x3 + L x).
unfold plus in |- * .
fold plus in |- * .
replace (x3 + L x) with (L x + x3).
apply eq_S.
rewrite H20.
rewrite <- H64.
auto.
apply plus_comm.
apply plus_comm.
cut (x = l1 \/ x <> l1).
intros.
inversion_clear H72.
absurd (code x0 x1 x2 = tuple V).
discriminate.
apply (mlookup_eq _ _ x (code x0 x1 x2) (tuple V) h).
apply eqorneq_nat.
rewrite H49 in H69.
auto.
rewrite H74.
auto.
auto.
apply eqorneq_nat.
apply tr_pc_trinstrseq with H.
auto.
Qed.




Lemma Preservation_unfold :
 forall (S : (nat -> nat) * ((_Reg -> nat) * nat)) 
   (H : Heap) (R : RegFile) (a : AP) (rd rs : Reg) 
   (I : InstrSeq) (P P' : Program),
 P = (H, (R, (a, iseq (unfold rd rs) I))) ->
 WFProgram P ->
 TrProgram P S -> Eval P P' -> WFProgram P' -> TrProgram P' (Step S).
Proof.
intros St.
intros.
rewrite H0 in H1.
inversion H1.
clear H14.
rewrite H0 in H2.
inversion H2.
unfold Step in |- * .
cut (TrInstrSeq L (iseq (unfold rd rs) I) M pc).
intros.
inversion H25.
rewrite H30.
inversion H28.
rewrite H0 in H3.
inversion H3.
apply trprogram with L.
auto.
auto.
unfold updatereg in |- * .
constructor.
intros.
compare _r _rd.
intros.
replace (beq_reg _r _rd) with true.
replace (_R _rs + 0) with (0 + _R _rs).
unfold plus in |- * .
replace v0 with v.
inversion H22.
cut (TrWordVal L (wfold v t0) (_R _rs)).
intros.
inversion H56.
auto.
apply H52 with rs.
auto.
auto.
apply (mlookup_eq _ _ r v v0 R').
apply eqorneq_reg.
replace r with rd.
apply mupdext_imp_mlookup with R.
apply eqorneq_reg.
auto.
apply trregf_eq_eq with _r.
rewrite e.
auto.
auto.
auto.
apply plus_comm.
rewrite e.
apply sym_eq.
apply beq_reg_eq_true; auto.
intros.
replace (beq_reg _r _rd) with false.
inversion H22.
apply H52 with r.
apply mupdext_preserv_rev with R' rd v; 
auto.
apply trregf_neq_neq with _r _rd; auto.
auto.
auto.
constructor.
unfold updatereg in |- * .
replace (beq_reg _r31 _rd) with false.
inversion H23.
auto.
apply sym_eq.
apply beq_reg_neq_false.
cut (_rd <> _r31).
auto.
apply trreg_neq_31 with rd.
auto.
apply tr_pc_Spc with (unfold rd rs).
auto.
apply tr_pc_trinstrseq with H.
auto.
Qed.




Theorem Preservation :
 forall S : (nat -> nat) * ((_Reg -> nat) * nat), Inv S -> Inv (Step S).
Proof.
intros.
elim H; clear H; intros P H.
elim H; clear H; intros D H.
cut (exists P' : _, Eval P P').
intros.
elim H0; clear H0; intros P' H0.
cut (WFProgram P').
intros D'.
split with P'.
split; auto.
inversion D.
inversion H5.
apply Preservation_add with H1 R l rd rs rs' I0 P.
rewrite H15.
auto.
auto.
auto.
auto.
auto.
apply Preservation_addi with H1 R l rd rs i I0 P.
rewrite H14.
auto.
auto.
auto.
auto.
auto.
apply Preservation_alloc with H1 R l rd V I0 P.
rewrite H14.
auto.
auto.
auto.
auto.
auto.
apply Preservation_bgt with H1 R l rs rt l0 I0 P.
rewrite H16.
auto.
auto.
auto.
auto.
auto.
apply Preservation_bump with H1 R l n I0 P.
rewrite H12.
auto.
auto.
auto.
auto.
auto.
apply Preservation_fold with H1 R l rd rs (recty f) I0 P.
rewrite H15.
auto.
auto.
auto.
auto.
auto.
apply Preservation_jd with H1 R l l0 P.
rewrite H13.
auto.
auto.
auto.
auto.
auto.
apply Preservation_jmp with H1 R l r P.
rewrite H13.
auto.
auto.
auto.
auto.
auto.
apply Preservation_ld with H1 R l rd rs i I0 P.
rewrite H16.
auto.
auto.
auto.
auto.
auto.
apply Preservation_mov with H1 R l rd rs I0 P.
rewrite H14.
auto.
auto.
auto.
auto.
auto.
apply Preservation_movi with H1 R l rd i I0 P.
rewrite H13.
auto.
auto.
auto.
auto.
auto.
apply Preservation_movl with H1 R l rd l0 I0 P.
rewrite H15.
auto.
auto.
auto.
auto.
auto.
apply Preservation_st with H1 R l rd rs i I0 P.
rewrite H17.
auto.
auto.
auto.
auto.
auto.
apply Preservation_unfold with H1 R l rd rs I0 P.
rewrite H16.
auto.
auto.
auto.
auto.
auto.
apply preserv with P.
auto.
auto.
apply progress.
auto.
Qed.


(*

Proofs needed:
  H1 : (Inv S0).
  H2 : (S:State) (Inv S) -> (SP S).
  H3 : (S:State) (Inv S) -> (Inv (Step S)).

  H1 must be supplied for each individual program.
  H2 and H3 can be proven once and used for all TAL programs compiled to TIS.

  Thus:
*)

Theorem Safety_aux :
 forall (S : (nat -> nat) * ((_Reg -> nat) * nat)) (n : nat),
 Inv S -> Inv (MultiStep n S).
Proof.
simple induction n.
intros.
unfold MultiStep in |- * .
apply Preservation; auto.

intros.
unfold MultiStep in |- * .
fold MultiStep in |- * .
apply Preservation.
apply H.
auto.
Qed.


Theorem Safety :
 forall S : (nat -> nat) * ((_Reg -> nat) * nat), Inv S -> Safe S.
Proof.
unfold Safe in |- * .
simple induction n.
apply Progress.
unfold MultiStep in |- * .
apply Preservation; auto.

intros.
unfold MultiStep in |- * .
fold MultiStep in |- * .
apply Progress.
apply Preservation.
apply Safety_aux.
auto.
Qed.





(***************************************************************************)
(***************************************************************************)
